
'use strict';

import fs from 'fs';
import _ from 'lodash';
import path from 'path';
import promise from '../src/promise';
import exception from '../src/exception';

/**
 * @name _escape_special:
 *   Escape control characters in a Javascript string.
 *   This function is synchronous.
 */
let _escape_special = (_string, _escape_whitespace) => {

  const control = {
    0: '\0', b: '\b', f: '\f', v: '\v'
  };

  const whitespace = {
    n: '\n', r: '\r', t: '\t'
  };

  let rv = _string.toString();
  let replacements = _.extend({}, control);

  if (_escape_whitespace) {
    _.extend(replacements, whitespace);
  }

  /* Safe: _v is derived from const data */
  _.each(replacements, (_v, _k) => {
    rv = rv.replace(new RegExp(_v, 'g'), '\\' + _k);
  });

  return rv;
};

/**
 * @name escape_string:
 *   Escape a Javascript string, such that it's suitable for
 *   literal inclusion in source code. This function supports
 *   three different string-quoting styles; supply a single
 *   character to `_quote` (either '`', '"', or "'") to pick.
 *   This function is synchronous.
 */
let escape_string = (_string, _quote) => {

  let rv = '';
  let quote = (_quote || "'");

  switch (quote) {
    case '"':
      rv = _string.replace(/([\"\\])/g, '\\$1');
      break;
    case '`':
      rv = _string.replace(/([\`\$\{\}\\])/g, '\\$1');
      break;
    case "'":
      rv = _string.replace(/([\'\\])/g, '\\$1');
      break;
    default:
      throw new exception.argument('Invalid quote style');
  }

  rv = _escape_special(rv, true);
  return [ quote, rv, quote ].join('');
};

/**
 * @name read:
 *   Read a test fixture, given a test type in `_type`, a module
 *   name in `_module`, and a fixture name in `_name`. This
 *   function is asynchronous and returns a promise.
 */
let read = (_type, _module, _name) => {

  var o = {
    encoding: 'utf-8', flag: 'r'
  };

  var p = path.join(
    __dirname, '..',
      'tests', 'fixtures', _type, _module, _name
  );

  /* jshint node: true */
  return new promise.base((_resolve, _reject) => {
    fs.readFile(p, o, (_e, _data) => {
      return (_e ? _reject(_e) : _resolve(_data));
    });
  });
};

/**
 * @name populate:
 *   Use `read` to populate an entire object of fixtures, where
 *   fixture names are represented as property names. This function
 *   is asynchronous and returns a promise.
 */
let populate = (_type, _module, _fixtures) => {

  var promises = [];

  _.each(_.keys(_fixtures), (_name, _next_fn) => {
    promises.push(
      read(_type, _module, _name).then((_fixture) => {
        _fixtures[_name] = _fixture;
      })
    );
  });

  return promise.all(promises).then(() => {
    return _fixtures;
  });
};

/**
 * @name _compile:
 *   Compile a single already-populated fixture in to source
 *   code, suitable for literal inclusion in a Javascript
 *   source file. The name of the fixture should be provided in
 *   `_name`; the fixture content should be provided in `_data`.
 *   This function is synchronous.
 */
let _compile = (_name, _data) => {

  let rv = '';
  let type = path.extname(_name).replace(/^\./, '');

  switch (type.toLowerCase()) {
    case 'json':
      rv += 'JSON.parse(' + escape_string(_.trim(_data), "'") + ')';
      break;
    default:
    case 'txt':
      rv += escape_string(_data, '`');
      break;
  }

  return [ rv, ';', "\n" ].join('');
};

/**
 * @name _build_path:
 *   Create the left hand side of a fixture assignment. The
 *   value of `_lhs` is the root object beneath which to perform
 *   an assignment (defaults to `exports`), and `_path` is an
 *   array of nested properties beneath `_lhs` that the assignment
 *   should follow. This function is synchronous.
 */
let _build_path = (_lhs, _path) => {

  return [
    (_lhs || 'fixtures'), '[',
    _.map(_path, (_s) => escape_string(_s)).join(']['), ']'
  ].join('');
};

/**
 * @name compile:
 *   Compile all fixtures for the test type `_type`, for the
 *   module named `_module`, using the `_fixtures` object as a
 *   template. This function is asynchronous and returns a promise.
 */
let compile = (_type, _module,
               _fixtures, _visited_paths, _options) => {

  let options = (_options || {});

  let paths = [
    _build_path(options.lhs, [ _type ]),
    _build_path(options.lhs, [ _type, _module ])
  ];

  /* Avoid duplicates */
  if (_visited_paths[_type]) {
    paths.shift();
  }

  let header = 'let fixtures = {};';
  let footer = 'export default fixtures;';

  let initializers = _.map(paths, (_p) => {
    return [
      '/* istanbul ignore else */', '\n',
      'if (!', _p, ') { ', _p, ' = {}; }'
    ].join('');
  });

  return populate(_type, _module, _fixtures).then(() => {
    return [
      header,
      initializers.join('\n'),
      _.reduce(_fixtures, (_r, _fixture, _name) => {
        return [
          _r, _build_path(options.lhs, [ _type, _module, _name ]),
            ' = ', _compile(_name, _fixture)
        ].join('');
      }, ''),
      footer
    ].join('\n');
  });
};

/**
 * @name compile_tree:
 *   Compile all fixtures inside of a "fixture tree". A fixture
 *   tree is a hierarchical object keyed by test type, then by
 *   module name, then by fixture name. An example of a valid
 *   fixture tree is `{ unit: { splitter: { 'f.txt': true } } }`.
 *   This function is asynchronous and returns a promise.
 */
let compile_tree = (_fixture_tree) => {

  let rv = {};
  let promises = [];
  let fixtures = _.extend({}, _fixture_tree);

  _.each(fixtures, (_modules, _type) => {
    _.each(_modules, (_tests, _module) => {
      promises.push(
        compile(_type, _module, _tests, rv).then((_fixture) => {
          fixtures[_type][_module] = _fixture;
        })
      );
    });
  });

  return promise.all(promises).then(() => {
    return fixtures;
  });
};

/**
 * @name serialize_tree:
 *   Serialize an already-compiled "fixture tree" into a string
 *   of executable javascript code. A fixture tree is a hierarchical
 *   object keyed by test type, then by module name, then by test
 *   fixture name. This function is synchronous.
 */
let serialize_tree = (_fixture_tree, _lhs) => {

  var rv = '';

  _.each(_fixture_tree, (_modules, _type) => {
    _.each(_modules, (_tests, _module) => {
      rv += _tests;
    });
  });

  return rv;
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export {
  read,
  compile,
  populate,
  compile_tree,
  serialize_tree,
};

export default {
  read: read,
  compile: compile,
  populate: populate,
  compile_tree: compile_tree,
  serialize_tree: serialize_tree
};

