/* globals __dirname */

'use strict';

import fs from 'fs';
import path from 'path';
import process from 'process';
import { rollup } from 'rollup';
import babelrc from 'babelrc-rollup';
import flow from 'rollup-plugin-flow';
import babel from 'rollup-plugin-babel';
import commonjs from 'rollup-plugin-commonjs';
import node_resolve from 'rollup-plugin-node-resolve';

/**
 * @name fatal:
 *   Dump a message and stack trace to stderr and exit.
 */
let fatal = (_e, _message) => {

  process.stderr.write('Fatal: ' + _message + '\n');
  process.stderr.write(_e.stack + '\n');
  process.exit(1);
};

/**
 * @name dist_dir:
 */
let dist_dir = path.join(__dirname, '..', 'dist');

/**
 * @name app.js
 */
rollup({
  sourceMap: true,
  entry: 'src/api.js',
  external: [ 'pouchdb' ],
  plugins: [
    flow(),
    babel(
      babelrc('.babelrc-rollup')
    ),
    node_resolve({
      main: true,
      jsnext: true,
      browser: true,
      preferBuiltins: true
    }),
    commonjs({
      sourceMap: true,
      extensions: [ '.js' ],
      include: [
        'node_modules/**'
      ],
      exclude: [
        'node_modules/pouchdb/**'
      ]
    })
  ]
}).then((_bundle) => {

  let rv = _bundle.generate({
    format: 'umd',
    sourceMap: true,
    exports: 'named',
    moduleName: 'api',
    globals: {
      pouchdb: 'PouchDB'
    }
  });

  fs.writeFileSync(
    path.join(dist_dir, 'api.js'), rv.code
  );

  fs.writeFileSync(
    path.join(dist_dir, 'api.map'), rv.map
  );

}).catch((_e) => {
  fatal(_e, 'compilation failure (api.js)');
});

/**
 * @name tests.js
 */
rollup({
  sourceMap: true,
  entry: 'tests/unit/all.js',
  external: [ 'pouchdb', 'chai', 'mocha' ],
  plugins: [
    flow(),
    babel(
      babelrc('.babelrc-rollup')
    ),
    node_resolve({
      main: true,
      jsnext: true,
      browser: true,
      preferBuiltins: true
    }),
    commonjs({
      sourceMap: true,
      extensions: [ '.js' ],
      include: [
        'node_modules/**'
      ],
      exclude: [
        'node_modules/pouchdb/**'
      ]
    })
  ]
}).then((_bundle) => {

  let rv = _bundle.generate({
    format: 'umd',
    sourceMap: true,
    exports: 'named',
    moduleName: 'api',
    globals: {
      pouchdb: 'PouchDB',
      chai: 'chai', mocha: 'mocha'
    }
  });

  fs.writeFileSync(
    path.join(dist_dir, 'tests.js'), rv.code
  );

  fs.writeFileSync(
    path.join(dist_dir, 'tests.map'), rv.map
  );

}).catch((_e) => {
  fatal(_e, 'compilation failure (tests.js)');
});

