/* globals emit */

'use strict';

import _ from './utils/lodash';

/**
 * @name Base:
 *   A universal base class. Handles exceedingly-abstract
 *   shared functionality that's of use to all major components.
 */
var Base = class {

  /**
   * @name constructor:
   */
  constructor (_options) {

    this._set_options(_options);
    return this;
  }

  /**
   * @name get options:
   *   Return the set of options currently in use by this
   *   instance. This will return the options originally passed
   *   to the constructor, plus any changes made via `option`.
   */
  get options () {

    return this._options;
  }

  /**
   * @name set options:
   */
  set options (_options) {

    this._set_options(_options);
  }

  /**
   * @name option:
   *   Read or change the value of a single option. If `_value` is
   *   exactly `undefined`, locate the option using the lodash-style
   *   object path `_path` and return it. If `_value` is any other
   *   value than `undefined`, locate the option along `_path`, set
   *   it to `_value`, and return the value that was there previously.
   *   If you wish to restrict which options can be modified after
   *   instantiation, override this method in your subclass(es).
   */
  option (_path, _value) {

    let rv = _.get(this._options, _path);

    if (!_.isUndefined(_value)) {
      _.set(this._options, _path, _value);
    }

    return rv;
  }

  /** @protected **/

  /**
   * @name _set_options:
   */
  _set_options (_options) {

    this._options = _.clone(_options || {});
    return this;
  }
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Base;
export { Base as base };

