/* @flow */

'use strict';

import Base from './base';
import utils from './utils';
import schema from './schema';
import _ from './utils/lodash';
import promise from './promise';
import indexer from './indexer';
import exception from './exception';
import validator from './validator';
import renderer from './engine/renderer';
import transform from './schema/transform';
import join_iterator from './iterator/join';
import view_iterator from './iterator/view';
import array_iterator from './iterator/array';
import words_transform from './schema/transform/words';
import nodes_by_tag_transform from './schema/transform/nodes-by-tag';

/**
 * @name iterator:
 */
type iterator = Object;

/**
 * @name zone_id:
 */
type zone_id = (number | String);

/**
 * @name word_index_item:
 */
type word_index_item = [ string, number, number ];

/**
 * @name document_stub:
 */
type document_stub = {
  _id: string,
  _rev: string
};

/**
 * @name node_input:
 */
type node_input = {
  type: ?'node',
  title: string,
  content: string,
  zone_id: ?zone_id,
  tags: ?Array<string>
};

/**
 * @name node:
 */
type node = document_stub & node_input & {
  index_id: string
};

/**
 * @name index:
 */
type index = document_stub & {
  type: 'index',
  node_id: string,
  data: {
    content: string,
    words: Array<[ string, number, number ]>,
    tags: Array<[ string, word_index_item ]>
  },
  links: ?Array<[
    zone_id, string, string, number, number,
      Array<word_index_item>
  ]>
};

/**
 * @name node_stub:
 */
type node_stub = document_stub & {
  type: 'node'
};

/**
 * @name index_stub:
 */
type index_stub = document_stub & {
  type: 'index'
};

/**
 * @name node_and_index:
 */
type node_and_index = {
  node: node,
  index: index
};

/**
 * @name n

/**
 * @name Engine:
 */
var Engine = class extends Base {

  /**
   * @name constructor:
   */
  constructor (_options: ?Object) {

    /* Must be first */
    super(_options);

    this._is_running = false;
    this._default_result_limit = 16;
    this._validator = new validator();

    this._type = (this.option('type') || 'pouchdb');
    this._name = (this.option('name') || 'noteable');

    this._schema = new schema[this._type](
      (this.option('path') || this.name), (this.option('schema') || {})
    );

    this._indexer = new indexer.text(null, {
      sort: true, collator: this._schema
    });

    return this._index_cache_reset();
  }

  /**
   * @name get name:
   */
  get name (): string {

    return this._name;
  }

  /**
   * @name get options:
   */
  get options (): Object {

    return this._options;
  }

  /**
   * @name get type:
   *   Return a string describing the schema backend currently
   *   in use. This name corresponds to a `Schema` subclass.
   */
  get type (): string {

    return this._type;
  }

  /**
   * @name get is_running:
   *   Return true if the engine is running; false otherwise.
   */
  get is_running (): boolean {

    return this._is_running;
  }

  /**
   * @name schema:
   *   Return the `Schema` instance currently in use by this engine.
   */
  schema (): Object {

    return this._schema;
  }

  /**
   * @name indexer:
   *   Return the `Indexer` instance currently in use by this engine.
   */
  indexer (): Object {

    return this._indexer;
  }

  /**
   * @name start:
   *   Start the cross-referencing engine.
   */
  start () {

    return this.schema().deploy().then((_last_version) => {
      this._is_running = true;
    });
  }

  /**
   * @name get:
   */
  get (_id: string, _options: ?Object) {

    return this.schema().get(_id, _options);
  }

  /**
   * @name get_node:
   */
  get_node (_id: string, _options: ?Object) {

    return this.get(_id, _options);
  }

  /**
   * @name get_index:
   */
  get_index (_id: string, _options: ?Object) {

    return this.get(_id, _options);
  }

  /**
   * @name store:
   */
  store (_object: Object, _options: ?Object) {

    return this.schema().store(_object, _options);
  }

  /**
   * @name store_node:
   */
  store_node (_node: node_input, _options: ?Object) {

    let node;

    try {
      node = this._validate_node(_node);
    } catch (_e) {
      return promise.reject(_e);
    }

    return this.store(node);
  }

  /**
   * @name store_index:
   */
  store_index (_index: Object, _options: ?Object) {

    return this.store(_index, _options);
  }

  /**
   * @name remove:
   */
  remove (_object: Object, _options: ?Object) {

    return this.schema().remove(_object);
  }

  /**
   * @name remove_node:
   */
  remove_node (_node: Object, _options: ?Object) {

    return this.remove(_node, _options);
  }

  /**
   * @name remove_index:
   */
  remove_index (_index: Object, _options: ?Object) {

    return this.remove(_index, _options);
  }

  /**
   * @name compact:
   */
  compact (_options: ?Object) {

    return this.schema().compact();
  }

  /**
   * @name destroy:
   */
  destroy (_options: ?Object) {

    return this.schema().destroy();
  }

  /**
   * @name list:
   */
  list (_options: Object) {

    return promise.reject();
  }

  /**
   * @name create_node:
   *   High-level node creation API. Handles node creation,
   *   index creation, and bidirectional linking.
   */
  create_node (_node: node_input, _options: ?Object) {

    let rv = {};

    return this.store_node(_node, _options).then((_n) => {
      return this.index_node(_n, _options);
    }).then((_rv) => {
      rv.node = _rv.node;
      return this.link_index(_rv.index, _options);
    }).then((_index) => {
      rv.index = _index;
      return rv;
    });
  }

  /**
   * @name delete_node:
   *   High-level node deletion API. Deletes a node and
   *   all of its associated objects, indexes, and links.
   */
  delete_node (_node: node, _options: ?Object) {

    return this.get_index(_node.index_id).then((_i) => {
      return this.remove_index(_i);
    }).then(() => {
      return this.remove_node(_node);
    });
  }

  /**
   * @name index_node:
   */
  index_node (_node: node, _options: ?Object): node_and_index {

    let node;
    let schema = this.schema();

    try {
      node = this._validate_node(_node);
    } catch (_e) {
      return promise.reject(_e);
    }

    let rv = {
      data: {},
      type: 'index',
      zone_id: node.zone_id
    };

    let index_options = {
      tags: node.tags,
      sort: true, collator: schema
    };

    try {
      rv.data = this.indexer().build(node.content, index_options).result;
    } catch (_e) {
      return promise.reject(_e);
    }

    let path_fn = (_is_backward) => {
      return (_is_backward ? 'node_id' : 'index_id');
    };

    return schema.store_linked(node, rv, path_fn).then((_r) => {
      return { node: _r.doc, index: _r.linked_doc };
    });
  }

  /**
   * @name link_index:
   */
  link_index (_index: index, _options: ?Object): index {

    let schema = this.schema();
    let options = (_options || {});

    return this._link(_index, options).then((_rv) => {
      _index.links = _rv;
      return this.store_index(_index);
    });
  }

  /**
   * @name render_node:
   *   Render the node `_node` using renderer type `_type`, if
   *   compatible. Returns a promise yielding the rendered node.
   */
  render_node (_node: Object, _type: string, _options: ?Object) {

    let r, renderer_class = renderer[_type];

    if (!renderer_class) {
      return promise.reject(
        new exception.argument('Invalid or unsupported renderer type')
      );
    }

    try {
      r = new renderer_class();
    } catch (_e) {
      return promise.reject(
        new exception.runtime('Renderer instantiation failed', _e)
      );
    }

    return this.render(
      _node, _.bind(r.render_text, r),
        _.bind(r.render_link, r), _options
    );
  }

  /**
   * @name render:
   *   Render the body of `_node`, using the `_text_fn(str)`
   *   and `_link_fn(str, node_ids)` callbacks to format text
   *   and link blocks, respectively. Returns a promise yielding
   *   fully-linked text.
   */
  render (_node: Object, _text_fn: Function,
          _link_fn: Function, _options: ?Object) {

    let rv = [];
    let options = (_options || {});

    let accumulator = null;
    let content = _node.content;

    let links = new view_iterator(this.schema(), 'links', {
      view: {
        limit: options.limit,
        inclusive_end: false, group: true,
        startkey: [ _node.zone_id, _node.index_id ],
        endkey: [ _node.zone_id, _node.index_id + '\u0000' ]
      }
    });

    /* Link tuple format:
        0: zone_id, 1: index_id,
        2: offset, 3: length, 4: target_node_id */

    return links.each((_r) => {

      /* Setup */
      let link = _r.key;
      let link_offset = link[2];

      /* First run */
      if (!accumulator) {
        accumulator = this._render_accumulator_setup(content, 0, link);
        return;
      }

      /* Subsequent link at same offset */
      if (link_offset === accumulator.offset) {
        return this._render_accumulator_update(accumulator, link);
      }

      /* Overlapping links */
      if (link_offset < accumulator.i) {
        return; /* Ignoreland */
      }

      /* Subsequent link at new offset */
      let accumulator_current = accumulator;

      accumulator = this._render_accumulator_setup(
        content, accumulator_current.i, link
      );

      return this._render_accumulator_flush(
        accumulator_current, _text_fn, _link_fn
      ).then((_fragments) => rv.push(..._fragments));

    }).then(() => {

      let i = 0;
      let p = promise.resolve();

      /* Flush final link */
      if (accumulator) {
        i = accumulator.i;
        p = this._render_accumulator_flush(
          accumulator, _text_fn, _link_fn
        ).then((_fragments) => rv.push(..._fragments));
      }

      /* Flush trailing text */
      return p.then(() => {
        if (i < content.length) {
          return this._render_block(
            _text_fn, content.substring(i, Infinity)
          ).then((_fragments) => rv.push(..._fragments));
        }
      }).then(() => rv.join(''));
    });
  }

  /** @protected: **/

  /**
   * @name _render_accumulator_setup:
   */
  _render_accumulator_setup (_content: string,
                             _i: number, _link: Array<any>) {

    /* Link tuple format:
        0: zone_id, 1: index_id,
        2: offset, 3: length, 4: target_node_id */

    let node_id = _link[4];
    let offset = _link[2], length = _link[3];

    let target_node_ids = {};
    target_node_ids[node_id] = true;

    let text_str = _content.substring(_i, offset);
    let link_str = _content.substring(offset, offset + length);

    return {
      i: offset + length, offset: offset,
      text_str: text_str, link_str: link_str,
      link: _link, target_node_ids: target_node_ids
    };
  }

  /**
   * @name _render_accumulator_update:
   */
  _render_accumulator_update (_accumulator: Object, _link: Array<any>) {

    /* Link tuple format:
        0: zone_id, 1: index_id,
        2: offset, 3: length, 4: target_node_id */

    _accumulator.target_node_ids[_link[4]] = true;
    return this;
  }

  /**
   * @name _render_accumulator_flush:
   */
  _render_accumulator_flush (_accumulator: Object,
                             _text_fn: Function, _link_fn: Function) {

    return this._render_block(
      _text_fn, _accumulator.text_str, _link_fn,
        _accumulator.link_str, _.keys(_accumulator.target_node_ids)
    );
  }

  /**
   * @name _render_block:
   */
  _render_block (_text_fn: Function,
                 _text_str: string, _link_fn: ?Function,
                 _link_str: ?string, _target_node_ids: ?Array<string>) {

    let p = [
      utils.make_promise(_text_fn, _text_str)
    ];

    if (_link_fn) {
      p.push(utils.make_promise(_link_fn, _link_str, _target_node_ids));
    }

    return promise.all(p);
  }

  /**
   * @name _link:
   */
  _link (_index: Object, _options: Object) {

    let rv = [];

    this._index_cache_reset();

    return this._link_outgoing(_index, _options).then((_outgoing) => {
      rv = _outgoing;
      return this._link_incoming(_index, _options);
    }).then((_incoming) => {
      return [ ...rv, ..._incoming ];
    });
  }

  /**
   * @name _link_incoming:
   */
  _link_incoming (_index: Object, _options: Object) {

    let iterator_options = this._compute_iterator_options(
      true, _options
    );

    let t = new transform(words_transform, {
      sort: true
    });

    let words = new array_iterator(
      t.map([ _index ]), iterator_options.array
    );

    words.events.on('key.map', (_k) => {
      return _k.key;
    });

    let tags = new view_iterator(
      this.schema(), 'nodes_by_tag', iterator_options.view
    );

    return this._link_generic(words, tags, iterator_options);
  }

  /**
   * @name _link_outgoing:
   */
  _link_outgoing (_index: Object, _options: Object) {

    let iterator_options = this._compute_iterator_options(
      false, _options
    );

    let t = new transform(nodes_by_tag_transform, {
      sort: true
    });

    let words = new view_iterator(
      this.schema(), 'words', iterator_options.view
    );

    let tags = new array_iterator(
      t.map([ _index ]), iterator_options.array
    );

    tags.events.on('key.map', (_k) => {
      return _k.key;
    });

    return this._link_generic(words, tags, iterator_options);
  }

  /**
   * @name _link_generic:
   */
  _link_generic (_words: iterator, _tags: iterator, _options: Object) {

    let rv = [];

    let j = new join_iterator(
      _words, _tags, _options.join
    );

    j.events.on('key.subset', (_r) => {
      return _r.slice(0, 2);
    });

    return j.each((_r) => {
      return this._link_one(rv, _r, _options).then((_link) => {

        if (_link !== false) {
          rv.push(_link);
        }
      });
    }).then(() => rv);
  }

  /**
   * @name _link_one:
   */
  _link_one (_rv: Array<Object>,
             _join_match: Object, _options: Object) {

    return this._link_one_slow(_rv, _join_match, _options);
  }

  /**
   * @name _link_one_slow:
   */
  _link_one_slow (_rv: Array<Object>,
                  _join_match: Object, _options: Object) {

    let schema = this.schema();

    let left = _join_match.left;
    let right = _join_match.right;

    let p = this._index_cache_retrieve(left.id);

    return p.then((_index_doc) => {

      this._index_cache_store(_index_doc);

      let r = this._match_keyphrase(
        _index_doc, left.key, right.value.tag, _options
      );

      if (!r) {
        return false;
      }

    /* Link tuple format:
        0: zone_id, 1: index_id, 2: offset,
        3: length, 4: word_index, 5: target_node_id  */

      return [ 0, left.id, r[0], r[1], right.value._id, r[2] ];
    });
  }

  /**
   * @name _match_keyphrase:
   */
  _match_keyphrase (_index_doc: Object, _word_key: Array<any>,
                    _indexed_tag: Array<any>, _options: Object) {

    /* Word tuple format:
        0: zone_id, 1: index_id,
        2: word, 3: offset, 4: length, 5: ordinal */

    let ordinal = _word_key[5];

    /* Tag tuple format:
        0: word, 1: offset, 2: length, 3: ordinal */

    let tag_words = _indexed_tag[1];
    let words = _index_doc.data.words;

    if (words.length <= 0 || tag_words.length <= 0) {
      return false;
    }

    let i = 0;
    let len = tag_words.length;
    let last_word = words[ordinal + len - 1];

    /* Ensure all words match */
    for (;;) {

      let word = words[ordinal + i];

      if (_.isNil(word) || word[0] !== tag_words[i][0]) {
        return false;
      }

      if (++i >= len) {
        break;
      }
    }

    /* Result tuple format:
         0: offset, 1: length, 2: word_index */

    return [
      words[ordinal][1], /* Word offset */
      last_word[1] + last_word[0].length - words[ordinal][1], /* Length */
      words.slice(ordinal, ordinal + len) /* Matched portion of word index */
    ];
  }

  /** @private: **/

  /**
   * @name _index_cache_reset:
   */
  _index_cache_reset() {

    this._index_cache = {};
    return this;
  }

  /**
   * @name _index_cache_retrieve:
   */
  _index_cache_retrieve (_id: String) {

    let schema = this.schema();
    let index = this._index_cache[_id];

    return (
      _.isUndefined(index) ?
        schema.get(_id) : promise.resolve(index)
    );
  }

  /**
   * @name _index_cache_store:
   */
  _index_cache_store (_index_doc: Object) {

    this._index_cache[_index_doc._id] = _index_doc;
    return this;
  }

  /**
   * @name _validate_node:
   */
  _validate_node (_node: Object, _options: ?Object) {

    let v = this._validator;

    if (!_.isObject(_node)) {
      throw new exception.argument('Node must be an object')
    }

    _node.type = 'node';

    let r = v.validate({
      title: _.isString,
      content: _.isString,
      tags: [ v.optional, _.isArray ],
      zone_id: [ v.optional, utils.logical.or(_.isString, _.isInteger) ]
    }, _node, false);

    if (!r.ok) {
      throw new exception.argument(r.message, false, r)
    }

    if (_.isNil(_node.tags)) {
      _node.tags = [];
    }

    if (_.isNil(_node.zone_id)) {
      _node.zone_id = 0;
    }

    return this._timestamp_document(_node);
  }

  /**
   * @name _timestamp_document:
   */
  _timestamp_document (_doc: Object, _options: ?Object) {

    let date = (new Date()).toISOString();

    if (_.isNil(_doc.created_at)) {
      _doc.created_at = date;
    }

    _doc.updated_at = date;
    return _doc;
  }

  /**
   * @name _compute_iterator_options:
   */
  _compute_iterator_options (_is_incoming: boolean,
                             _iterator_options: Object) {

    let iterator_options = (_iterator_options || {});
    let join_options = (iterator_options.join || {});
    let view_options = (iterator_options.view || {});
    let array_options = (iterator_options.array || {});

    let default_join_options = {};
    let default_array_options = {};

    let default_view_options = {
      limit: this._default_result_limit
    };

    join_options = _.assign(default_join_options, join_options, {
      ordered: true
    });

    array_options = _.assign(default_array_options, array_options, {
      ordered: true
    });

    view_options = _.assign(default_view_options, view_options, {
      ordered: true
    });

    return {
      join: join_options,
      view: view_options,
      array: array_options
    };
  }

};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Engine;
export { Engine as engine };

