/* globals emit */

'use strict';

import Base from './base';
import utils from './utils';
import _ from './utils/lodash';
import exception from './exception';

/**
 * @name Events:
 *   Register, emit, listen for, and manage synchronous events.
 */
var Events = class extends Base {

  /**
   * @name constructor:
   */
  constructor (_options) {

    /* Must be first */
    super(_options);

    this._events = {};
    return this;
  }

  /**
   * @name register:
   *   Register a new type of event named `_type`, a string. Start
   *   allowing others to attach to it via the `on` method and/or
   *   trigger it using the `emit` method. If `_is_singleton` is
   *   true-like, then only allow one listener at a time to be attached
   *   via the `on` method. Otherwise, allow unlimited listeners.
   */
  register (_type, _is_singleton) {

    if (!_.isNil(this._events[_type])) {
      throw new exception.reference('Event type is already registered');
    }

    this._events[_type] = {
      listeners: {},
      is_singleton: !!_is_singleton
    };

    return this;
  }

  /**
   * @name on:
   */
  on (_type, _fn) {

    let ev = this.get(_type);

    if (!_.isFunction(_fn)) {
      throw new exception.argument('Handler must be a function');
    }

    if (ev.is_singleton && _.size(ev.listeners) >= 1) {
      throw new exception.reference('Event already has a listener');
    }

    let rv = {
      fn: _fn,
      id: utils.uuid.memory(),
      remove: () => {
        return this.remove(_type, rv);
      }
    };

    ev.listeners[rv.id] = rv;
    return rv;
  }

  /**
   * @name send:
   *   A simple flat-argument wrapper for `emit`.
   */
  send (_type, _default_fn, ..._args) {

    return this.emit(_type, _default_fn, _args);
  }


  /**
   * @name emit:
   *   Synchronously emit an event of type `_type` with arguments `_args`.
   *   If `_default_fn` is specified and there are no listeners for the
   *   event type `_type`, call it with the arguments in `_args` and
   *   return its result instead of `undefined`.
   */
  emit (_type, _default_fn, _args) {

    let ev = this.get(_type);
    let args = (_args || []);

    if (_default_fn === false) {
      _default_fn = null;
    }

    if (_.size(ev.listeners) <= 0 && !_.isNil(_default_fn)) {
      return _default_fn.apply(this, args);
    }

    let rv = _.map(ev.listeners, (_l) => {
      return _l.fn.apply(this, args);
    });

    if (ev.is_singleton) {
      rv = rv[0];
    }

    return rv;
  }

  /**
   * @name get:
   */
  get (_type) {

    let ev = this._events[_type];

    if (!_.isPlainObject(ev)) {
      throw new exception.reference('Event type is not registered');
    }

    return ev;
  }

  /**
   * @name listeners:
   *   Return the number of active listeners for event type
   *   `_type`. If the event was specified as a singleton vent
   *   type, the return value will always be either zero or one.
   */
  listeners (_type) {

    return _.size(this.get(_type).listeners);
  }

  /**
   * @name remove:
   */
  remove (_type, _listener) {

    let ev = this.get(_type);

    if (!_.isPlainObject(_listener) || !_.isString(_listener.id)) {
      throw new exception.argument('Invalid listener object');
    }

    delete ev.listeners[_listener.id];
    return true;
  }

  /**
   * @name unregister:
   */
  unregister (_type) {

    if (!_.isPlainObject(this._events[_type])) {
      throw new exception.reference('Event type is not registered');
    }

    delete this._events[_type];
    return this;
  }

};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Events;
export { Events as events };

