
'use strict';

import Base from './base';
import _ from './utils/lodash';
import splitter from './splitter';
import exception from './exception';

/**
 * @name Indexer:
 *   Build an indexable document from raw text.
 */
var Indexer = class extends Base {

  /**
   * @name constructor:
   */
  constructor (_str, _options) {

    /* Must be first */
    super(_options);

    if (!_.isNil(_str)) {
      this.string = _str;
    }

    this._text_splitter = (
      new splitter.text(null, { fold: true })
    );

    return this;
  }

  /**
   * @name set string:
   */
  set string (_str) {

    if (!_.isString(_str)) {
      throw new exception.argument('Argument must be a string');
    }

    this._string = _str;
    return this;
  }

  /**
   * @name get string:
   */
  get string () {

    return this._string;
  }

  /**
   * @name get result:
   */
  get result () {

    return this._result;
  }

  /**
   * @name get splitter:
   *   Return the splitter instance currently being used by
   *   this indexer to process text.
   */
  get splitter () {

    return this._text_splitter;
  }

  /**
   * @name build:
   *   Build an index. This base implementation only processes
   *   tags, since the exact language/format of `_str` might vary.
   */
  build (_str, _options) {

    let options = (_options || {});

    /* Shortcut */
    if (!_.isNil(_str)) {
      this.string = _str;
    }

    /* Empty result */
    this._result = {
      content: null, tags: [], words: []
    };

    /* Index all tags */
    if (options.tags) {

      if (!_.isArrayLikeObject(options.tags)) {
        throw new exception.argument('Tags must be an array');
      }

      try {
        this._build_tag_index(options.tags, options);
      } catch (_e) {
        throw new exception.runtime('Failed to build tag index', _e);
      }
    }

    return this;
  }

  /**
   * @name clear:
   */
  clear () {

    this._string = this._result = undefined;
    return this;
  }

  /**
   * @name _build_tag_index:
   *   Helper function. Builds a word index for each tag in
   *   `_tags`, and adds them to the `result` property of `this`.
   */
  _build_tag_index (_tags, _options) {

    /* Allow option overrides */
    let options = _.assignIn({}, this.options, (_options || {}));

    let tags = _tags;
    let collator = options.collator;

    if (options.sort) {
      tags.sort(
        !collator ? undefined :
          (_l, _r) => collator.collate(_l, _r)
      );
    }

    return _.each(tags, (_tag) => {
      this._result.tags.push([
        _tag, this._text_splitter.split(_tag).result
      ]);
    });
  }
};

/**
 * @name Splitter.Text:
 *   A word splitter for plain Unicode text.
 */
var indexer_text = Indexer.Text = class extends Indexer {

  /**
   * @name build:
   */
  build (_str, _options) {

    /* Must be first */
    super.build(_str, _options);

    this._result.content = this.string;
    this._result.words = this._text_splitter.split(this.string).result;

    return this;
  }
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default {
  base: Indexer,
  text: Indexer.Text
};

export {
  Indexer as base,
  indexer_text as text
};

