
'use strict';

import Base from '../base';
import utils from '../utils';
import events from '../events';
import _ from '../utils/lodash';
import promise from '../promise';
import exception from '../exception';
import { collate } from 'pouchdb-collate';

/**
 * @name Iterator:
 *   An asynchronous, buffered, seekable iterator.
 */
var Iterator = class extends Base {

  /**
   * @name constructor:
   */
  constructor (_options) {

    /* Must be first */
    super(_options);

    let e = this._events = new events();

    this._autobegin_disabled = (
      (this.options.autobegin === false)
    );

    e.register('filter', true);
    e.register('populate', true);
    e.register('key.map', true);
    e.register('transform', true);

    return this.reset();
  }

  /**
   * @name get events:
   *   Return the event manager currently in use.
   *
   *  @event populate:
   *    Add a populate transformation function to this iterator. A
   *    populate transformation function makes changes to entire batches
   *    of results, all at once, as they are paged in from an upstream
   *    data source. The function will be called as `populate(_results,
   *    _data)`, where `_data` is an initially-empty object that can be
   *    used to make the transform stateful (i.e. store data between
   *    individual calls), and `_results` is a modifiable array of
   *    results that may be transformed. The new value for `_results`
   *    should be returned.
   *
   *   @event filter:
   *    Add a result filter to this iterator. A result filter is a
   *    function that "prescreens" individual results before they are
   *    returned to the caller. It has no direct effect on
   *    non-sequential operations like `seek`, but will screen results
   *    out following a non-sequential operation. The result filter
   *    will be called as `filter(_result, _data)`, where `_data` is
   *    an initially-empty object that can be used to make a filter
   *    stateful (i.e. store data between calls), and `_result` is
   *    the result object to examine. Return true if the result object
   *    passes the filter and should be included in the iterator
   *    result set, or false otherwise.
   *
   *   @event transform:
   *    Add a transformation function to this iterator. A
   *    transformation function can change individual results as
   *    they're encountered, before they're returned to the caller.
   *    The function will be called as `transform(_result, _data)`,
   *    where `_data` is an initially-empty object that can be used to
   *    make the transform stateful, i.e. store data between individual
   *    calls. The new value for `_result` should be returned.
   *
   *   @event key.map:
   *    Add a key mapping function to this iterator. This function allows
   *    you to specify a particular way of locating a key inside of an
   *    iterator result object. For example, an array iterator with items
   *    of the form `{ key: 1 }` would map keys using `(_r) => _r.key`.
   *    If provided, this function is used in the `key_for` method.
   */
  get events () {

    return this._events;
  }

  /**
   * @name get statistics:
   *   Return an object containing performance information for
   *   the current iterator instance. The `requests` property will
   *   contain the number of times `populate` has executed successfully,
   *   (usually following exhaustion of internally-buffered results).
   */
  get statistics () {

    return {
      requests: this._total_requests,
      filter: { data: this._filter_data },
      transform: { data: this._transform_data }
    };
  }

  /**
   * @name get key:
   *   Return the most recent value passed to `seek`, or undefined if
   *   `seek` hasn't yet been called. Note that this is the most
   *   recently *requested* key, and may not match the actual current
   *   result key. If you need the latter, use `current_key`.
   */
  get key () {

    return this._key;
  }

  /**
   * @name current_key:
   *   Return the current item from in-memory buffer.
   */
  current_key () {

    let current = this.current(true);

    if (_.isUndefined(current)) {
      return undefined;
    }

    return this.key_for(current);
  }

  /**
   * @name get last_key:
   *   Return the last item from in-memory buffer.
   */
  last_key () {

    if (this._buffer.length <= 0) {
      return undefined;
    }

    return this.key_for(_.last(this._buffer));
  }

  /**
   * @name begin:
   *   Fill the internal buffer, and prepare for a barrage of calls
   *   to the `current` and `next` methods. This function is asynchronous
   *   and returns a promise.
   */
  begin () {

    return this.reset().seek();
  }

  /**
   * @name current:
   *   Return the item at the current position. If `_is_original`
   *   is true-like, return the original value, taken after `populate`
   *   was handled but before `transform` was handled.
   */
  current (_is_original) {

    if (this.is_bof() || this.is_eof()) {
      return undefined;
    }

    return (
      _is_original ?
        this._buffer[this._offset] :
          this._apply_transform(this._offset)
    );
  }

  /**
   * @name next:
   *   Advance the internal pointer to the next item, and return a
   *   promise that yields that next item. Though the interface is
   *   asynchronous, most requests (read: those that don't exhaust
   *   the internal buffer and end up calling down to `populate`)
   *   will be synchronously and immediately resolved/rejected.
   */
  next () {

    let is_bof = this.is_bof();

    return this._require_started().then(() => {

      /* Potentially-fast path */
      if (this._offset + 1 < this._buffer.length) {

        this._offset++;
        return this._filter_iterate_nonresident();
      }

      /* Guaranteed-slow path */
      return this.clear().fill(false, is_bof);
    });
  }

  /**
   * @name previous:
   *   Move the internal pointer backward one item, and return a
   *   promise that yields that item. Though the interface is
   *   asynchronous, most requests (read: those that don't exhaust
   *   the internal buffer and end up calling down to `populate`)
   *   will be synchronously and immediately resolved/rejected.
   */
  previous () {

    let is_eof = this.is_eof();

    return this._require_started().then(() => {

      /* Potentially-fast path */
      if (this._offset > 0) {
        this._offset--;
        return this._filter_iterate_nonresident(true);
      }

      /* Guaranteed-slow path */
      return this.clear().fill(true, is_eof);
    });
  }

  /**
    * @name fill:
    *   Asynchronously fill the internal buffer with whatever
    *   results come next. Returns the new current value.
    */
  fill (_in_reverse, _is_initial) {

    return this._require_started().then(() => {

      if (_in_reverse) {
        if (this._previous_keys.length > 0) {
          this._is_eof = false;
          this._key = this._previous_keys.pop();
        } else {
          this._is_bof = true;
          return this.current();
        }
      } else {
        this._previous_keys.push(this._key);
      }

      let p = this._seek_slow(
        this._key, _in_reverse, !this.is_eof(), false, _is_initial
      );

      return p.then(() => {
        return this._filter_iterate_nonresident(_in_reverse);
      });
    });
  }

  /**
   * @name drain:
   *   Synchronously return an array of any remaining buffered
   *   results, emptying the internal buffer beforehand.
   */
  drain () {

    this._require_started_synchronous();

    let buffer = this._buffer;
    let offset = this._offset;

    if (this.is_depleted()) {
      return false;
    }

    if (this.events.listeners('transform')) {

      /* Transform anything that still needs it */
      for (let i = this._offset; i < buffer.length; ++i) {
        this._apply_transform(i);
      }

      /* Use transformed results */
      buffer = this._transform_map;
    }

    let rv = buffer.slice(offset);
    this.clear();

    return rv;
  }

  /**
   * @name is_depleted:
   *   Returns true if the buffered results have been depleted,
   *   either by calling `drain` or by reaching the end with some
   *   potentially-trivial combination of `seek` and `next`.
   */
  is_depleted () {

    return (this._buffer.length < this._offset + 1);
  }

  /**
   * @name is_first_buffered:
   *   Returns true if the iterator is currently pointing at the first
   *   item in its internal buffer.
   */
  is_first_buffered () {

    return (this._offset <= 0);
  }

  /**
   * @name is_last_buffered:
   *   Returns true if the current iterator position is at or beyond
   *   the last buffered item. This function guarantees that a subsequent
   *   call to `next` would result in a buffer miss and a trip back
   *   through `populate`.
   */
  is_last_buffered () {

    return (this._buffer.length <= this._offset + 1);
  }

  /**
   * @name is_final_buffered:
   */
  is_final_buffered (_in_reverse) {

    return (
      _in_reverse ?
        this.is_first_buffered() : this.is_last_buffered()
    );
  }

  /**
   * @name is_eof:
   *   Return true if end-of-file was reached, i.e. an attempt was made
   *   to seek beyond the end of the internal buffer, and no more items
   *   can be found to fill the buffer. In any other case, return false.
   */
  is_eof () {

    return this._is_eof;
  }

  /**
   * @name is_bof:
   *   Return true if the stream is at its beginning, i.e. no operations
   *   have been performed yet, or an attempt was made to seek backward
   *   beyond the beginning of the internal buffer and no more items
   *   are available. In any other case, return false;
   */
  is_bof () {

    return this._is_bof;
  }

  /**
   * @name is_ordered:
   *   Return true if items pointed to by this iterator are sorted
   *   according to the `compare` method, or false if they appear
   *   in some other ordering.
   */
  is_ordered () {

    return !!this.options.ordered;
  }

  /**
   * @name seek:
   *   Seek to the record identified by `_key`. The format of `_key`
   *   is iterator-defined, and can be of any supported, defined type.
   *   If `undefined` is provided for `_key`, seek to the beginning.
   *   If `_is_inexact` is false-like, seeking will continue until
   *   an exact match for `_key` is found. If `_is_inexact` is true-like,
   *   seeking will stop at the first item that measures greater-than
   *   or equal-to `_key`. If `_force_advance` is true-like, never match
   *   the current iterator position in inexact mode - always advance
   *   by at least one item.
   */
  seek (_key, _is_inexact, _force_advance) {

    this._previous_keys = [ _key ];

    return this.seek_generic(
      _key, false, _is_inexact, _force_advance
    );
  }

  /**
   * @name rewind:
   *   Seek backward to the record identified by `_key`. All other
   *   parameters behave identically to those described in `seek`.
   */
  rewind (_key, _is_inexact, _force_advance) {

    this._is_eof = false;
    this._previous_keys = [];

    return this.seek_generic(
      _key, true, _is_inexact, _force_advance
    );
  }

  /**
   * @name reset:
   *   Reset the iterator's internal state and discard buffers.
   */
  reset () {

    this._key = undefined;
    this._previous_keys = [];

    this._is_seeking = false;
    this._is_started = false;

    this._total_requests = 0;

    this._filter_data = {};
    this._transform_data = {};

    return this.clear();
  }

  /**
   * @name clear:
   *   Discard any internally-buffered results. Any subsequent
   *   call to `next` or `seek` will result in a call to `populate`.
   */
  clear () {

    /* Save final key */
    this._key = (
      (this._buffer && this._buffer.length > 0) ?
        this.last_key() : this._key
    );

    this._offset = 0;
    this._buffer = [];
    this._transform_map = [];
    this._is_eof = this._is_bof = false;

    return this;
  }

  /**
   * @name each:
   *   Call `_fn` for each item. Returns a promise.
   */
  each (_fn) {

    return this._require_started().then(() => {
      return promise.while(() => !this.is_eof(), () => {

        let p = utils.make_promise(
          _fn.call(this, this.current(), this)
        );

        return p.then((_rv) => {
          return this.next();
        });
      });
    });
  }

  /**
   * @name all:
   *   Return an array containing all remaining items.
   */
  all () {

    return this._require_started().then(() => {

      let rv = [];

      return this.each((_item) => {
        rv.push(_item);
      }).then(() => {
        return rv;
      });
    });
  }

  /**
   * @name get offset:
   *   Return the current position within the in-memory buffer.
   *   If a `next` or `seek` operation causes a buffer miss,
   *   this value is reset.
   */
  get offset () {

    return this._offset;
  }

  /**
   * @name get buffer:
   *   Return a reference to the currently-in-use buffer. This
   *   method may be called repeatedly and should be fast. If
   *   substantial computation is required to produce a buffer,
   *   it should be performed elsewhere.
   */
  get buffer () {

    return this._buffer;
  }

  /** @protected: **/

  /**
   * @name set offset:
   */
  set offset (_offset) {

    return (this._offset = _offset);
  }

  /**
   * @name set buffer:
   *   Replace this iterator's internal buffer with a new array.
   */
  set buffer (_buffer) {

    return (this._buffer = _buffer);
  }

  /**
   * @name repopulate:
   *   Pop an internal stack frame and call `populate`. This helper
   *   function can be used from within subclasses to reinvoke
   *   `populate` when current results are insufficient.
   */
  repopulate (_in_reverse, _is_sequential, _is_inexact, _is_initial) {

    this._key = this._previous_keys.pop();

    return this.populate(
      _in_reverse, _is_sequential, _is_inexact, _is_initial
    );
  }

  /**
   * @name populate:
   *   Assign a new array to `this.buffer`. This array should contain
   *   a new set of items from the iterator's backing data source,
   *   starting at `this.key` and ending at an implementation-defined
   *   limit.
   *
   *   If `_in_reverse` is true, the iterator is seeking and/or iterating
   *   backward, toward the beginning of the collection.
   *
   *   If `_is_sequential` is true, then the `populate` operation
   *   was prompted by a call to `next`; in this case, the iterator's
   *   new current item should be *after* `this.key` rather than directly
   *   on `this.key`. This can be achieved by manipulating `this.offset`
   *   or by simply not returning the matching item - either is fine.
   *   If `this.key` is defined, then the `populate` operation should start
   *   from the beginning of the backing data.
   *
   *   If `_is_inexact` is true-like, then the `populate` operation
   *   should start from the first item that is greater-than or equal-to
   *   `this.key` according to `this.compare`. Otherwise, the `populate`
   *   operation should start from an exact match.
   *
   *   If `_is_initial` is true-like, then this request represents a
   *   sequential access of first known key of the underlying collection
   *   (or, if `_in_reverse` is true, the last known key). This information
   *   can be useful for handling pagination edge cases. Use this in
   *   subclasses instead of `is_bof` and `is_eof`.
   *
   *   Before this function is invoked, `this.offset` must be reset
   *   to zero by the caller.
   *
   *   If `_is_sequential` is true-like, then `_is_inexact` should
   *   always be false-like: an inexact search makes no sense when
   *   merely advancing to the next item.
   *
   *   This function is asynchronous and returns a promise. If no
   *   additional results are available, the promise should evaluate
   *   to `false`. If the buffer was filled, the promise should
   *   evaluate to `true`. The promise should be rejected in any other
   *   error condition.
   */
  populate (_in_reverse, _is_sequential, _is_inexact, _is_initial) {

    /* Full-buffer transform support */
    this._buffer = this._events.send(
      'populate', _.identity, this._buffer, this._fill_data
    );

    return promise.resolve(true);
  }

  /**
   * @name find:
   *   Starting at `this.offset`, find the first item in `this.buffer`
   *   that matches `_key` and return its offset. If the item can't be
   *   found, return `false`. This function is synchronous, as it only
   *   operates on the internally-held `this.buffer`. If `_is_inexact`
   *   is false-like, find an exact match for `_key`. Otherwise, find the
   *   first item that measures greater-than or equal-to `_key` according
   *   to `compare`, and return it. Returns an integer offset on success,
   *   or false if the key could not be found. If `_in_reverse` is
   *   true-like, search backward (i.e. toward offset zero) rather than
   *   forward (i.e. toward offset `this.buffer.length`).
   */
  find (_key, _in_reverse, _is_inexact) {

    if (this.is_ordered()) {
      return this.find_logarithmic(_key, _in_reverse, _is_inexact);
    }

    return this.find_linear(_key, _in_reverse, _is_inexact);
  }

  /**
   * @name key_for:
   *   Return the `seek` key that can be used to locate `_item`.
   */
  key_for (_item) {

    return this.events.emit(
      'key.map', _.identity, [ _item ]
    );
  }

  /**
   * @name compare:
   *   A replacable key-comparison function. Returns zero if `_l` and
   *   `_r` are lexically indistinguishable; -1 if `_l` should come before
   *   `_r` in lexical order, and 1 if `_r` should come before `_l`. The
   *   default implementation should be overridden if the upstream data
   *   source defines its own key collation scheme.
   */
  compare (_l, _r) {

    return collate(_l, _r);
  }

  /**
   * @name find_linear:
   *   A linear-time find function. Suitable for all input.
   */
  find_linear (_key, _in_reverse, _is_inexact) {

    if (_in_reverse) {
      return this._find_linear_reverse(_key, _is_inexact);
    }

    return this._find_linear_forward(_key, _is_inexact);
  }

  /**
   * @name find_logarithmic:
   *   A logarithmic-time find function for pre-sorted input.
   *   Uses a binary search.
   */
  find_logarithmic (_key, _in_reverse, _is_inexact) {

    let buffer = this._buffer;
    let offset = this._offset;

    let map = (_r) => {
      return this.key_for(_r);
    };

    let compare = (_l, _r) => {
      return this.compare(_l, _r);
    };

    let n = utils.search.binary(
      buffer, _key, compare, map,
        (_in_reverse ? 0 : offset), (_in_reverse ? offset : undefined)
    );

    /* Empty input */
    /* istanbul ignore next */
    if (n === false) {
      return false;
    }

    /* Exact match */
    if (n >= 0) {
      this._offset = n;
      return this._find_linear_reverse(_key, _is_inexact);
    }

    /* Inexact match */
    if (_is_inexact && n < 0) {

      n = ~n;

      if (n < buffer.length) {
        return n;
      }
    }

    /* No match */
    return false;
  }

  /**
   * @name seek_generic:
   *   Backend for the `seek` and `rewind` methods. Control the
   *   search direction using the boolean `in_reverse` parameter.
   */
  seek_generic (_key, _in_reverse, _is_inexact, _force_advance) {

    let p = this._seek(
      _key, _in_reverse, _is_inexact, _force_advance
    );

    return p.then(() => {
      return this._filter_iterate_nonresident(_in_reverse);
    });
  }

  /** @private **/

  /**
   * @name _require_started:
   *   Call `begin` if we're allowed to and it hasn't already
   *   been called. Asynchronous; returns a promise.
   */
  _require_started () {

    if (this._is_started) {
      return promise.resolve();
    }

    if (this._autobegin_disabled) {
      throw new exception.argument(
        'You must call `begin` before using this iterator'
      );
    }

    return this.begin();
  }

  /**
   * @name _require_started_synchronous:
   *   Throw an exception if `begin` hasn't yet been called.
   *   Otherwise, return `this`.
   */
  _require_started_synchronous () {

    if (!this._is_started) {
      throw new exception.argument(
        'You must call `begin` before using this iterator'
      );
    }

    return this;
  }

  /**
   * @name _find_linear_forward:
   */
  _find_linear_forward (_key, _is_inexact) {

    let buffer = this._buffer;
    let offset = this._offset;

    for (; offset < buffer.length; offset++) {

      let n = this.compare(this.key_for(buffer[offset]), _key);

      if (n === 0 || (_is_inexact && n > 0)) {
        return offset;
      }
    }

    /* Not found */
    return false;
  }

  /**
   * @name _find_linear_reverse:
   */
  _find_linear_reverse (_key, _is_inexact) {

    let rv = false;
    let offset = this._offset;

    for (;;) {

      let n = this.compare(
        this.key_for(this._buffer[offset]), _key
      );

      if (_is_inexact && n < 0) {
        return offset + (offset + 1 < this._buffer.length ? 1 : 0);
      }

      if (n === 0) {
        rv = offset;
      } else if (rv !== false) {
        break;
      }

      if (offset <= 0) {
        break;
      }

      offset--;
    }

    return rv;
  }

  /**
   * @name _seek:
   *   The internal implementation of `seek`. This version lacks any
   *   and all support for result filtering. See the `seek` method
   *   for additional information and/or documentation.
   */
  _seek (_key, _in_reverse, _is_inexact, _force_advance) {

    let offset = false;

    /* Handle undefined key:
     *  This seeks to beginning of the stream. */

    if (_.isUndefined(_key)) {
      this._is_started = true;
      return this._seek_slow(_key, _in_reverse, false, _is_inexact);
    }

    /* Precondition:
     *  Either `begin` or `seek` must have been called. */

    this._require_started_synchronous();

    /* Fast path */
    if (!this.is_final_buffered(_in_reverse)) {

      /* You gotta move */
      this._offset += (_in_reverse ? -1 : 1);

      /* Search in-memory buffer first */
      offset = this.find(_key, _in_reverse, _is_inexact);

      /* Buffer hit: found */
      if (_.isInteger(offset) && offset < this._buffer.length) {

        this._offset = offset;
        return promise.resolve(this.current());
      }
    }

    /* Buffer miss: forced sequential seek */
    if (_force_advance && utils.is_subset(this.last_key(), _key)) {
      return this._seek_slow(
        this.last_key(), _in_reverse, true, _is_inexact
      );
    }

    /* Buffer miss: seek */
    if (offset === false) {
      return this._seek_slow(_key, _in_reverse, false, _is_inexact);
    }

    /* Invalid offset */
    return promise.reject(
      new exception.unexpected(
        'Iterator `find` method returned an invalid result'
      )
    );
  }

  /**
   * @name _filter_iterate_resident:
   *   Run the currently-active result filter against the in-memory
   *   buffer only.  If it matches the current item, do nothing. If
   *   it fails to match, advance to the next in-memory result and try
   *   again. Return false if the end of the buffer is reached without
   *   locating a match; true if a match is located. This function is
   *   synchronous, and only deals with results that are resident in
   *   memory.
   */
  _filter_iterate_resident (_in_reverse) {

    let buffer = this._buffer;
    let length = buffer.length;

    for (;;) {

      let condition = (
        _in_reverse ?
          this._offset >= 0 : this._offset < length
      );

      if (!condition) {
        break;
      }

      let r = buffer[this._offset];

      if (this._events.send('filter', false, r, this._filter_data)) {
        return true;
      }

      if (_in_reverse) {
        this._offset--;
      } else {
        this._offset++;
      }
    }

    return false;
  }

  /**
   * @name _filter_iterate_nonresident:
   *   Run `_filter_iterate_resident`, but don't stop when the end of
   *   the in-memory buffer is reached: ask the backing data source
   *   for another result set and continue onward until a result is
   *   found or end-of-file is reached. This function is asynchronous
   *   and returns a promise.
   */
  _filter_iterate_nonresident (_in_reverse) {

    let found = false;

    /* Fast path */
    if (!this._events.listeners('filter')) {
      return promise.resolve(this.current());
    }

    /* Slow path */
    return promise.while(() => {

      return (
        !found &&
          !(_in_reverse ? this.is_bof() : this.is_eof())
      );

    }, () => {

      if (!this._filter_iterate_resident(_in_reverse)) {
        return this.clear().fill(_in_reverse);
      }

      found = true;
      return promise.resolve(this.current());
    });
  }

  /**
   * @name _seek_slow:
   *   Slow path for the `seek` method. This handles the case where the
   *   internal buffer has been exhausted, and `populate` must be called.
   *   If `_is_sequential` is true, and the `seek` operation was prompted
   *   by a call to `next`, the internal buffer will be filled starting
   *   immediately *after* `this.key`. If `_is_inexact` is false-like,
   *   then this function was called from an inexact-matching `seek`
   *   invocation: pass this on to `populate`.
   */
  _seek_slow (_key, _in_reverse, _is_sequential, _is_inexact, _is_initial) {

    if (this._is_seeking) {
      return promise.reject(
        new exception.runtime('A seek operation is already in progress')
      );
    }

    this._offset = 0;
    this._key = _key;
    this._is_seeking = true;

    let p = this.populate(
      _in_reverse, _is_sequential, _is_inexact, _is_initial
    );

    return p.then((_is_filled) => {

      this._total_requests++;
      this._is_seeking = false;

      if (!_is_filled) {
        this._is_eof = true;
        return undefined;
      }

      /* Success */
      return this.current();

    }).catch((_e) => {

      /* Clean up and rethrow */
      this._is_seeking = false;
      throw _e;
    });
  }

  /**
   * @name _apply_transform:
   *   Apply the currently-registered transformation function to
   *   the result at offset `_offset`. The value in `_offset` must
   *   be a prescreened value that is less than `this.buffer.length`.
   *   Return the transformed result.
   */
  _apply_transform (_offset) {

    /* Base case */
    if (!this._events.listeners('transform')) {
      return this._buffer[_offset];
    }

    /* Transformation case */
    if (_.isUndefined(this._transform_map[_offset])) {
      this._transform_map[_offset] = this._events.send(
        'transform', _.identity,
          this._buffer[_offset], this._transform_data
      );
    }

    return this._transform_map[_offset];
  }
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Iterator;
export { Iterator as base };

