
'use strict';

import utils from '../utils';
import Iterator from './base';
import _ from '../utils/lodash';
import promise from '../promise';
import exception from '../exception';

/**
 * @name Iterator.Join:
 *  I heard you liked iterators.  Walk through the sorted iterators
 *  `_l` and `_r` in parallel. The join iterator will only emit
 *  pairs of items with matching keys. By default, "matching keys"
 *  means that the keys are strictly equal. If the `key.subset`
 *  event is handled, use it as a function to transform keys before
 *  comparing them, allowing for partial and/or prefix comparisons.
 */
var iterator_join = Iterator.Join = class extends Iterator {

  /**
   * @name constructor:
   */
  constructor (_l, _r, _options) {

    /* Must be first */
    super(_options);

    this._left = _l;
    this._right = _r;
    this._limit = (this.options.limit || Infinity);

    this.events.register('key.subset', true);
    return this._join_initialize();
  }

  /**
   * @name is_ordered:
   *   See superclass documentation for details.
   */
  is_ordered () {

    return true;
  }

  /**
   * @name seek:
   *   See superclass documentation for details.
   */
  seek (_key, _is_inexact, _force_advance) {

    let p = [
      this._left.seek(_key, _is_inexact, _force_advance),
      this._right.seek(_key, _is_inexact, _force_advance)
    ];

    return promise.all(p).then(() => {
      return super.seek(_key, _is_inexact, _force_advance);
    });
  }

  /**
   * @name rewind:
   *   See superclass documentation for details.
   */
  rewind (_key, _is_inexact, _force_advance) {

    let p = [
      this._left.rewind(_key, _is_inexact, _force_advance),
      this._right.rewind(_key, _is_inexact, _force_advance)
    ];

    this._join_initialize();

    return promise.all(p).then(() => {
      return super.rewind(_key, _is_inexact, _force_advance);
    });
  }

  /**
   * @name populate:
   *   See superclass documentation for details.
   */
  populate (_in_reverse, _is_sequential, _is_inexact) {

    let l = this._left;
    let r = this._right;
    let n = this._limit;

    let rv = [];
    let join = this._join;

    if (!_in_reverse && this._is_delayed_eof) {
      return promise.resolve(false);
    }

    /* While records remain on either side... */
    return promise.while(() => !l.is_eof() && !r.is_eof() && n > 0, () => {

      /* Map and compare */
      this._join_compare();

      /* Possible result, perform state transition */
      return this._join_result().then((_rv) => {

        if (_rv.length > 0) {
          --n;
        }

        rv = _.concat(rv, _rv);
        join.state = !join.state;

        if (join.comparison === 0) {
          return promise.resolve();
        }

        return (
          join.state ?
            r.seek(join.left_key, true, false) :
              l.seek(join.right_key, true, false)
        );
      });

    }).then(() => {

      /* Handle end-of-file */
      if (l.is_eof() || r.is_eof()) {
        if (rv.length > 0) {
          this._is_delayed_eof = true;
        } else {
          return promise.resolve(false);
        }
      }

      /* Success */
      this._buffer = rv;
      return super.populate(_is_sequential, _is_inexact);
    });
  }

  /** @private **/

  /**
   * @name _map_key:
   *   Pass `_key` through the key-mapping function that was
   *   provided to the class constructor at instantiation time,
   */
  _map_key (_key, _iterator, _is_left) {

    let join = this._join;

    if (_.isNil(_key)) {
      return _key;
    }

    return this.events.emit(
      'key.subset', _.identity, [ _key, _is_left, _iterator ]
    );
  }

  /**
   * @name _join_initialize:
   *   Build the necessary data structures and internal state
   *   required to start a relational join operation. This data
   *   is stored beneath `this._join`.
   */
  _join_initialize () {

    let options = this.options;

    let rv = {
      state: null, key: null, refresh_keys: true,
      left_key: null, right_key: null, comparison: null
    };

    if (!this._left.is_ordered() || !this._right.is_ordered()) {
      throw new exception.argument('Iterators must be ordered');
    }

    this._join = rv;
    return this;
  }

  /**
   * @name _join_compare:
   *   Perform a single key mapping and comparison operation for
   *   the relational join currently in progress.
   */
  _join_compare () {

    let l = this._left;
    let r = this._right;
    let join = this._join;

    if (join.refresh_keys || !join.state) {
      join.left_key = this._map_key(l.current_key(), l, true);
    }

    if (join.refresh_keys || join.state) {
      join.right_key = this._map_key(r.current_key(), r, false);
    }

    if (join.refresh_keys) {
      join.refresh_keys = false;
    }

    join.comparison = l.compare(
      join.left_key, join.right_key
    );

    /* Current internal state of the join operation:
        False if the next seek operation should occur on the left
        iterator; true if it should occur on the right iterator. */

    join.state = (join.comparison < 0);
    return this;
  }

  /**
   * @name _join_result:
   *   If the comparison performed in `_join_compare` indicated
   *   equality, consume all consecutive records from both iterators
   *   that happen to have the same key. Return an array containing all
   *   possible pairs of records, i.e. the cross-join of all records that
   *   have identical keys.
   */
  _join_result () {

    let l = this._left;
    let r = this._right;
    let join = this._join;

    if (join.comparison !== 0) {
      return promise.resolve([]);
    }

    var p = [
      this._consume_matching(l, l.current_key(), true),
      this._consume_matching(r, r.current_key(), false)
    ];

    return promise.all(p).then((_rv) => {
      return this._join_refresh()._cartesian_product(_rv[0], _rv[1]);
    });
  }

  /**
   * @name _join_refresh:
   *   Forcibly refresh the internal join data, regardless of the
   *   current join `state`.
   */
  _join_refresh () {

    this._join.refresh_keys = true;
    return this._join_compare();
  }

  /**
   * @name _consume_matching:
   *   Return an array of all results from `_iterator` that match
   *   `_key`. We pull these into memory instead of rewinding iterators,
   *   as iterators only know how to seek/rewind to the first occurrance
   *   of a particular key (currently by design).
   */
  _consume_matching (_iterator, _key, _is_left) {

    let rv = [];
    let join = this._join;
    let key = this._map_key(_key, _iterator, _is_left);

    return promise.while(() => {

      let k = key;

      key = this._map_key(
        _iterator.current_key(), _iterator, _is_left
      );

      return (_iterator.compare(key, k) === 0);

    }, () => {

      rv.push(_iterator.current());
      return _iterator.next();

    }).then(() => rv);
  }

  /**
   * @name _cartesian_product:
   *   Generate the Cartesian product of `_left_array` and
   *   `_right_array`, also known as a cross join. The Cartesian
   *   product of `_left_array` and `_right_array` is all possible
   *   2-tuples `{ left: a, right: b }`, where `a` is in `_left_array`
   *   and `b` is in `_right_array`.
   */
  _cartesian_product (_left_array, _right_array) {

    let rv = [];

    for (let i = 0, l1 = _left_array.length; i < l1; ++i) {
      for (let j = 0, l2 = _right_array.length; j < l2; ++j) {
        rv.push({ left: _left_array[i], right: _right_array[j] });
      }
    }

    return rv;
  }
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Iterator.Join;
export { iterator_join as join };

