
'use strict';

import Iterator from './base';
import _ from '../utils/lodash';
import promise from '../promise';

/**
 * @name Iterator.View:
 *   An iterator subclass for database views.
 */
var iterator_view = Iterator.View = class extends Iterator {

  /**
   * @name constructor:
   */
  constructor (_schema, _view_name, _options) {

    /* Must be first */
    super(_options);

    this._schema = _schema;
    this._view_name = _view_name;
    this._view_options = (this.options.view || {});

    /* Special case:
        A limit of one will prevent us from seeking properly.
        Enforce a minimum of two records per fill operation. */

    let limit = this._view_options.limit;

    if (_.isInteger(limit) && limit <= 1) {
      this._view_options.limit = 2;
    }

    return this;
  }

  /**
   * @name populate:
   *   See superclass documentation for details.
   */
  populate (_in_reverse, _is_sequential, _is_inexact, _is_initial) {

    let schema = this.schema;
    let view_name = this._view_name;
    let view_options = _.clone(this._view_options);

    /* Build upper/lower bound */
    if (!_.isNil(this.key)) {

      let endkey = view_options.endkey;

      if (!_.isUndefined(endkey)) {
        if (this.compare(this.key, endkey) >= 0) {
          return promise.resolve(false);
        }
      }

      view_options.startkey = this.key;
    }

    /* Perform direct database request */
    return schema.view(view_name, view_options).then((_r) => {

      let rv = _r.rows;

      if (rv.length <= 0) {
        return false;
      }

      /* Non-sequential random seek:
          Easy, just return the results. */

      if (!_is_sequential) {
        this.buffer = rv;
        return super.populate(_is_sequential, _is_inexact);
      }

      /* Handle sequential access:
          Rules differ depending upon iteration direction. */

      if (_in_reverse) {

        if (_is_initial) {
          return this.repopulate(
            _in_reverse, _is_sequential, _is_inexact, false
          );
        }

        if (!_.isNil(this.key)) {
          rv = rv.slice(1);
        }

        this.offset = (rv.length > 0 ? rv.length - 1 : 0);

      } else {

        if (!_is_initial && rv.length <= 1) {
          return false;
        }

        this.offset = 1;
      }

      /* Success */
      this.buffer = rv;
      return super.populate(_is_sequential, _is_inexact);
    });
  }

  /**
   * @name key_for:
   *   See superclass documentation for details.
   */
  key_for (_item) {

    return super.key_for(_item.key);
  }

  /**
   * @name is_ordered:
   *   See superclass documentation for details.
   */
  is_ordered () {

    return true;
  }

  /**
   * @name compare:
   *   See superclass documentation for details.
   */
  compare (_l, _r) {

    return (
      this.schema.collate(_l, _r) *
        (this._view_options.descending ? -1 : 1)
    );
  }

  /**
   * @name get schema:
   */
  get schema () {

    return this._schema;
  }
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Iterator.View;
export { iterator_view as view };

