
'use strict';

import promise from 'lie';
import _ from './utils/lodash';

/**
 * @name _while:
 *   A promise-based asynchronous `while` loop. Repeatedly call
 *   `_promise_fn` while `_condition_fn` returns a true-like value.
 *   The return value of `_promise_fn` must be a promise; rejection
 *   will stop the loop and be passed back along to the caller. This
 *   function is asynchronous; it returns a promise that is fulfilled
 *   or rejected upon completion of the loop.
 */
let _while = (_condition_fn, _promise_fn) => {

  /* Promise wrapper */
  return new promise((_resolve_fn, _reject_fn) => {

    /* Single step */
    let iteration_fn = (...args) => {

      if (!_condition_fn()) {
        return _resolve_fn(...args);
      }

      return _promise_fn().then(iteration_fn).catch(_reject_fn);
    };

    /* Start looping */
    return iteration_fn();
  });
};

/**
 * @name similar:
 *   Return true if `_value` is a promise-like object with
 *   promise-like methods. Otherwise, return false.
 */
let similar = (_value) => {

  return (
    _.isObject(_value) &&
      _.isFunction(_value.then) && _.isFunction(_value.catch)
  );
};

/**
 * @name all:
 *   Passthrough function.
 */
let all = _.bind(promise.all, promise);

/**
 * @name race:
 *   Passthrough function.
 */
let race = _.bind(promise.race, promise);

/**
 * @name reject:
 *   Passthrough function.
 */
let reject = _.bind(promise.reject, promise);

/**
 * @name resolve:
 *   Passthrough function.
 */
let resolve = _.bind(promise.resolve, promise);

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default {
  all: all, base: promise,
  race: race, reject: reject,
  resolve: resolve, similar: similar, while: _while
};

