/* globals emit */

'use strict';

import Base from '../base';
import utils from '../utils';
import _ from '../utils/lodash';
import promise from '../promise';
import iterator from '../iterator';
import exception from '../exception';
import { collate } from 'pouchdb-collate';

/**
 * @name Schema:
 */
var Schema = class extends Base {

  /**
   * @name constructor:
   */
  constructor (_path, _options) {

    /* Must be first */
    super(_options);

    this._name = (this.option('name') || 'default');
    this._path = (_path || this.name);

    return this;
  }

  /**
   * @name get path:
   */
  get path () {

    return this._path;
  }

  /**
   * @name get name:
   */
  get name () {

    return this._name;
  }

  /**
   * @name get options:
   */
  get options () {

    return this._options;
  }

  /**
   * @name deploy:
   *   Update the database schema to the most recent version.
   */
  deploy () {

    return promise.reject(
      new exception.argument('Abstract class; nothing to deploy')
    );
  }

  /**
   * @name undeploy:
   *   Irreversibly remove the currently-installed schema to the
   *   extent possible, but leave any and all stored data intact.
   *   Returns a promise that resolves to the previously-installed
   *   schema version number, or zero if no schema was installed.
   */
  undeploy () {

    return promise.resolve();
  }

  /**
   * @name destroy:
   *   Irreversibly remove the currently-installed schema and all
   *   documents. This results in total destruction of the database.
   *   Returns a promise that resolves to the previously-installed
   *   schema version number, or zero if no schema was installed.
   */
  destroy () {

    return promise.resolve();
  }

  /**
   * @name cleanup:
   *   Release any resources that are held by this instance.
   *   Returns a promise that is fulfilled on success.
   */
  cleanup () {

    return promise.resolve();
  }

  /**
   * @name compact:
   *   Find and release any unused storage back to the
   *   underlying database system and/or OS.
   */
  compact() {

    return promise.resolve();
  }

  /**
   * @name version:
   *   Get the remote database's schema version. Returns a
   *   promise that yields an integer value once fulfilled.
   */
  version () {

    return promise.resolve(-1);
  }

  /**
   * @name get latest_version:
   *   Return the schema version that's implemented by this
   *   class. Returns a single integer value.
   */
  get latest_version () {

    return -1;
  }

  /**
   * @name store:
   *   Store a new document beneath this schema.
   */
  store (_doc, _options) {

    return promise.reject();
  }

  /**
   * @name remove:
   *   Remove a document from beneath this schema. If the `force`
   *   option is set to a true-like value, issue a hard deletion
   *   request. Otherwise, mark the document as deleted but preserve
   *   the contents of the document so that the deletion can be
   *   properly replicated to other nodes.
   */
  remove (_doc, _options) {

    return promise.reject();
  }

  /**
   * @name all:
   *   Retrieve all documents from beneath this schema.
   */
  all (_options) {

    return promise.reject();
  }

  /**
   * @name get:
   *   Retrieve a document given its unique identifier.
   */
  get (_id, _options) {

    return promise.reject();
  }

  /**
   * @name view:
   *   Retrieve the contents of a database view.
   */
  view (_name, _options) {

    return promise.reject();
  }

  /**
   * @name view_iterator:
   *   Retrieve the contents of a database view, but wrap the
   *   results in an `Iterator.View` instance. This allows you
   *   to use `_options.limit` to lazily request chunks of data.
   */
  view_iterator (_name, _options) {

    return new iterator.view(this, _name, _options);
  }

  /**
   * @name remove_forced:
   *   Forcibly remove a document, retrying repeatedly if a
   *   conflict is encountered. This function is not safe for
   *   every concurrent scenario; use caution.
   */
  remove_forced (_doc, _retries) {

    let n = 0;
    let retries = (_retries || 4);

    if (!_.isInteger(retries) && !_.isFunction(retries)) {
      throw new exception.runtime(
        'Retry count must be an integer or function'
      );
    }

    return utils.make_promise(retries, _doc).then((_n) => {

      n = _n;
      return this.remove(_doc);

    }).catch((_e) => {

      if (n <= 0 || _e.status !== 409) {
        throw _e;
      }

      return this.get(_doc._id).then((_d) => {
        return this.remove_forced(
          _d, (_.isFunction(retries) ? retries : n - 1)
        );
      });
    });
  }

  /**
   * @name store_linked:
   *   Create a new anonymous document from `_linked_doc`, then update
   *   `_doc` with the identifier of that new anonymous document. To figure
   *   out what property to update, call `_update_path` with a single
   *   boolean argument indicating link direction; the result of this will
   *   be passed to the lodash `_.set` method. You can also use a literal
   *   string if you don't need to differentiate based upon link direction.
   *   If a conflict occurs, attempt to force-remove the anonymous document;
   *   if this fails, note the error in `_err.linked` and defer garbage
   *   collection to the caller.
   *
   *   If `_doc` has a valid `_id` at the time this function is called, that
   *   identifier is added to `_linked_doc` along `_update_path`. While
   *   `_doc` will be auto-created if necessary, its new unique identifier
   *   will not be added to `_linked_doc` in this case (i.e. you'll get only
   *   one direction of linkage, from `_doc` to `_linked_doc`). If you need
   *   to be able to find `_doc` given only `_linked_doc` at some point in
   *   the future, you should create `_doc` before calling this function.
   */
  store_linked (_doc, _linked_doc, _update_path, _options) {

    let linked_doc = _linked_doc;
    let options = (_options || {});

    let update_path = (
      (_.isNil(_update_path) || _.isString(_update_path)) ?
        ((_is_linked_doc) => (_update_path || 'linked_id')) : _update_path
    );

    if (!_.isFunction(update_path)) {
      return promise.reject(
        new exception.argument(
          'The `update_path` parameter must be a function'
        )
      );
    }

    /* Backward link */
    if (_doc._id) {
      _.set(linked_doc, update_path(true), _doc._id);
    }

    /* Store */
    return this.store(linked_doc).then((_ld) => {

      /* Forward link */
      linked_doc = _ld;
      _.set(_doc, update_path(false), linked_doc._id);

      return this.store(_doc).catch((_err) => {

        if (_err.status !== 409) {
          throw _err;
        }
        return this.remove_forced(_ld, options.retries).catch((_e) => {
          _err.linked = _.assignIn(_e, { doc: _ld });
        }).then((_r) => {
          throw _err;
        });
      });
    }).then((_d) => {
      return {
        doc: _d,
        linked_doc: linked_doc
      };
    });
  }

  /**
   * @name remove_linked:
   *   Retrieve the linked document identified by `_doc[_property_path]`,
   *   delete `_doc`, then force-delete the linked document.
   */
  remove_linked (_doc, _property_path, _options) {

    let linked_doc = null;
    let options = (_options || {});
    let property_path = (_property_path || 'linked_id');

    return this.get(_doc[property_path]).then((_linked_doc) => {
      linked_doc = _linked_doc;
      return this.remove(_doc);
    }).then((_doc) => {
      return this.remove_forced(linked_doc).then((_linked_doc) => {
        return { doc: _doc, linked_doc: _linked_doc };
      });
    });
  }

  /** @protected: */

  /**
   * @name collate:
   *   An object comparison function. Returns -1 if `_l` comes lexically
   *   prior to `_r`, 1 if `_l` comes lexically after `_r`, and 0 if the
   *   two strings are lexically indistinguishable. This function should
   *   represent the underlying database's idea of collation order, and
   *   should not change based upon the execution environment or locale.
   *   The function is used to ensure consistent sorting and binary
   *   searches, not consistent text ordering for user presentation.
   */
  collate (_l, _r) {

    return collate(_l, _r);
  }

};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Schema;
export { Schema as base };

