
'use strict';

import view from '../view';
import Schema from './base';
import utils from '../utils';
import pouchdb from 'pouchdb';
import _ from '../utils/lodash';
import promise from '../promise';
import pouchdb_views from './pouchdb/views';

/**
 * @name Schema.PouchDB
 */
var schema_pouchdb = Schema.PouchDB = class extends Schema {

  /**
   * @name constructor:
   */
  constructor (_path, _options) {

    /* Must be first */
    super(_path, _options);

    /* Currently-implemented schema version */
    this._latest_version = 1;

    /* Design document unique identifier */
    this._ddoc_id = [ '_design', this.name ].join('/');

    return this._initialize_pouchdb();
  }

  /**
   * @name design_document:
   *   Return the raw design document if one exists. Returns
   *   a promise. If no schema has been deployed, reject the
   *   returned promise with an error.
   */
  design_document () {

    return this.get(this._ddoc_id);
  }

  /**
   * @name get latest_version:
   *   Return the latest schema version that's implemented
   *   by Schema.PouchDB. Returns a single integer value.
   */
  get latest_version () {

    return this._latest_version;
  }

  /**
   * @name version:
   *   Get the remote database's schema version. Returns a
   *   promise that yields an integer value once fulfilled.
   */
  version () {

    return this.design_document()
      .then((_ddoc) => _ddoc.version)
      .catch(this._handle_missing_version);
  }

  /**
   * @name deploy:
   *   Update the database schema to the most recent version.
   */
  deploy () {

    return this.version().then((_version) => {

      if (_version >= this.latest_version) {
        return _version;
      }
      return this.store(this._build_design_document()).then((_doc) => {
        return _version;
      });
    });
  }

  /**
   * @name undeploy:
   *   Irreversibly remove the currently-installed schema to the
   *   extent possible, but leave any and all stored data intact.
   *   Returns a promise that resolves to the previously-installed
   *   schema version number, or zero if no schema was installed.
   */
  undeploy () {

    return this.design_document().then((_ddoc) => {
      return this.remove(_ddoc).then(() => {
        return _ddoc.version;
      });
    }).catch(this._handle_missing_version);
  }

  /**
   * @name destroy:
   *   Irreversibly remove the currently-installed schema and all
   *   documents. This results in total destruction of the database.
   *   Returns a promise that resolves to the previously-installed
   *   schema version number, or zero if no schema was installed.
   */
  destroy () {

    let p = this.undeploy().then((_previous_version) => {
      return this._db.destroy().then(() => {
        this._initialize_pouchdb();
        return _previous_version;
      });
    });

    return this._propagate_pouchdb_error(p);
  }

  /**
   * @name compact:
   *   Find and release any unused storage back to the
   *   underlying database system and/or OS.
   */
  compact () {

    let p = this._db.compact();
    return this._propagate_pouchdb_error(p);
  }

  /**
   * @name store:
   *   Store a new document beneath this schema. If a `_rev`
   *   field is provided in `_doc`, update the document.
   *   Otherwise, create a new document. If an `_id` field is
   *   provided in `_doc`, use that as the unique identifier.
   *   Otherwise, generate a new universally-unique identifier.
   */
  store (_doc, _options) {

    var doc = this._preprocess_document(_doc);

    if (_.isNil(doc._id)) {
      doc._id = utils.uuid.compact();
    }

    let p = this._db.put(doc).then((_r) => {
      return this._postprocess_document(doc, _r);
    });

    return this._propagate_pouchdb_error(p);
  }

  /**
   * @name remove:
   *   Remove a document from beneath this schema.
   */
  remove (_doc, _options) {

    let p = null;
    let options = (_options || {});
    let doc = this._preprocess_document(_doc);

    if (options.force) {
      return this._propagate_pouchdb_error(
        this._db.remove(doc)
      );
    }

    return this.store(
      _.assign(doc, { _deleted: true })
    );
  }

  /**
   * @name get:
   *   Retrieve a document given its unique identifier.
   */
  get (_id, _options) {

    let p = this._db.get(_id);
    return this._propagate_pouchdb_error(p);
  }

  /**
   * @name all:
   *   Retrieve all documents from beneath this schema.
   */
  all (_options) {

    let p = this._db.allDocs(_options || {});
    return this._propagate_pouchdb_error(p);
  }

  /**
   * @name view:
   *   Retrieve the contents of a database view.
   */
  view (_name, _options) {

    let p = this._db.query(
      [ this.name, _name ].join('/'), (_options || {})
    );

    return this._propagate_pouchdb_error(p);
  }

  /** @private: **/

  /**
   * @name _build_design_document:
   *   Return an object representing the latest database schema.
   *   This object is PouchDB-specific, and can contain both
   *   application code and map/reduce database views.
   */
  _build_design_document () {

    let views = this._build_views();

    _.each(views, (_v, _k) => {
      views[_k] = (new view.pouchdb(_v)).build().result;
    });

    return {
      views: views,
      _id: this._ddoc_id,
      version: this.latest_version
    };
  }

  /**
   * @name _build_views:
   *   Return an object containing all map/reduce views that
   *   are required for a functioning database schema.
   */
  _build_views () {

    return pouchdb_views;
  }

  /**
   * @name _initialize_pouchdb:
   *   Create a new PouchDB instance.
   */
  _initialize_pouchdb () {

    /* Support global loading */
    /* istanbul ignore next */
    let constructor = (pouchdb || PouchDB);

    this._db = new constructor(
      _.assign({}, (this.option('pouchdb') || {}), {
        name: this.path
      })
    );

    return this;
  }

  /**
   * @name _propagate_pouchdb_error:
   */
  _propagate_pouchdb_error (_promise) {

    return _promise.catch((_e) => {
      throw this._rewrite_pouchdb_error(_e);
    });
  }

  /**
   * @name _rewrite_pouchdb_error:
   */
  _rewrite_pouchdb_error (_err) {

    return _err;
  }

  /**
   * @name _handle_missing_version:
   *   Throw `_err` unless `_err.status` is a 404.
   */
  _handle_missing_version (_err) {

    /* istanbul ignore if */
    if (_err.status !== 404) {
      throw _err;
    }

    /* Default version */
    return 0;
  }

  /**
   * @name _preprocess_document:
   *   Perform pre-storage rewriting on a document.
   */
  _preprocess_document (_doc) {

    return _.clone(_doc);
  }

  /**
   * @name _postprocess_document:
   *   Perform post-storage rewriting on a document.
   */
  _postprocess_document (_doc, _result) {

    let rv = _doc;

    _.each([ 'id', 'rev' ], (_k) => {
      if (!_.isNil(_result[_k])) {
        rv[`_${_k}`] = _result[_k];
      }
    });

    return rv;
  }
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Schema.PouchDB;
export { schema_pouchdb as pouchdb };

