/* @flow */

'use strict';

import Base from './base';
import _ from './utils/lodash';
import foldcase from 'fold-case';
import porter from 'porter-stemmer';
import exception from './exception';
import unicode from './utils/unicode';
import htmlparser2 from 'htmlparser2-es6';

/**
 * @name offset_map:
 */
type offset_map = Array<{
  delta: number,
  offset: number
}>;

/**
 * @name Splitter:
 *   Base class for Unicode word splitters.
 */
var Splitter = class extends Base {

  /**
   * @name constructor:
   */
  constructor (_str: string, _options: ?Object) {

    /* Must be first */
    super(_options);

    this.reset();
    this._call_count = 0;
    this._skip_autoreset = (this.option('autoreset') === false);

    if (!_.isNil(_str)) {
      this.string = _str;
    }

    return this;
  }

  /**
   * @name set string:
   */
  set string (_str: string) {

    if (!_.isString(_str)) {
      throw new exception.argument('Argument must be a string');
    }

    this._string = _str;
    return this;
  }

  /**
   * @name get string:
   */
  get string (): string {

    return this._string;
  }

  /**
   * @name get result:
   */
  get result (): Array<any> {

    return this._result;
  }

  /**
   * @name split:
   */
  split (_str: string, _options: ?Object) {

    if (!_.isNil(_str)) {

      /* Reset if needed */
      if (!this._skip_autoreset) {
        this.reset();
      }

      /* String override */
      this.string = _str;
    }

    /* Sanity check */
    if (!_.isString(this.string)) {
      throw new exception.argument('No string provided');
    }

    this._call_count++;
    return this;
  }

  /**
   * @name reset:
   */
  reset () {

    this._tuple_count = 0;
    this._string = this._result = undefined;

    return this;
  }

  /**
   * @name verify:
   *   Verify that the `result` getter returns an object that
   *   is logically consistent with the `string` getter. This
   *   is essentially self-test code, and may be removed in a
   *   later version.
   */
  verify () {

    var rv = { ok: true, errors: 0 };
    var s = this.string, r = this.result;

    if (!_.isArrayLikeObject(r)) {
      rv.ok = false;
      return rv;
    }

    _.each(r, (_r) => {
      if (s.substring(_r[1], _r[1] + _r[0].length) !== _r[0]) {
        rv.errors++;
        rv.ok = false;
      }
    });

    return rv;
  }

  /** @protected **/

  /**
   * @name set result:
   */
  set result (_result: Array<any>) {

    this._result = _result;
  }

  /**
   * @name fold:
   *   Return a version of `_string` with all uppercase characters
   *   turned in to lowercase characters, according to the Unicode
   *   "case folding" algorithm.
   */
  fold (_string: string) {

    return foldcase(_string);
  }

  /**
   * @name build_tuple:
   *   Build a result object for the word `_word` appearing at
   *   integer byte offset `_offset`.
   */
  build_tuple (_word: string, _offset: number,
               _length: number, _options: ?Object) {

    let extra = {};
    let folded_word = _word;

    let options = this.options;
    let tuple_options = (_options || {});
    let offset_start = (tuple_options.offset || 0);

    if (options.fold || options.stem) {
      folded_word = this.fold(_word);
    }

    let rv = [
      (options.fold ? folded_word : _word),
        offset_start + _offset, _length, this._tuple_count++
    ];

    if (options.stem) {
      extra.stem = this.stem(folded_word, options.stem);
    }

    if (!_.isEmpty(extra)) {
      rv.push(extra);
    }

    return rv;
  }

  /**
   * @name stem:
   */
  stem (_word: string, _language: string) {

    switch (_language) {
      case 'en': case true:
        return porter.stemmer(_word);
      default:
        throw new exception.argument('Invalid language');
    }
  }
};

/**
 * @name Splitter.Text:
 *   A word splitter for plain Unicode text.
 */
var splitter_text = Splitter.Text = class extends Splitter {

  /**
   * @name constructor:
   */
  constructor (_str: string, _options: ?Object) {

    /* Must be first */
    super(_str, _options);

    return this._initialize_regex();
  }

  /**
   * @name split:
   */
  split (_str: string, _options: ?Object) {

    let delta = 0;
    let offset_map = [];
    let options = (_options || {});

    /* HTML entity decoding:
        If requested, decode HTML entities and keep an ordered offset
        map in order to locate the entities in the original markup. */

    if (options.entities) {
      _str = this._decode_entities(_str, offset_map);
    }

    /* Must be first */
    super.split(_str);

    switch (options.type) {
      default:
      case 'native':
        this.result = this._split_regex(this.string, offset_map, _options);
        break;
    }

    return this;
  }

  /** @private **/

  /**
   * @name _initialize_regex:
   */
  _initialize_regex () {

    let f1, f2;

    let p = '(?:' + [
      unicode.regex_for_category('P', true),
      unicode.regex_for_category('Sc', false),
      unicode.regex_for_category('Sk', false),
      unicode.regex_for_category('Sm', false)
    ].join('|') + ')';

    let z = unicode.regex_for_category('Z', true);
    let cf = unicode.regex_for_category('Cf', true);
    let cc = unicode.regex_for_category('Cc', true);

    this._regex = new RegExp(
      `(?:^${p}+)|(?:${p}*(?:(?:${z}|${cf}|${cc})+|$)${p}*)`, 'g'
    );

    return this;
  }

  /**
   * @name _decode_entities:
   */
  _decode_entities (_s: string, _offset_map: offset_map) {

    let total_delta = 0;

    return htmlparser2.entities.decodeHTML(_s, (_from, _to, _i) => {

      let dx = (_to.length - _from.length);

      _offset_map.push({
        delta: dx, offset: _i + total_delta
      });

      total_delta += dx;
      return _to;
    });
  }

  /**
   * @name _split_regex:
   *   Returns an array of two-element tuples; the first element of
   *   each tuple is the word, and the second is the zero-based code
   *   unit offset.
   */
  _split_regex (_str: string, _offset_map: offset_map, _options: ?Object) {

    let i = 0;
    let map_i = 0, delta = 0;

    let m = {}, rv = [];
    let re = this._regex;

    let len = _str.length;
    let map_len = _offset_map.length;

    /* For every delimiter match */
    while (i < len && (m = re.exec(_str)) !== null) {

      let word = _str.substring(i, m.index);

      let word_length = word.length;
      let word_adjust = 0, word_end = i + word_length;

      /* HTML entities fixup:
          Map locations back to their original offsets. */

      for (; map_i < map_len; ++map_i) {

        if (word_end <= _offset_map[map_i].offset) {
          break;
        }

        if (i > _offset_map[map_i].offset) {
          delta += _offset_map[map_i].delta;
        } else {
          word_adjust += _offset_map[map_i].delta;
        }
      }

      /* Handle start of string:
          Return token preceeding delimiter match. */

      if (i < m.index) {
        rv.push(this.build_tuple(
          word, i - delta, word_length - word_adjust, _options
        ));
      }

      /* Next delimiter */
      delta += word_adjust;
      i = m.index + m[0].length;
    }

    re.lastIndex = 0;
    return rv;
  }

};

/**
 * @name Splitter.Html:
 *   A word splitter for Unicode HTML markup.
 */
var splitter_html = Splitter.Html = class extends Splitter {

  /**
   * @name constructor:
   */
  constructor (_str: string, _options: ?Object) {

    /* Must be first */
    super(_str, _options);

    this._text_options = (this.option('text') || {});

    this._text_splitter = new Splitter.Text(
      null, _.assign({}, this._text_options, {
        autoreset: false, entities: true
      })
    );

    return this;
  }

  /**
   * @name split:
   */
  split (_str: string, _options: ?Object) {

    let options = (_options || {});

    /* Must be first */
    super.split(_str, options);

    switch (options.type) {
      default:
      case 'native':
        this.result = this._split_html(this.string);
        break;
    }

    return this;
  }

  /**
   * @name reset:
   */
  reset () {

    /* Must be first */
    super.reset();

    if (this._text_splitter) {
      this._text_splitter.reset();
    }

    return this;
  }

  /** @private **/

  /**
   * @name _split_html:
   *   Returns an array of two-element tuples; the first element of each
   *   tuple is the word, and the second is the zero-based code unit offset.
   */
  _split_html (_str: string) {

    let rv = [];

    let parser = new htmlparser2({
      ontext: (_text) => {
        rv = rv.concat(
          this._text_splitter.split(_text, {
            entities: true, offset: parser.startIndex
          }).result
        );
      }
    });

    parser.write(_str);
    parser.end();

    return rv;
  }
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export {
  Splitter as base,
  splitter_text as text,
  splitter_html as html
};

export default {
  base: Splitter,
  text: Splitter.Text,
  html: Splitter.Html
};

