
'use strict';

import jsdump from 'jsDump';
import _ from './utils/lodash';
import uuid_module from 'uuid';
import promise from './promise';
import exception from './exception';
import basex from './utils/basex-tiny';

/**
 * @name logical:
 *   Functional versions of common boolean logic operations.
 */
let logical = {

  /**
   * @name or:
   *   Return a function that valuate each function in `_fns` in order.
   *   If any function in the list yields a true-like value, immediately
   *   return true.  Otherwise, continue evaluating and return false
   *   after all functions have been evaluated. In other words: compute
   *   the logical `or` of the return value of each function in `_fns`,
   *   short-circuiting if necessary.
   */
  or: (..._fns) => {

    let rv = false;

    return ((..._args) => {

      for (let i = 0, len = _fns.length; i < len; ++i) {

        if (_fns[i](..._args)) {
          rv = true; break;
        }
      }

      return rv;
    });
  },

  /**
   * @name and:
   *   Return a function that evaluates each function in `_fns` in
   *   order. If any function in the list yields a false-like value,
   *   immediately return false.  Otherwise, continue evaluating and
   *   return true after all functions have been evaluated. In other
   *   words: compute the logical `and` of the return value of each
   *   function in `_fns`, short-circuiting if necessary.
   */
  and: (..._fns) => {

    let rv = true;

    return ((..._args) => {

      for (let i = 0, len = _fns.length; i < len; ++i) {

        if (!_fns[i](..._args)) {
          rv = false; break;
        }
      }

      return rv;
    });
  }
};

/**
 * @name search:
 *   General-purpose searching and sorting algorithms.
 */
let search = {

  /**
   * @name binary:
   *   A simple binary search implementation. Adapted from
   *   a public domain implementation available on `npm` at
   *   http://npmjs.com/package/binary-search.
   */
  binary: (_haystack, _needle,
           _compare_fn, _map_fn, _low, _high) => {

    if (_haystack.length <= 0) {
      return false;
    }

    if (!_map_fn) {
      _map_fn = ((_x) => _x); /* Identity */
    }

    if (_.isNil(_low)) {
      _low = 0;
    } else {
      _low = (_low | 0);
      if (_low < 0 || _low >= _haystack.length) {
        throw new exception.range('Invalid lower bound');
      }
    }

    if (_.isNil(_high)) {
      _high = _haystack.length - 1;
    }
    else {
      _high = (_high | 0);
      if (_high < _low || _high >= _haystack.length) {
        throw new exception.range('Invalid upper bound');
      }
    }

    while (_low <= _high) {

      /* Prevent overflow:
          Don't use `(_low + _high) / 2`, it can wrap around. */

      let mid = _low + (_high - _low >> 1);

      let cmp = _compare_fn(
        _map_fn(_haystack[mid]), _needle, mid, _haystack
      );

      if (cmp < 0) {
        /* Too low */
        _low  = mid + 1;
      } else if (cmp > 0) {
        /* Too high */
        _high = mid - 1;
      } else {
        /* Key found */
        return mid;
      }
    }

    /* Key not found */
    return ~_low;
  }
};

/**
 * @name uuid:
 *   Functions for generating (sometimes-universally) unique
 *   identifiers. Warning: these functions make dramatically
 *   different security and predictability guarantees.
 */
let uuid = {

  /** @public **/

  /**
   * @name memory:
   *   The fastest possible implementation. This function makes no
   *   guarantees of universal uniqueness or the unpredictability of
   *   identifiers, only that a single process is unlikely to ever see
   *   a duplicate identifier. Returns a string containing the (U)UID.
   *   Do not use this for any externally-visible identifiers: the
   *   output of this function is intended for in-memory use only.
   */
  memory: () => {

    return [ Date.now(), ++uuid._memory.counter ].join('-');
  },

  /**
   * @name compact:
   *   A production-suitable UUID implementation that produces short
   *   structured identifiers using a range of ASCII letters. The
   *   prefix portion of each identifier is primarily time-derived,
   *   while the suffix portion is randomly-generated session identifier
   *   that *may* remain unchanged for the lifetime of the interpreter.
   *   You should *not* rely on the session identifier remaining stable.
   *   These identifiers may be expensive to compute, and are not
   *   suitable for in-memory-only uses.
   */
  compact: () => {

    let b = new Uint8Array(16);
    let rv = new Uint8Array(16);

    uuid_module.v1(null, b);

    rv[0] = b[10]; rv[1] = b[11];
    rv[2] = b[12]; rv[3] = b[13];
    rv[4] = b[14]; rv[5] = b[15];

    rv[6] = b[6]; rv[7] = b[7];
    rv[8] = b[4]; rv[9] = b[5];

    rv[10] = b[0]; rv[11] = b[1];
    rv[12] = b[2]; rv[13] = b[3];

    rv[14] = b[8]; rv[15] = b[9];

    return basex.encode(rv, uuid._compact.alphabet);
  },

  /**
   * @name rfcv1:
   *   A production-suitable UUID implementation that produces long
   *   identifiers based upon the RFC4122 v1 UUID specification.
   *   Includes a timestamp portion.
   */
  rfcv1: () => {

    return uuid_module.v1();
  },

  /**
   * @name rfcv4:
   *   A production-suitable UUID implementation that produces long
   *   identifiers based upon the RFC4122 v1 UUID specification.
   *   Random only; does not include a timestamp portion.
   */
  rfcv4: () => {

    return uuid_module.v4();
  },


  /** @protected **/

  /**
   * _memory:
   *   Private storage for the `uuid.memory` implementation.
   */
  _memory: {
    counter: 0,
    timestamp: Date.now()
  },

  /**
   * _compact:
   *   Private storage for the `uuid.compact` implementation.
   */
  _compact: {
    alphabet: '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
  }
};

/**
 * @name is_subset:
 *   Recursively determine if `_needle` is a subset of `_haystack`.
 *   Return true if `_needle` is a subset of `_haystack`, and false
 *   otherwise.
 */
let is_subset = (_haystack, _needle) => {

  return _is_subset(_haystack, _needle, 10);
};

/**
 * @name _is_subset:
 *   Internal implementation of `is_subset`. This function should
 *   not be called directly; use the public `is_subset` instead.
 */
let _is_subset = (_haystack, _needle, _limit) => {

  let rv = true;

  if (_limit <= 0) {
    throw new exception.argument(
      'Data structure recursion limit exceeded'
    );
  }

  let is_enumerable = (
    (_.isArray(_needle) && _.isArray(_haystack)) ||
      (_.isPlainObject(_needle) && _.isPlainObject(_haystack))
  );

  if (!is_enumerable) {
    return (_needle === _haystack);
  }

  if (_.size(_haystack) < _.size(_needle)) {
    return false;
  }

  _.each(_needle, (_v, _k) => {
    if (!_is_subset(_haystack[_k], _v, _limit - 1)) {
      return (rv = false);
    }
  });

  return rv;
};

/**
 * @name paths_for:
 *   Return an array of all possible lodash-style property paths that
 *   have a usable value in `_object`. If `_options` is an object and
 *   has a true-like `skip_arrays` property, the function will report
 *   the presence of arrays but refuse to examine their contents. If
 *   `_options` has a `skip_objects` property, the function will report
 *   the presence of objects beneath `_object` but will refuse to examine
 *   their contents. Returns an array of string-based object paths.
 */
let paths_for = (_object, _options) => {

  return _paths_for({}, '', _object, _options, 10);
};

/**
 * @name _paths_for:
 *   Internal implementation of `paths_for`. This function should
 *   not be called directly; use the public `paths_for` instead.
 */
let _paths_for = (_rv, _path, _object, _options, _limit) => {

  let options = (_options || {});
  let dot = (_path !== '' ? '.' : '');

  if (_limit <= 0) {
    throw new exception.argument(
      'Data structure recursion limit exceeded'
    );
  }

  if (_.isPlainObject(_object)) {

    /* Object: recur on contents */
    if (!options.skip_objects) {
      _.each(_object, (_v, _k) => {
        _paths_for(_rv, `${_path}${dot}${_k}`, _v, options, _limit - 1);
      });
    }

  } else if (_.isArrayLike(_object)) {

    /* Array: recur on contents */
    if (!options.skip_arrays) {
      _.each(_object, (_v, _i) => {

        let path = `${_path}[${_i}]`;

        if (_.isEmpty(_v)) {
          _rv[path] = true;
        } else {
          _paths_for(_rv, path, _v, options, _limit - 1);
        }
      });
    }

  } else {

    /* Scalar: finished */
    _rv[_path] = true;
  }

  return _rv;
};

/**
 * @name make_promise:
 *   Turn `_v` into a promise. If `_v` is not a function, return
 *   a resolved promise that evaluates to `_v`. Otherwise, call `_v`
 *   with `_args` as an argument list, and return a promisified
 *   version of its return value.
 */
let make_promise = (_v, ..._args) => {

  if (!_.isFunction(_v)) {
    return promise.resolve(_v);
  }

  let v = _v(..._args);

  if (!promise.similar(v)) {
    v = promise.resolve(v);
  }

  return v;
};

/**
 * @name pretty_print:
 *   Print a human-readable version of the argument `_v`,
 *   which may be of any type.
 */
let pretty_print = (_v) => {

  return jsdump.parse(_v);
};

/**
 * @name is_node:
 *   Return true if we're running under Node.js.
 */
let is_node = () => {

  return (typeof process === 'object');
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export {
  uuid, search, logical, paths_for,
    is_subset, make_promise, pretty_print
};

export default {
  uuid: uuid,
  search: search,
  logical: logical,
  is_node: is_node,
  paths_for: paths_for,
  is_subset: is_subset,
  make_promise: make_promise,
  pretty_print: pretty_print
};

