
'use strict';

import Base from './base';
import utils from './utils';
import _ from './utils/lodash';

/**
 * @name Validator:
 *   A simple object validation API for user-supplied data
 *   structures. Uses lodash-style object paths.
 */
var Validator = class extends Base {

  /**
   * @name validate:
   *   Validate the structure, content, or type signature of a user-supplied
   *   object using a non-recursive "schema" object. Within the schema, keys
   *   are lodash-style object paths; values are constraint sets - a fixed
   *   function or array of functions to be evaluated in order.  Constraints
   *   may set the special property `short_circuit` in their return value to
   *   stop processing immediately. If `_allow_extraneous` is true-like,
   *   don't fail if there are properties in `_object` that don't appear in
   *   `_schema`.
   */
  validate (_schema, _object, _allow_extraneous) {

    let rv = {
      ok: true,
      success: {}, failure: {}
    };

    if (!_.isPlainObject(_schema)) {
      return {
        ok: false, internal: true,
        message: 'Invalid schema format'
      };
    }

    if (!_.isPlainObject(_object)) {
      return {
        ok: false, internal: true,
        message: 'Object required for validation'
      };
    }

    let expected_keys = _.keys(_schema);

    /* For each constraint set... */
    for (let i = 0, len = expected_keys.length; i < len; ++i) {

      let key = expected_keys[i];
      let value = _.get(_object, key);

      /* Evaluate all constraints */
      let r = this.validate_constraints(_schema[key], value);

      if (r.ok) {
        rv.success[key] = r;
        continue;
      }

      rv.ok = false;
      rv.failure[key] = r;
    }

    /* Provide top-level error messages */
    this._lift_messages(rv, expected_keys);

    /* Detect unexpected properties */
    if (!_allow_extraneous) {

      /* Set difference: provided minus expected */
      var diff = _.omit(
        utils.paths_for(_object, { skip_arrays: true }), expected_keys
      );

      if (_.size(diff) > 0) {
        rv.ok = false;
        rv.unexpected = diff;
        rv.message = 'Unexpected parameter(s) provided';
      }
    }

    return rv;
  }

  /**
   * @name validate_constraints:
   */
  validate_constraints (_constraints, _value) {

    let rv = {
      ok: true,
      success: [], failure: []
    };

    let constraints = (
      _.isFunction(_constraints) ?
        [ _constraints ] : _constraints
    );

    if (!_.isArray(constraints)) {
      return {
        ok: false, internal: true,
        message: 'Invalid constraint set'
      };
    }

    /* For each constraint... */
    for (let i = 0, len = constraints.length; i < len; ++i) {

      let r = this.evaluate_constraint(constraints[i], _value);

      if (r.ok) {
        rv.success.push(r);
      } else {
        this._copy_result_properties(rv, r);
        rv.failure.push(r);
      }

      if (r.short_circuit) {
        break;
      }
    }

    return rv;
  }

  /**
   * @name evaluate_constraint:
   *   Evaluate the constraint `_constraint` against the user-supplied
   *   value `_value`. Constraint functions can return a simple boolean
   *   value or an object. Object-returning constraint functions *must*
   *   containing a boolean `ok`property. If the constraint evaluates to
   *   false, the return value *should* include a `message` property
   *   explaining the failure in human-understandable terms. Constraint
   *   functions *may* also include a boolean `short_circuit` property,
   *   regardless of success or failure: if true, the `ok` property will
   *   be considered final for the entire user-supplied value, overriding
   *   any previously-determined results.
   */
  evaluate_constraint (_constraint, _value) {

    /* Type check */
    if (!_.isFunction(_constraint)) {
      return {
        ok: false,
        internal: true, short_circuit: true,
        message: 'Constraint must be a function'
      };
    }

    /* Invoke */
    let rv = _constraint.call(_value, _value);

    /* Normal result format */
    if (_.isPlainObject(rv)) {
      return rv;
    }

    /* Stringify */
    let s = (_constraint.name || 'unknown');

    /* Simplified result format */
    if (_.isBoolean(rv)) {

      if (rv) {
        return { ok: true };
      }

      return {
        ok: false,
        message: `Constraint '${s}' must be true`
      };
    }

    return {
      ok: false,
      internal: true, short_circuit: true,
      message: `Constraint '${s}' yielded an invalid value`
    };
  }

  /**
   * @name optional:
   *   A short-circuiting validation rule for optional values.
   *   If `_value` is null-like, force the entire constraint
   *   to evaluate to trueusing the `short_circuit` property.
   *   Otherwise, evaluate to true and continue.
   */
  optional (_value) {

    if (_.isNil(_value)) {
      return {
        ok: true,
        short_circuit: true,
        message: 'Optional value omitted'
      };
    }

    return { ok: true };
  }

  /** @protected **/

  /**
   * @name _copy_result_properties:
   */
  _copy_result_properties (_to, _from) {

    _to.ok = _from.ok;
    _to.message = _from.message;
    _to.internal = _from.internal;
    _to.short_circuit = _from.short_circuit;

    return this;
  }

  /**
   * @name _lift_messages:
   *   "Lift" one or more result messages out of the subresult arrays
   *   and into `_result.message`. This provides a convenient top-level
   *   result/error message for human beings to read.
   */
  _lift_messages (_rv, _keys) {

    /* For each top-level result object... */
    _.each([ _rv.failure, _rv.success ], (_list) => {

      /* For each property... */
      _.each(_list, (_v, _k) => {

        let success = (_v.success || []);
        let failure = (_v.failure || []);

        /* Lift message to result object */
        if (failure.length > 0 && !_.isNil(failure[0].message)) {
          _v.message = failure[0].message;
        } else if (success.length > 0 && !_.isNil(success[0].message)) {
          _v.message = success[0].message;
        }

        /* Lift message to top */
        if (_.isNil(_rv.message) && !_.isNil(_v.message)) {
          _rv.message = _k + ': ' + _v.message;
        }
      });
    });

    return _rv;
  }
};


/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Validator;
export { Validator as validator };

