/* @flow */

'use strict';

import Base from './base';
import utils from './utils';
import _ from './utils/lodash';
import promise from './promise';
import exception from './exception';

/**
 * @name Worker:
 */
var Worker = class extends Base {

  /**
   * @name constructor:
   */
  constructor (_options: ?Object) {

    /* Must be first */
    super(_options);

    this._error_strings = {
      terminated: 'Worker thread has terminated'
    };

    /* Magic termination message:
        In browsers, we need to use a special message internally
        to determine if the worker has called `close`. Don't use
        this value in a call to `WorkerGlobalScope.postMessage`. */

    this._termination_message = (
      'worker:op:internal:\ud3ad\ubeef\ud3ad\ubeef:0'
    );

    return this.reset();
  }

  /**
   * @name reset:
   */
  reset () {

    this._worker = null;
    this._blob_url = null;
    this._is_running = false;
    this._is_terminated = false;

    this._wait_queue = [];
    this._send_buffer = [];
    this._message_queue = [];

    /* istanbul ignore next */
    this._worker_class = (
      utils.is_node() ?
        require('webworker-threads').Worker : window.Worker
    );

    return this;
  }

  /**
   * @name run:
   */
  run (_dependencies: ?Array<String>, _options: ?Object) {

    let code = this.main;
    let bootstrap_code = [];
    let send_buffer = this._send_buffer;
    let dependencies = (_dependencies || []);

    /* Dependency injection */
    if (dependencies.length > 0) {
      bootstrap_code = [
        'self.importScripts.apply(self, ',
          JSON.stringify(dependencies),
        ');',
      ];
    }

    /* Browser only */
    /* istanbul ignore if */
    if (!utils.is_node()) {

      /* Termination-tracking code */
      let termination_setup = [
        'self._worker_internal_close = self.close;',
        'self.close = function () {',
          'self.postMessage(',
            JSON.stringify(this._termination_message),
          ');',
          'return self._worker_internal_close();',
        '};'
      ];

      /* Generate trampoline */
      let trampoline = [
        '(function () { ',
          ...bootstrap_code, code.toString(), ...termination_setup,
          'main();',
        '})();'
      ];

      /* Generate URL */
      code = this._blob_url = URL.createObjectURL(
        new Blob(trampoline, {
          type: 'application/javascript; charset=utf-8'
        })
      );
    }

    this._is_running = true;
    this._worker = new this._worker_class(code);
    this._worker.onmessage = _.bind(this._handle_worker_message, this);

    /* Node only */
    if (utils.is_node() && bootstrap_code.length > 0) {
      this._worker.thread.eval(bootstrap_code.join(''));
    }

    /* Drain send buffer */
    if (send_buffer.length > 0) {
      for (var i = 0, len = send_buffer.length; i < len; ++i) {
        this._worker.postMessage(send_buffer[i]);
      }
    }

    this._send_buffer = [];
    this._is_running = true;

    return this;
  }

  /**
   * @name is_running:
   */
  is_running () {

    return this._is_running;
  }

  /**
   * @name is_terminated:
   */
  is_terminated () {

    return this._is_terminated;
  }

  /**
   * @name send:
   */
  send (_message: any) {

    const termination_message = 'Worker has terminated';

    if (this.is_terminated()) {
      throw new exception.argument(termination_message);
    }

    if (!this.is_running()) {
      this._send_buffer.push(_message);
      return this;
    }

    try {
      this._worker.postMessage(_message);
    } catch (_e) {
      this.terminate(); /* Handle termination (node) */
      throw new exception.argument(termination_message);
    }

    return this;
  }

  /**
   * @name receive:
   */
  receive () {

    let rv = {};

    if (this.is_terminated() && this._message_queue.length <= 0) {
      return promise.reject(
        new exception.runtime(this._error_strings.terminated)
      );
    }

    /* Create promise */
    rv.promise = new promise.base((_resolve_fn, _reject_fn) => {
      rv.reject = _reject_fn;
      rv.resolve = _resolve_fn;
    });

    /* Add to wait queue */
    this._wait_queue.push(rv);

    /* Notify waiters */
    this._notify_wait_queue();

    return rv.promise;
  }

  /**
   * @name terminate:
   */
  terminate () {

    let wait_queue = this._wait_queue;

    if (this.is_running()) {
      try {
        this._worker.terminate();
      } catch (_e) {
        /* Ignore error (node) */
      }
    }

    /* Drain message queue */
    this._notify_wait_queue();

    /* Cancel waiters, avoiding deadlock */
    setTimeout(() => {
      for (var i = 0, len = wait_queue.length; i < len; ++i) {
        return wait_queue[i].reject(
          new exception.runtime(this._error_strings.terminated)
        );
      }
    }, 0);

    /* Reset */
    this._is_running = false;
    this._is_terminated = true;

    return this;
  }

  /** @protected: **/

  /**
   * @name _handle_worker_message:
   */
  _handle_worker_message (_ev: Object) {

    /* istanbul ignore if */
    /* Track termination in browsers */
    if (!utils.is_node() && _ev.data === this._termination_message) {
      return this.terminate();
    }

    this._message_queue.push(_ev.data);
    this._notify_wait_queue();
  }

  /**
   * @name _notify_wait_queue:
   */
  _notify_wait_queue () {

    let len = Math.min(
      this._wait_queue.length,
        this._message_queue.length
    );

    for (var i = 0; i < len; ++i) {

      let r = this._wait_queue.shift();
      let m = this._message_queue.shift();

      setTimeout(() => r.resolve(m), 0);
    }

    return this;
  }

  /** @static: **/

  /**
   * @name main:
   *   Worker thread entry point. Static method: no access to
   *   instance methods or library code unless explicitly loaded.
   */
  main () {

    /* istanbul ignore next */
    throw new Error(
      'Worker thread entry point is unimplemented'
    );
  }
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */

export default Worker;
export { Worker as base };

