/* @flow */

'use strict';

import Base from '../base';
import utils from '../utils';
import Worker from '../worker';
import _ from '../utils/lodash';
import promise from '../promise';
import exception from '../exception';
import proxy_worker from './proxy/worker';

/**
 * @name Proxy:
 */
var Proxy = class extends Base {

  /**
   * @name constructor:
   */
  constructor (_dependencies: Array<string>,
               _controller: Object, _options: ?Object) {

    /* Must be first */
    super(_options);

    this._args = [];
    this._transactions = {};
    this._is_running = false;
    this._global_error = false;
    this._worker = new proxy_worker();

    return this;
  }

  /**
   * @name run:
   */
  run (_dependencies: ?Array<any>, ..._args: Array<any>) {

    this._args = _args;

    /* Ensure we've been subclassed */
    if (!_.isFunction(this.instantiate)) {
      throw new exception.argument(
        'Abstract method; nothing to instantiate'
      );
    }

    /* Set up error reporting */
    this.worker.onerror = (_e) => {
      this._handle_global_error(_e);
    };

    /* Start worker thread */
    this.worker.run(_dependencies);

    /* Initialize worker */
    this.worker.send({
      op: 'initialize', args: this._args,
      instantiation_fn: this.instantiate.toString()
    });

    /* Wait for acknowledgement */
    return this.worker.receive().then((_m) => {

      /* Handle errors */
      if (_m.op !== 'ok') {
        return promise.reject(
          new exception.runtime('Worker initialization failed')
        );
      }

      /* Begin normal operation */
      this._is_running = true;
    });
  }

  /**
   * @name call:
   */
  call (_method: string, ..._args: Array<any>) {

    let rv = {};
    let id = utils.uuid.memory();

    this.worker.send({
      op: 'call', id: id,
      method: _method, args: _args
    });

    /* Promise for this transaction */
    rv.promise = new promise.base((_resolve_fn, _reject_fn) => {
      rv.reject = _reject_fn;
      rv.resolve = _resolve_fn;
    });

    /* Save pending transaction */
    this._transactions[id] = rv;

    /* Schedule another receive */
    this.worker.receive().then((_m) => {
      this._handle_call_response(_m);
    });

    return rv.promise;
  }

  /**
   * @name terminate:
   */
  terminate () {

    this.worker.terminate();
    return this;
  }

  /**
   * @name is_running:
   */
  is_running () {

    /* Just-in-time check */
    if (!this.worker.is_running()) {
      this._is_running = false;
    }

    return this._is_running;
  }

  /**
   * @name get worker:
   */
  get worker (): Object {

    return this._worker;
  }

  /** @protected: **/

  _handle_call_response (_m: Object) {

    let tx = this._transactions[_m.id];

    /* Avoid deadlock */
    setTimeout(() => {

      /* Deliver result */
      if (_m.op === 'ok') {
        tx.resolve(_m.rv);
      } else {
        tx.reject(new exception.runtime('Method call failed', _m.err));
      }

      /* Transaction complete */
      delete this._transactions[_m.id];

    }, 0);

    return this;
  }

  /**
   * @name _handle_global_error:
   */
  _handle_global_error (_e: Object) {

    this._global_error = _e;
    return this.terminate();
  }
};

/**
 * @name Proxy.worker:
 */
Proxy.worker = proxy_worker;

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default Proxy;
export { Proxy as proxy };

