/* @flow */

'use strict';

import Worker from '../../worker';

/**
 * @name ProxyWorker:
 */
var ProxyWorker = class extends Worker {

  /** @static: **/

  /**
   * @name main:
   *   Worker thread entry point.
   */
  main () {

    self.args = [];
    self.transactions = {};
    self.is_initialized = false;

    /**
     * @name invalid_message_error
     */
    let invalid_message_error = {
      op: 'error', id: 'invalid',
      detail: 'Invalid or unsupported message'
    };

    /**
     * @name safe_eval:
     */
    let safe_eval = (_code) => {

      let global_eval = eval;
      return global_eval(_code);
    };

    /**
     * @name is_promise:
     */
    let is_promise = (_v) => {

      return (
        Object(_v) === _v &&
          _v.then instanceof Function &&
            _v.catch instanceof Function
      );
    };

    /**
     * @name handle_instantiation_failure:
     */
    let handle_instantiation_failure = (_e) => {

      self.postMessage({
        op: 'error',
        detail: _e.toString(),
        message: 'Instantiate method failed'
      });

      self.is_initialized = false;
      return self.close();
    }

    /**
     * @name handle_initialization:
     */
    function handle_initialization (_m) {

      let factory_code = _m.instantiation_fn;

      self.args = _m.args;
      self.is_initialized = true;

      try {
        self.instance = safe_eval([
          factory_code,
          'instantiate.apply(',
            'self,', JSON.stringify(self.args),
          ');'
        ].join(''));
      } catch (_e) {
        return handle_instantiation_failure(_e);
      }

      self.postMessage({ op: 'ok', id: _m.id });
    };

    /**
     * @name handle_call:
     */
    let handle_call = (_m) => {

      let rv = self.instance[_m.method];

      if (typeof rv === 'undefined') {
        return self.postMessage({
          op: 'error', id: _m.id,
          detail: 'Requested method is undefined'
        });
      }

      if (rv instanceof Function) {
        rv = rv.apply(self.instance, _m.args);
      }

      let p = self.transactions[_m.id] = (
        is_promise(rv) ? rv : Promise.resolve(rv)
      );

      return p.then((_rv) => {
        self.postMessage({ op: 'ok', id: _m.id, rv: _rv });
      }).catch((_e) => {
        self.postMessage({ op: 'error', id: _m.id, detail: _e.toString() });
      }).then(() => {
        delete self.transactions[_m.id];
      });
    };

    /**
     * @name onmessage:
     */
    self.onmessage = (_ev: Object) => {

      let m: Object = _ev.data;

      /* Type check */
      if (Object(m) !== m) {
        return self.postMessage(invalid_message_error);
      }

      /* Handle initialization */
      if (!self.is_initialized) {
        if (m.op === 'initialize') {
          return handle_initialization(m);
        } else {
          m.op = 'invalid';
        }
      }

      /* Begin operating normally */
      switch (m.op) {
        case 'call':
          return handle_call(m);
        default: case 'invalid':
          invalid_message_error.id = m.id;
          return self.postMessage(invalid_message_error);
      }
    };
  }
};

/**
 * @name exports:
 *   Named and default exports for this module.
 */
export default ProxyWorker;
export { ProxyWorker as proxy_worker };

