
'use strict';

import chai from 'chai';
import _ from './utils/lodash';
import utils from '../../src/utils';
import engine from '../../src/engine';
import schema from '../../src/schema';
import indexer from '../../src/indexer';
import promise from '../../src/promise';
import promises from 'chai-as-promised';
import test_utils from '../include/utils';
import proxy from '../../src/worker/proxy';
import fixtures from '../include/fixtures';
import exception from '../../src/exception';

/**
 * @name engine_proxy:
 *   A Proxy subclass that instantiates an Engine object.
 */
let engine_proxy = class extends proxy {

  /** @static: **/
  instantiate (_options) {

    return new self.api.engine(_options);
  }
};

/**
 * @name engine:
 *   Unit tests for the engine module.
 */
describe('engine', () => {

  let should = chai.should();
  chai.use(promises);

  let engine_options = {
    type: 'pouchdb',
    path: 'output/databases/unit/engine'
  };

  let base_dir = test_utils.base_directory();

  let dependencies = [
    `${base_dir}/node_modules/babel-polyfill/dist/polyfill.js`,
    `${base_dir}/dist/pouchdb.min.js`,
    `${base_dir}/dist/api.js`
  ];

  /* Engine */
  let e = null;

  /**
   * @name before:
   *   Destroy and compact the database to release storage.
   */
  before(() => {
    e = new engine(engine_options);
  });

  /**
   * @name after:
   *   Destroy and compact the database to release storage.
   */
  after(() => {
    return e.destroy().catch((_e) => {
      /* Ignore error */
    });
  });

  /**
   * @name should instansiate with no options:
   */
  it('should instansiate with no options', () => {

    e.should.be.an.instanceof(engine);
  });

  /**
   * @name should have a type:
   */
  it('should have a type', () => {

    return e.type.should.equal(engine_options.type);
  });

  /**
   * @name should start properly:
   */
  it('should start properly', () => {

    return e.start().then(() => {
      return e.schema().version();
    }).then((_version) => {
      _version.should.be.above(0);
      e.is_running.should.be.true;
    });
  });

  /**
   * @name should expose a working schema:
   */
  it('should expose a working schema', () => {

    should.exist(e.schema());
    e.schema().should.be.an.instanceof(schema.base);
  });

  /**
   * @name should expose a working indexer:
   */
  it('should expose a working indexer', () => {

    should.exist(e.indexer());
    e.indexer().should.be.an.instanceof(indexer.base);
  });

  /**
   * @name stores and indexes a simple node:
   */
  it('stores and indexes a simple node', () => {

    let n = {
      title: 'Test',
      content: 'This is a test'
    };

    return e.store_node(n).then((_node) => {
      _node.type.should.equal('node');
      _node.title.should.equal(n.title);
      _node.content.should.equal(n.content);
      _node.tags.should.be.an.array;
      _node.tags.should.have.length.of(0);
      return e.index_node(_node);
    }).then((_r) => {
      _r.node.type.should.equal('node');
      _r.node.title.should.equal(n.title);
      _r.node.content.should.equal(n.content);
      _r.node.tags.should.be.an.array;
      _r.node.tags.should.have.length.of(0);
      _r.index.type.should.equal('index');
      _r.index.data.content.should.be.a.string;
      _r.index.data.words.should.be.an.array;
      _r.index.data.words.should.have.lengthOf(4);

      return promise.all([
        e.remove_node(_r.node), e.remove_index(_r.index)
      ]);
    });
  });

  /**
   * @name stores and indexes a node with tags:
   */
  it('stores and indexes a node with tags', () => {

    let n = {
      title: 'Another test',
      content: 'This is another test',
      tags: [ 'test', 'testing', 'another test' ]
    };

    return e.store_node(n).then((_node) => {
      _node.title.should.equal(n.title);
      _node.content.should.equal(n.content);
      _node.tags.should.be.an.array;
      _node.tags.should.have.length.of(3);
      return e.index_node(_node);
    }).then((_r) => {
      _r.index.data.tags.should.be.an.array;
      _r.index.node_id.should.equal(_r.node._id);

      for (let i = 0; i < n.tags.length; ++i) {
        _r.index.data.tags[i][0].should.equal(n.tags[i]);
      }
      return promise.all([
        e.remove_node(_r.node), e.remove_index(_r.index)
      ]);
    });
  });

  /**
   * @name rejects a title-free node:
   */
  it('rejects a title-free node', () => {

    e.store_node({ content: 'A' }).should.eventually.be.rejectedWith(
      exception.argument
    );
  });

  /**
   * @name rejects a content-free node:
   */
  it('rejects a content-free node', () => {

    e.store_node({ title: 'B' }).should.eventually.be.rejectedWith(
      exception.argument
    );
  });

  /**
   * @name allows a numeric zone identifier:
   */
  it('allows a numeric zone identifier', () => {

    let n = {
      zone_id: 123,
      title: 'Test',
      content: 'Hello, world!'
    };

    return e.store_node(n).then((_r) => {
      return e.remove_node(_r);
    });
  });

  /**
   * @name index linking:
   */
  describe('index linking', () => {

    /**
     * @name simple:
     */
    describe('simple', () => {

      let n1, n2, i1, i2;
      let t1 = 'This is an example';
      let t2 = fixtures.all().unit.text['link.txt'];

      /**
       * @name beforeEach:
       *   Add, index, and link multiple fixtures manually.
       */
      beforeEach(() => {

        n1 = {
          type: 'node', title: 'Test Node #1',
          tags: [ 'should link' ], content: t1
        };

        n2 = {
          type: 'node', title: 'Test Node #2',
          tags: [ 'example' ], content: t2
        };

        /* Manually store, index, and link nodes */
        return e.store_node(n1).then((_node) => {
          n1 = _node;
          return e.index_node(n1);
        }).then((_rv) => {
          n1 = _rv.node;
          i1 = _rv.index;
          return e.store_node(n2);
        }).then((_node) => {
          n2 = _node;
          return e.index_node(n2);
        }).then((_rv) => {
          n2 = _rv.node;
          i2 = _rv.index;
          return e.link_index(i2);
        }).then((_index) => {
          i2 = _index;
        });
      });

      /**
       * @name afterEach:
       *   Remove all fixtures and indexes added in `before`.
       */
      afterEach(() => {

        return promise.all([
          e.remove_node(n1), e.remove_node(n2),
          e.remove_index(i1), e.remove_index(i2)
        ]);
      });

      /**
       * @name links a simple set of documents:
       */
      it('links a simple set of documents', () => {

        i2.links.should.deep.equal([
          [ 0, n1.index_id, 11, 7, n2._id,
            [ [ "example", 11, 7, 3 ] ] ],
          [ 0, n2.index_id, 0, 11, n1._id,
            [ [ "should", 0, 6, 0 ], [ "link", 7, 4, 1 ] ] ],
          [ 0, n2.index_id, 43, 11, n1._id,
            [ [ "should", 43, 6, 6 ], [ "link", 50, 4, 7 ] ] ],
          [ 0, n2.index_id, 101, 11, n1._id,
            [ [ "should", 101, 6, 15 ], [ "link", 108, 4, 16 ] ] ],
          [ 0, n2.index_id, 160, 12, n1._id,
            [ [ "should", 160, 6, 24 ], [ "link", 168, 4, 25 ] ] ]
        ]);

        return e.schema().view('links', { group: 1 }).then((_v) => {
          _v.rows.should.deep.equal([
            { value: 1, key: [ 0, n1.index_id, 11, 7, n2._id  ] },
            { value: 1, key: [ 0, n2.index_id, 0, 11, n1._id ] },
            { value: 1, key: [ 0, n2.index_id, 43, 11, n1._id ] },
            { value: 1, key: [ 0, n2.index_id, 101, 11, n1._id ] },
            { value: 1, key: [ 0, n2.index_id, 160, 12, n1._id ] }
          ]);
          return e.link_index(i1);
        }).then((_index) => {
          i1 = _index;
          return e.schema().view('links', { group: true });
        }).then((_v) => {
          _v.rows.should.deep.equal([
            { value: 2, key: [ 0, n1.index_id, 11, 7, n2._id ] },
            { value: 2, key: [ 0, n2.index_id, 0, 11, n1._id ] },
            { value: 2, key: [ 0, n2.index_id, 43, 11, n1._id ] },
            { value: 2, key: [ 0, n2.index_id, 101, 11, n1._id ] },
            { value: 2, key: [ 0, n2.index_id, 160, 12, n1._id ] }
          ]);
        });
      });

      /**
       * @name renders strings:
       */
      it('renders strings', () => {

        let link_fn = (_s) => `[${_s}]`;

        return e.render(n1, _.identity, link_fn).then((_rv) => {
          _rv.should.equal(t1.replace(/(example)/gi, '[$1]'));
          return e.render(n2, _.identity, link_fn);
        }).then((_rv) => {
          _rv.should.equal(t2.replace(/(should\s+link)/gi, '[$1]'));
        });
      });

      /**
       * @name renders links with multiple targets:
       */
      it('renders links with multiple targets', () => {

        let link_fn = (_s, _target_node_ids) => {
          return `[${_s}|${_target_node_ids.join(',')}]`;
        }

        let n3 = {
          type: 'node', title: 'Test Node #3',
          tags: [ 'example' ], content: 'This is an example.'
        };

        let n4 = {
          type: 'node', title: 'Test Node #4',
          tags: [ 'example' ], content: '¡Example!'
        };

        return e.create_node(n3).then((_rv) => {
          n3 = _rv.node;
          return e.render_node(n3, 'text');
        }).then((_s) => {
          _s.should.equal(`This is an [example|${n2._id},${n3._id}].`);
          return e.create_node(n4);
        }).then((_rv) => {
          n4 = _rv.node;
          return e.render(n4, _.identity, link_fn);
        }).then((_s) => {
          _s.should.equal(`¡[Example|${n2._id},${n3._id},${n4._id}]!`);
        }).then(() => {
          return promise.all([
            e.delete_node(n3), e.delete_node(n4)
          ]);
        });
      });
    });

    /**
     * @name synchronous:
     */
    describe('synchronous', () => {

      /**
       * @name renders overlapping links:
       */
      it('renders overlapping links', () => {

        let n1 = {
          title: 'Test Node #1',
          content: 'This is a test.',
          tags: [ 'dank memes', 'post dank memes', 'memes' ]
        };

        let n2 = {
          title: 'Test Node #2',
          content: 'Post dank memes.',
          tags: [ 'this', 'this is', 'this is a', 'this is a test' ]
        };

        return e.create_node(n1).then((_rv) => {
          n1 = _rv.node;
          return e.create_node(n2);
        }).then((_rv) => {
          n2 = _rv.node;
          return e.render_node(n1, 'text');
        }).then((_s) => {
          _s.should.equal(`[This|${n2._id}] is a test.`);
          return e.render_node(n2, 'text');
        }).then((_s) => {
          _s.should.equal(`[Post dank memes|${n1._id}].`);
          return promise.all([ e.delete_node(n1), e.delete_node(n2) ]);
        });
      });

      /**
       * @name renders links using different result limits:
       */
      it('renders links using different result limits', () => {

        const n = [ 1, 2, 3, 4, 5 ];

        let offset = 0
        let limits = [];
        let t1 = fixtures.all().unit.text['link.txt'];
        let t2 = fixtures.all().unit.text['mantra.txt'];

        /* Cartesian product */
        for (var i = 0, len = n.length; i < len; ++i) {
          for (var j = 0; j < len; ++j) {
            limits.push({
              view: { limit: n[i] }, join: { limit: n[j] }
            });
          }
        }

        /* For every combination of limits... */
        return promise.while(() => offset < limits.length, () => {

          let options = { iterator: limits[offset] };

          let n1 = {
            title: 'Test Node #1',
            tags: [ 'example one', 'example two', 'by will alone' ],
            content: [ 'Post memes. Post dank memes.', t1 ].join(' ')
          };

          let n2 = {
            title: 'Test Node #2',
            tags: [ 'post memes', 'post memes', 'post dank memes' ],
            content: [ 'Example one and example two.', t2 ].join(' ')
          };

          return e.create_node(n1, options).then((_rv) => {
            n1 = _rv.node;
            return e.create_node(n2, options);
          }).then((_rv) => {
            n2 = _rv.node;
            return e.render_node(n1, 'text');
          }).then((_s) => {
            _s.should.include([
              `[Post memes|${n2._id}].`,
              `[Post dank memes|${n2._id}].`
            ].join(' '));
            return e.render_node(n2, 'text');
          }).then((_s) => {
            _s.should.include([
              `[Example one|${n1._id}] and [example two|${n1._id}].`,
              `It is [by will alone|${n1._id}] I set my mind in motion.`
            ].join(' '));
          }).then(() => {
            offset++;
            return promise.all([ e.delete_node(n1), e.delete_node(n2) ]);
          });
        });
      });

    });

    /**
     * @name threaded:
     */
    describe('threaded', () => {

      /**
       * @name renders in parallel:
       */
      it('renders in parallel', () => {

        if (utils.is_node()) {
          return;
        }

        let n = [{
          type: 'node', title: 'Test Node #0',
          tags: [ 'example', 'an example' ], content: 'Hello world!'
        }, {
          type: 'node', title: 'Test Node #1',
          tags: [ 'world', 'earth' ], content: 'Example: an example'
        }];

        let n2 = {
          type: 'node', title: 'Test Node #2',
          tags: [ 'a different', 'example' ], content: 'Hello world.'
        };

        /* Use a web worker */
        let p = new engine_proxy();

        return p.run(dependencies, engine_options).then(() => {
          p.is_running().should.be.true;
          return p.call('start');
        }).then(() => {
          return p.call('create_node', n2);
        }).then((_rv) => {
          n2 = _rv.node;
          return p.call('render_node', n2, 'text');
        }).then((_s) => {
          _s.should.equal('Hello world.');
          return promise.all(
            _.map(n, (_n) => p.call('create_node', _n))
          );
        }).then((_rv) => {
          n = _.map(_rv, (_r) => _r.node);
          return p.call('render_node', n2, 'text');
        }).then((_s) => {
          _s.should.equal(`Hello [world|${n[1]._id}].`);
          return p.call('render_node', n[0], 'text');
        }).then((_s) => {
          _s.should.equal(`Hello [world|${n[1]._id}]!`);
          return p.call('render_node', n[1], 'text');
        }).then((_s) => {
          _s.should.equal(
            `[Example|${n2._id},${n[0]._id}]: [an example|${n[0]._id}]`
          );
        }).then(() => {
          return promise.all(_.map(n.concat(n2), (_n) => {
            return p.call('delete_node', _n);
          }));
        });
      });
    });
  });

  /**
   * @name node creation:
   */
  describe('node creation', () => {

    /**
     * @name creates a single trivial node:
     */
    it('creates a single trivial node', () => {

      let n = {
        title: 'Test',
        content: 'Hello, world!'
      };

      return e.create_node(n).then((_rv) => {
        _rv.node.type.should.equal('node');
        _rv.index.type.should.equal('index');
        _rv.node._id.should.equal(_rv.index.node_id);
        _rv.index._id.should.equal(_rv.node.index_id);

        _rv.node.zone_id.should.equal(0);
        _rv.node.title.should.equal('Test');
        _rv.node.tags.should.deep.equal([]);
        _rv.node.content.should.equal('Hello, world!');
      });
    });

    /**
     * @name rejects invalid nodes:
     */
    it('rejects invalid nodes', () => {

      e.store_node().should.be.rejectedWith(exception.argument);
      e.store_node([]).should.be.rejectedWith(exception.argument);
      e.store_node({}).should.be.rejectedWith(exception.argument);
      e.store_node(false).should.be.rejectedWith(exception.argument);

      e.create_node().should.be.rejectedWith(exception.argument);
      e.create_node([]).should.be.rejectedWith(exception.argument);
      e.create_node({}).should.be.rejectedWith(exception.argument);
      e.create_node(false).should.be.rejectedWith(exception.argument);

      e.store_node({ content: 'A' }).should.be.rejectedWith(
        exception.argument
      );

      e.create_node({ content: 'A' }).should.be.rejectedWith(
        exception.argument
      );

      e.store_node({ title: 'B' }).should.be.rejectedWith(
        exception.argument
      );

      e.create_node({ title: 'B' }).should.be.rejectedWith(
        exception.argument
      );

      e.store_node({ extra: 'C' }).should.be.rejectedWith(
        exception.argument
      );

      e.create_node({ extra: 'C' }).should.be.rejectedWith(
        exception.argument
      );
    });
  });
});

