
'use strict';

import chai from 'chai';
import events from '../../src/events';
import exception from '../../src/exception';

/**
 * @name events:
 *   Unit tests for the events module.
 */
describe('events', () => {

  let should = chai.should();

  /**
   * @name can be constructed:
   */
  it('can be constructed', () => {

    (() => { new events(); }).should.not.throw();
    (() => { new events({}); }).should.not.throw();
  });

  /**
   * @name can register a new event type:
   */
  it('can register a new event type', () => {

    let e = new events();

    (() => { e.register('t1'); }).should.not.throw();
    (() => { e.on('t1', () => {}); }).should.not.throw();
    (() => { e.unregister('t1'); }).should.not.throw();
  });

  /**
   * @name throws on unregistered event types:
   */
  it('throws on unregistered event types', () => {

    let e = new events();

    (() => { e.send('t1'); }).should.throw();
    (() => { e.unregister('t1'); }).should.throw();
    (() => { e.on('t1', () => {}); }).should.throw();
  });

  /**
   * @name emits an event correctly:
   */
  it('emits an event correctly', () => {

    let pass = false;
    let e = new events();

    (() => { e.register('t1'); }).should.not.throw();
    (() => { e.register('t1'); }).should.throw();

    (() => {
      e.on('t1', (_v) => (_v + _v));
    }).should.not.throw();

    (() => {
      e.on('t1', (_v) => ((pass = _v) * _v));
    }).should.not.throw();

    (() => {
      e.emit('t1', false, [ 3 ]).should.deep.equal([ 6, 9 ]);
    }).should.not.throw();

    pass.should.equal(3);

    (() => { e.unregister('t1'); }).should.not.throw();
    (() => { e.unregister('t1'); }).should.throw();
  });

  /**
   * @name emits a singleton event correctly:
   */
  it('emits a singleton event correctly', () => {

    let pass = false;
    let e = new events();

    (() => { e.register('t1', true); }).should.not.throw();

    (() => {
      e.on('t1', (_v) => ((pass = _v) + _v));
    }).should.not.throw();

    (() => {
      e.on('t1', (_v) => _v);
    }).should.throw();

    (() => {
      e.send('t1', false, 3).should.equal(6);
    }).should.not.throw();

    pass.should.equal(3);

    (() => { e.unregister('t1'); }).should.not.throw();
    (() => { e.unregister('t1'); }).should.throw();
  });

  /**
   * @name validates listener object on remove:
   */
  it('validates listener object on remove', () => {

    let e = new events();

    (() => { e.register('t1', true); }).should.not.throw();
    (() => { e.remove('t1', 1); }).should.throw(exception.argument);
    (() => { e.remove('t1', {}); }).should.throw(exception.argument);
  });
});

