
'use strict';

import chai from 'chai';
import indexer from '../../src/indexer';
import exception from '../../src/exception';

/**
 * @name indexer:
 *   Unit tests for the text indexer.
 */
describe('indexer', () => {

  let should = chai.should();

  /**
   * @name includes content:
   */
  it('includes content', () => {

    let i = new indexer.text();

    let t = 'This is a test.';
    let r = i.build(t).result;

    r.words.length.should.equal(4);
    r.tags.length.should.equal(0);
    r.content.should.be.a.string;
    r.content.should.equal(t);
  });

  /**
   * @name creates ASCII word lists correctly:
   */
  it('creates ASCII word lists correctly', () => {

    let i = new indexer.text();

    let r = i.build('This is a test.', {
      tags: [ 'Important', 'But not really important.', 'High priority!' ]
    }).result;

    r.words.length.should.equal(4);
    r.words[1][0].should.equal('is');

    r.tags.length.should.equal(3);

    r.tags[0][1][0][0].should.equal('important');
    r.tags[1][1][0][0].should.equal('but');
    r.tags[1][1][3][0].should.equal('important');
    r.tags[2][1][0][0].should.equal('high');
    r.tags[2][1][1][0].should.equal('priority');

    r.tags[0][1].length.should.equal(1);
    r.tags[1][1].length.should.equal(4);
    r.tags[2][1].length.should.equal(2);
  });

  /**
   * @name rejects non-string values:
   */
  it('rejects non-string values', () => {

    let i = new indexer.text();

    (() => { i.build(undefined); }).should.throw();
    (() => { i.build(1); }).should.throw(exception.argument);
    (() => { i.build(true); }).should.throw(exception.argument);
    (() => { i.clear().build(undefined); }).should.throw();
    (() => { i.clear().string = 'a'; i.build(); }).should.not.throw();
  });

  /**
   * @name retains a string correctly:
   */
  it('retains a string correctly', () => {

    let i = new indexer.text();

    i.build('dank');
    i.string.should.equal('dank');
    i.clear().string = 'not dank';
    i.string.should.equal('not dank');
  });

  /**
   * @name rejects invalid tag lists:
   */
  it('rejects invalid tag lists', () => {

    let i = new indexer.text();

    (() => { i.build('', { tags: [] }); }).should.not.throw();
    (() => { i.build('', { tags: 1 }); }).should.throw(exception.argument);
    (() => { i.build('', { tags: {} }); }).should.throw(exception.argument);
  });

  /**
   * @name rejects improperly-formatted tags:
   */
  it('rejects improperly-formatted tags', () => {

    let i = new indexer.text();

    (() => { i.build('', { tags: [ 1 ] }); }).should.throw();
    (() => { i.build('', { tags: [ false ] }); }).should.throw();
    (() => { i.build('', { tags: [ undefined ] }); }).should.throw();
    (() => { i.build('', { tags: [ '', '', '' ] }); }).should.not.throw();
  });

  /**
   * @name sorts tags correctly:
   */
  it('sorts tags correctly', () => {

    let i = new indexer.text();

    let r = i.build('This is a test.', {
      sort: true,
      tags: [ 'Important', 'But not really important.', 'High priority!' ]
    }).result;

    r.tags[0][1][0][0].should.equal('but');
    r.tags[1][1][0][0].should.equal('high');
    r.tags[2][1][0][0].should.equal('important');
  });

  /**
   * @name sorts tags correctly with a custom collator:
   */
  it('sorts tags correctly with a custom collator', () => {

    let i = new indexer.text();

    let r = i.build('This is a test.', {
      sort: true,
      tags: [ 'Important', 'But not really important.', 'High priority!' ],
      collator: {
        collate: (_l, _r) => {
          return (_l == _r ? 0 : (_l < _r ? 1 : -1));
        }
      },
    }).result;

    r.tags[0][1][0][0].should.equal('important');
    r.tags[1][1][0][0].should.equal('high');
    r.tags[2][1][0][0].should.equal('but');
  });

  /**
   * @name accepts input via its constructor:
   */
  it('accepts input via its constructor', () => {

    let i = new indexer.text('memes');
    let r = i.build().result;

    r.should.be.an.object;
    r.tags.should.be.empty;
    r.tags.should.be.an.instanceof(Array);
    r.words.length.should.equal(1);
  });
});

