
'use strict';

import chai from 'chai';
import _ from './utils/lodash';
import promise from '../../src/promise';
import iterator from '../../src/iterator';
import exception from '../../src/exception';

/**
 * @name iterator:
 *   Unit tests for the asynchronous iterator.
 */
describe('iterator', () => {

  /**
   * @name invalid_find_subclass:
   */
  let invalid_find_subclass = class extends iterator.base {

    find () {
      return 'Invalid offset';
    }

    populate () {
      return new promise.base((_resolve_fn) => {
        _.defer(() => {
          this.buffer = [ 1, 2, 3 ];
          return _resolve_fn();
        });
      });
    }
  };

  /**
   * @name rejecting_subclass:
   */
  let rejecting_subclass = class extends iterator.base {

    populate () {
      return promise.reject(
        new exception.unexpected('Intentionally rejected')
      );
    }
  };

  let should = chai.should();

  let a = new iterator.base();
  let r = new rejecting_subclass();
  let f = new invalid_find_subclass();

  let b = new iterator.base({
    autobegin: false
  });

  /**
   * @name base:
   */
  describe('base', () => {

    /**
     * @name initializes correctly:
     */
    it('initializes correctly', () => {

      should.not.exist(b.key);
      b.is_eof().should.equal.false;
      b.offset.should.equal(0);
      b.buffer.length.should.equal(0);
    });

    /**
     * @name does not throw when autobegin is on:
     */
    it('does not throw when autobegin is on', () => {

      should.not.exist(a.key);
      should.not.exist(a.current());

      (() => { a.next(); }).should.not.throw();
      (() => { a.previous(); }).should.not.throw();
      (() => { a.seek(true); }).should.not.throw();
      (() => { a.rewind(true); }).should.not.throw();
    });

    /**
     * @name throws when autobegin is off:
     */
    it('throws when autobegin is off', () => {

      should.not.exist(b.key);
      should.not.exist(b.current());

      (() => { b.next(); }).should.throw(exception.argument);
      (() => { b.previous(); }).should.throw(exception.argument);
      (() => { b.seek(true); }).should.throw(exception.argument);
      (() => { b.rewind(true); }).should.throw(exception.argument);
    });

    /**
     * @name is not ordered:
     */
    it('is not ordered', () => {

      b.is_ordered().should.be.false;
    });

    /**
     * @name has no current key or last key:
     */
    it('has no current key or last key', () => {

      should.not.exist(b.last_key());
      should.not.exist(b.current_key());
    });

    /**
     * @name seeks to beginning:
     */
    it('seeks to beginning', () => {

      return b.begin().then((_r) => {
        should.not.exist(_r);
        b.is_eof().should.equal.true;
      });
    });

    /**
     * @name clears eof after reset:
     */
    it('clears eof after reset', () => {

      b.reset();
      b.is_eof().should.equal.false;
    });

    /**
     * @name seeks to a non-existent value:
     */
    it('seeks to a non-existent value', () => {

      return b.begin().then((_r) => {
        return b.seek('abc');
      }).then((_r) => {
        should.not.exist(_r);
        b.is_eof().should.equal.true;
      });
    });

    /**
     * @name refuses to drain an empty buffer:
     */
    it('refuses to drain an empty buffer', () => {

      b.drain().should.equal.false;
    });

    /**
     * @name handles invalid results from find:
     */
    it('handles invalid results from find', () => {

      f.begin().then(() => {
        f.seek(3).should.eventually.be.rejectedWith(
          exception.argument
        );
      });
    });

    /**
     * @name rejects invalid populate functions:
     */
    it('rejects invalid populate functions', () => {

      (() => {
        f.events.on('populate', 'dank');
      }).should.throw(exception.argument);
    });

    /**
     * @name rejects invalid filter functions:
     */
    it('rejects invalid filter functions', () => {

      (() => {
        f.events.on('filter', 'memes');
      }).should.throw(exception.argument);
    });

    /**
     * @name rejects invalid transform functions:
     */
    it('rejects invalid transform functions', () => {

      (() => {
        f.events.on('transform', 'dank');
      }).should.throw(exception.argument);
    });

    /**
     * @name disallows overlapping seek operations:
     */
    it('disallows overlapping seek operations', () => {

      f.begin(); /* Discard promise */
      f.seek().should.eventually.throw(exception.reference);
    });

    /**
     * @name throws if populate throws:
     */
    it('throws if populate throws', () => {

      r.begin().should.eventually.throw(exception.unexpected);
    });

    /**
     * @name initially has no event listeners:
     */
    it('initially has no event listeners', () => {

      let events = {
        filter: 1, populate: 2, transform: 3
      };

      _.each(events, (_n, _type) => {

        (() => {
          b.events.listeners(_type).should.equal(0);
        }).should.not.throw();

        (() => {
          should.exist(b.events.get(_type));
          b.events.get(_type).should.be.an('object');
        }).should.not.throw();
      });
    });
  });
});

