
'use strict';

import chai from 'chai';
import _ from '../utils/lodash';
import promise from '../../../src/promise';
import iterator from '../../../src/iterator';
import join from '../../../src/iterator/join';
import exception from '../../../src/exception';

/**
 * @name iterator:
 *   Unit tests for the asynchronous iterator.
 */
describe('iterator', () => {

  let should = chai.should();

  const numbers = [
    10, 20, 30, 40, 50
  ];

  /**
   * @name array:
   */
  describe('array', () => {

    /**
     * @name iteration:
     */
    describe('iteration', () => {

      /**
       * @name returns current/next:
       */
      it('returns current/next', () => {

        let i = new iterator.array(numbers, {
          repeat: 10
        });

        return i.begin().then((_n) => {
          _n.should.equal(10);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.next();
        }).then((_n) => {
          _n.should.equal(20);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
        });
      });

      /**
       * @name throws when autobegin is disabled:
       */
      it('throws when autobegin is disabled', () => {

        let r = [];

        let i = new iterator.array(numbers, {
          autobegin: false
        });

        (() => { i.next(); }).should.throw(exception.argument);
        (() => { i.seek(10); }).should.throw(exception.argument);
      });

      /**
       * @name works without calling begin:
       */
      it('works without calling begin', () => {

        let r = [];

        let i = new iterator.array(numbers, {
          ordered: true
        });

        i.all().then((_r) => {
          r.should.deep.equal(numbers);
        });
      });

      /**
       * @name iterates properly in ordered mode:
       */
      it('iterates properly in ordered mode', () => {

        let r = [];

        let i = new iterator.array(numbers, {
          ordered: true
        });

        return i.begin().then((_n) => {
          return promise.while(() => !i.is_eof(), () => {
            r.push(i.current());
            return i.next();
          });
        }).then(() => {
          r.should.deep.equal(numbers);
        });
      });

      /**
       * @name sets eof for nonexistent items:
       */
      it('sets eof for nonexistent items', () => {

        let i = new iterator.array(numbers, {
          ordered: true
        });

        return i.begin().then((_n) => {
          _n.should.equal(10);
          i.is_eof().should.be.false;
          i.current().should.equal(_n);
          return i.seek(60);
        }).then((_n) => {
          should.not.exist(_n);
          i.is_eof().should.be.true;
          should.not.exist(i.current());
        });

        /**
         * @name reads identically forward and backward:
         */
        it('reads identically forward and backward', () => {

          let forward = [], backward = [];

          let i = new iterator.array(numbers, {
            ordered: true
          });

          return i.all().then((_all) => {
            forward = _all;
            return i.previous();
          }).then(() => {
            return promise.while(() => !i.is_bof(), () => {
              backward.unshift(i.current());
              return i.previous();
            });
          }).then(() => {
            backward.should.deep.equal(forward);
            offset++;
          });
        });
      });

      /**
       * @name iterates over an empty array using each:
       */
      it('iterates over an empty array using each', () => {

        let rv = true;
        let i = new iterator.array([]);

        return i.begin().then((_n) => {

          should.not.exist(_n);
          i.is_eof().should.be.true;
          should.not.exist(i.current());

          return i.each((_n) => {
            rv = false; /* Should not be reached */
          }).then(() => {
            rv.should.be.true;
          });;
        });
      });

      /**
       * @name iterates over a simple array using each:
       */
      it('iterates over a simple array using each', () => {

        let rv = [];
        let i = new iterator.array(numbers);

        return i.begin().then((_n) => {

          should.exist(_n);
          should.exist(i.current());
          i.is_eof().should.be.false;

          return i.each((_n) => {
            rv.push(_n);
          }).then(() => {
            rv.should.deep.equal([ 10, 20, 30, 40, 50 ]);
          });;
        });
      });

    });

    /**
     * @name repeat:
     */
    describe('repeat', () => {

      /**
       * @name works sequentially in ordered/unordered modes:
       */
      it('works sequentially in ordered/unordered modes', () => {

        let i = new iterator.array(numbers, {
          repeat: 10
        });

        let is = new iterator.array(numbers, {
          repeat: 2, ordered: true
        });

        /* Test function */
        let fn = (_it) => {

          /* Map: item => count */
          let visited = {};

          return _it.begin().then((_n) => {

            /* Consume everything */
            return promise.while(() => !_it.is_eof(), () => {
              let n = _it.current();
              visited[n] = (visited[n] || 0) + 1;
              return _it.next();
            });

          }).then(() => {

            /* Ensure counts are correct */
            _it.is_eof().should.be.true;
            _.keys(visited).length.should.equal(numbers.length);

            _.each(visited, (_v, _k) => {
              _v.should.equal(_it.options.repeat);
              numbers.indexOf(_.parseInt(_k)).should.be.at.least(0);
            });
          });
        };

        /* Test two different iterator types */
        return promise.all([ fn(i), fn(is) ]);
      });

      /**
       * @name can continue indefinitely:
       */
      it('can continue indefinitely', () => {

        let n = 0;

        let i = new iterator.array(numbers, {
          repeat: true
        });

        return i.begin().then(() => {
          return promise.while(() => (n++ < 1000), () => {
            i.current().should.be.an.integer;
            return i.next();
          });
        });
      });
    });

    /**
     * @name seek:
     */
    describe('seek', () => {

      /**
       * @name seeks in ordered/unordered modes:
       */
      it('seeks in ordered/unordered modes', () => {

        let test_fn = (_is_ordered) => {
          let i = new iterator.array(numbers, {
            repeat: 2, ordered: _is_ordered
          });

          return i.begin().then((_n) => {
            i.is_eof().should.be.false;
            _n.should.equal(10);
            i.current().should.equal(_n);
            return i.seek(50);
          }).then((_n) => {
            i.is_eof().should.be.false;
            _n.should.equal(50);
            i.current().should.equal(_n);
            return i.seek(20);
          }).then((_n) => {
            i.is_eof().should.be.false;
            _n.should.equal(20);
            i.current().should.equal(_n);
            return i.seek(10);
          }).then((_n) => {
            i.is_eof().should.be.true;
            should.not.exist(_n);
            should.not.exist(i.current());
          });
        };

        return promise.all([ test_fn(true), test_fn(false) ]);
      });

      /**
       * @name seeks using inexact matching:
       */
      it('seeks using inexact matching', () => {

        let i = new iterator.array(numbers, {
          repeat: 10
        });

        let ir = new iterator.array(numbers, {
          repeat: true
        });

        let is = new iterator.array(numbers, {
          repeat: 2, ordered: true
        });

        /* Test function */
        let fn = (_it) => {
          return _it.begin().then(() => {
            return _it.seek(20);
          }).then((_n) => {
            _it.is_eof().should.be.false;
            _n.should.equal(20);
            _it.current().should.equal(20);
            return _it.seek(25, true);
          }).then((_n) => {
            _it.is_eof().should.be.false;
            _n.should.equal(30);
            _it.current().should.equal(30);
            return _it.seek(49, true);
          }).then((_n) => {
            _it.is_eof().should.be.false;
            _n.should.equal(50);
            _it.current().should.equal(50);
            return _it.seek(51, true);
          }).then((_n) => {
            /* Improper ordering treated as EOF */
            should.not.exist(_n);
            should.not.exist(i.current());
            _it.is_eof().should.be.true;
          });
        };

        return promise.all(
          [ fn(i), fn(ir), fn(is) ]
        );
      });

      /**
       * @name seeks properly with no-op filter:
       */
      it('seeks properly with no-op filter', () => {

        let i = new iterator.array(numbers, {
          repeat: true
        });

        let ev = i.events.on('filter', (_r) => {
          return true;
        });

        return i.begin().then((_n) => {
          _n.should.equal(10);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(30);
        }).then((_n) => {
          _n.should.equal(30);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(50);
        }).then((_n) => {
          _n.should.equal(50);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(31, true);
        }).then((_n) => {
          _n.should.equal(40);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          i.events.remove('filter', ev);
        });
      });

      /**
       * @name seeks to the first of duplicate keys:
       */
      it('seeks to the first of duplicate keys', () => {

        const duplicate_numbers = [
          10,
          20, 20,
          30, 30, 30,
          40, 40, 40, 40,
          50, 50, 50, 50, 50
        ];

        let fn = (_is_ordered) => {
          let i = new iterator.array(duplicate_numbers, {
            repeat: false, ordered: _is_ordered
          });

          return i.begin().then((_n) => {
            return i.seek(30);
          }).then((_n) => {
            _n.should.equal(30);
            i.offset.should.equal(3);
            i.current().should.equal(_n);
            i.is_eof().should.be.false;
            return i.seek(50);
          }).then((_n) => {
            _n.should.equal(50);
            i.offset.should.equal(10);
            i.current().should.equal(_n);
            i.is_eof().should.be.false;
          });
        };

        return promise.all([ fn(true), fn(false) ]);
      });

      /**
       * @name seeks with a key mapping function:
       */
      it('seeks with a key mapping function', () => {

        const number_objects = [
          { n: 10, text: 'ten' },
          { n: 20, text: 'twenty' },
          { n: 30, text: 'thirty' },
          { n: 40, text: 'forty' },
          { n: 50, text: 'fifty' }
        ];

        let fn = (_is_ordered) => {
          let i = new iterator.array(number_objects, {
            repeat: false,
            ordered: _is_ordered
          });

          i.events.on(
            'key.map', (_r) => _r.n
          );

          return i.begin().then((_n) => {
            return i.seek(30);
          }).then((_r) => {
            _r.n.should.equal(30);
            _r.text.should.equal('thirty');
            i.offset.should.equal(2);
            i.current().should.deep.equal(_r);
            i.is_eof().should.be.false;
            return i.seek(50);
          }).then((_r) => {
            _r.n.should.equal(50);
            _r.text.should.equal('fifty');
            i.offset.should.equal(4);
            i.is_eof().should.be.false;
            i.current().should.deep.equal(_r);
            return i.next();
          }).then((_r) => {
            should.not.exist(_r);
            i.is_eof().should.be.true;
          });
        };

        return promise.all([ fn(true), fn(false) ]);
      });
    });

    /**
     * @name populate:
     */
    describe('populate', () => {

      /**
       * @name populates properly via next:
       */
      it('populates properly via next', () => {

        let i = new iterator.array(numbers, {
          repeat: 10
        });

        return i.begin().then((_n) => {
          _n.should.equal(10);
          i.current().should.equal(_n);
          return i.seek(50);
        }).then((_n) => {
          _n.should.equal(50);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.next();
        }).then((_n) => {
          _n.should.equal(10);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
        });
      });

      /**
       * @name populates properly via seek:
       */
      it('populates properly via seek', () => {

        let i = new iterator.array(numbers, {
          repeat: 10
        });

        return i.begin().then((_n) => {
          _n.should.equal(10);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(40);
        }).then((_n) => {
          _n.should.equal(40);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(20);
        }).then((_n) => {
          _n.should.equal(20);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
        });
      });

      /**
       * @name applies a transformation properly:
       */
      it('applies a transformation properly', () => {

        let e = 0;

        let i = new iterator.array(numbers, {
          repeat: 2
        });

        let ev = i.events.on('populate', (_results) => {
          return ++e, _.map(_results, (_r) => {
            return _r * Math.pow(10, e - 1);
          });
        });

        return i.begin().then((_n) => {
          _n.should.equal(10);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(201, true);
        }).then((_n) => {
          _n.should.equal(300);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(400);
        }).then((_n) => {
          _n.should.equal(400);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.next();
        }).then((_n) => {
          _n.should.equal(500);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.next();
        }).then((_n) => {
          should.not.exist(_n);
          should.not.exist(i.current());
          i.is_eof().should.be.true;
          i.events.remove('populate', ev);
        });
      });

      /**
       * @name applies transformations in the correct order:
       */
      it('applies transformations in the correct order', () => {

        let e = 0;

        let i = new iterator.array(numbers, {
          repeat: 2
        });

        let ev1 = i.events.on('populate', (_results) => {
          return ++e, _.map(_results, (_r) => {
            return _r * Math.pow(10, e - 1);
          });
        });

        let ev2 = i.events.on('filter', (_result, _data) => {
          return (
            (_result % 2 === 0) &&
              _result.toString().substring(0, 1) !== '3'
          );
        });

        let ev3 = i.events.on('transform', (_result, _data) => {
          return _result + 1;
        });

        return i.begin().then((_n) => {
          _n.should.equal(11);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(30);
        }).then((_n) => {
          _n.should.equal(41);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(150, true);
        }).then((_n) => {
          _n.should.equal(201);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(550, true);
        }).then((_n) => {
          should.not.exist(_n);
          should.not.exist(i.current());
          i.is_eof().should.be.true;
          _.each([ ev1, ev2, ev3 ], (_e) => {
            (() => { _e.remove(); }).should.not.throw();
          });
        });
      });

      /**
       * @name applies transformations to drained results:
       */
      it('applies transformations to drained results', () => {

        let i = new iterator.array(numbers, {
          repeat: true
        });

        let ev = i.events.on('transform', (_result, _data) => {
          return _result + 1;
        });

        return i.begin().then((_n) => {
          _n.should.equal(11);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          return i.seek(25, true);
        }).then((_n) => {
          _n.should.equal(31);
          i.current().should.equal(_n);
          i.is_eof().should.be.false;
          i.drain().should.deep.equal([ 31, 41, 51 ]);
          return i.seek(150, true);
        }).then((_n) => {
          should.not.exist(_n);
          should.not.exist(i.current());
          i.is_eof().should.be.true;
          ev.remove();
        });
      });

      /**
       * @name works in inexact matching mode:
       */
      it('works in inexact matching mode', () => {

        let i = new iterator.array(numbers, {
          repeat: 2
        });

        let ev = i.events.on('transform', (_n, _data) => {
          return (_n * 10);
        });

        return i.begin().then(() => {
          return i.seek(50);
        }).then((_n) => {
          i.is_eof().should.be.false;
          _n.should.equal(500);
          i.current().should.equal(_n);
          return i.seek(10, true);
        }).then((_n) => {
          i.is_eof().should.be.false;
          _n.should.equal(100);
          i.current().should.equal(_n);
          return i.seek(39, true);
        }).then((_n) => {
          i.is_eof().should.be.false;
          _n.should.equal(400);
          i.current().should.equal(_n);
          return i.seek(51, true);
        }).then((_n) => {
          should.not.exist(_n);
          should.not.exist(i.current());
          i.is_eof().should.be.true;
          (() => { i.events.remove('transform', ev); }).should.not.throw();
        });
      });
    });

    /**
     * @name rewind:
     */
    describe('rewind', () => {

      const more_numbers = [
        1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
      ];

      const duplicate_numbers = [
        1, 2, 3, 4, 5, 6, 6, 6, 7, 8
      ];

      const matrix = [
        [ 1, 1 ], [ 1, 2 ], [ 2, 1 ], [ 2, 2 ], [ 2, 3 ],
        [ 4, 0 ], [ 4, 1 ], [ 7, 7 ], [ 9, 1 ], [ 9, 2 ], [ 9, 9 ]
      ];

      /**
       * @name works in ordered and unordered modes:
       */
      it('works in ordered and unordered modes', () => {

        let test_fn = (_is_ordered) => {
          let i = new iterator.array(more_numbers, {
            repeat: 2, ordered: _is_ordered
          });

          return i.begin().then(() => {
            i.statistics.requests.should.equal(1);
            return i.seek(10);
          }).then((_n) => {
            _n.should.equal(10);
            i.current().should.equal(_n);
            i.statistics.requests.should.equal(1);
            return i.rewind(4.5, true);
          }).then((_n) => {
            _n.should.equal(5);
            i.current().should.equal(_n);
            i.statistics.requests.should.equal(1);
            return i.seek(2);
          }).then((_n) => {
            _n.should.equal(2);
            i.current().should.equal(_n);
            i.statistics.requests.should.equal(2);
            return i.rewind(14);
          }).then((_n) => {
            _n.should.equal(14);
            i.current().should.equal(_n);
            i.statistics.requests.should.equal(3);
            return i.seek(4);
          }).then((_n) => {
            _n.should.equal(4);
            i.current().should.equal(_n);
            i.statistics.requests.should.equal(4);
            return i.rewind(1);
          }).then((_n) => {
            _n.should.equal(1);
            i.current().should.equal(_n);
            i.statistics.requests.should.equal(4);
            return i.seek(16);
          }).then((_n) => {
            should.not.exist(_n);
            should.not.exist(i.current());
            i.is_eof().should.be.true;
          });
        };

        return promise.all([ test_fn(true), test_fn(false) ]);
      });

      /**
       * @name rewinds to the first of duplicate keys:
       */
      it('rewinds to the first of duplicate keys', () => {

        let fn = (_is_ordered) => {
          let i = new iterator.array(duplicate_numbers, {
            repeat: false, ordered: _is_ordered
          });

          return i.begin().then(() => {
            return i.seek(7);
          }).then(() => {
            return i.rewind(6);
          }).then((_n) => {
            _n.should.equal(6);
            i.current().should.equal(_n);
            i.offset.should.equal(5);
          }).then(() => {
            return i.next();
          }).then(() => {
            return i.next();
          }).then(() => {
            return i.rewind(6);
          }).then((_n) => {
            _n.should.equal(6);
            i.current().should.equal(_n);
            i.offset.should.equal(5);
          });
        };

        return promise.all([ fn(true), fn(false) ]);
      });

      /**
       * @name works in inexact mode:
       */
      it('works in inexact mode', () => {

        let fn = (_is_ordered) => {
          let i = new iterator.array(matrix, {
            repeat: false, ordered: _is_ordered
          });

          return i.begin().then(() => {
            return i.seek([ 3, 5 ]);
          }).then((_r) => {
            should.not.exist(_r);
            i.is_eof().should.be.true;
            return i.begin();
          }).then((_r) => {
            _r.should.deep.equal([ 1, 1 ]);
            return i.seek([ 3, 5 ], true);
          }).then((_r) => {
            _r.should.deep.equal([ 4, 0 ]);
            i.current().should.equal(_r);
            i.offset.should.equal(5);
          }).then(() => {
            return i.rewind([ 2 ], true);
          }).then((_r) => {
            _r.should.deep.equal([ 2, 1 ]);
            i.current().should.equal(_r);
            i.offset.should.equal(2);
          }).then(() => {
            return i.seek([ 9, 8 ], true);
          }).then((_r) => {
            _r.should.deep.equal([ 9, 9 ]);
            i.current().should.equal(_r);
            i.offset.should.equal(10);
          }).then(() => {
            return i.rewind([ 10 ], true);
          }).then((_r) => {
            _r.should.deep.equal([ 9, 9 ]);
            i.current().should.equal(_r);
            i.offset.should.equal(10);
          }).then(() => {
            return i.rewind([ 9, 0 ], true);
          }).then((_r) => {
            _r.should.deep.equal([ 9, 1 ]);
            i.current().should.equal(_r);
            i.offset.should.equal(8);
          });
        };

        return promise.all([ fn(true), fn(false) ]);
      });
    });

    /**
     * @name previous:
     */
    describe('previous', () => {

      /**
       * @name iterates backward correctly:
       */
      it('iterates backward correctly', () => {

        let fn = (_is_ordered) => {
          let i = new iterator.array(numbers, {
            repeat: false, ordered: _is_ordered
          });

          return i.begin().then((_r) => {
            return i.seek(49, true);
          }).then((_r) => {
            _r.should.equal(50);
            i.current().should.equal(_r);
            i.offset.should.equal(4);
            return i.previous();
          }).then((_r) => {
            _r.should.equal(40);
            i.current().should.equal(_r);
            i.offset.should.equal(3);
            return i.rewind(9, true);
          }).then((_r) => {
            _r.should.equal(10);
            i.current().should.equal(_r);
            i.offset.should.equal(0);
            return i.previous();
          }).then((_r) => {
            should.not.exist(_r);
            i.is_bof().should.be.true;
          });
        };

        return promise.all([ fn(true), fn(false) ]);
      });
    });


    /**
     * @name filter:
     */
    describe('filter', () => {

      /**
       * @name skips items correctly:
       */
      it('skips items correctly', () => {

        let limit = 0;

        const step = 47;
        const limit_max = 190;

        let i = new iterator.array(numbers, {
          repeat: true
        });

        /* Vary iteration limit */
        return promise.while(() => ((limit += step) < limit_max), () => {

          let skip = 0;
          let skip_max = numbers.length;

          /* Vary filter criteria */
          return promise.while(() => (++skip < skip_max), () => {

            let n = 0;
            let threshold = skip * 10;
            let batch = (numbers.length - skip);

            /* Filter: survives resets */
            let ev = i.events.on('filter', (_r, _data) => {

              if (_r <= threshold) {
                _data.skipped = (_data.skipped || 0) + 1;
                return false;
              }
              _data.accepted = (_data.accepted || 0) + 1;
              return true;
            });

            return i.begin().then((_n) => {
              return promise.while(() => (n++ < limit), () => {
                i.current().should.be.an.integer;
                i.current().should.be.at.least(threshold);
                return i.next();
              });

            }).then(() => {
              i.statistics.filter.data.skipped.should.equal(
                skip * Math.ceil(limit / batch) +
                  (limit % batch === 0 ? skip : 0)
              );

              i.statistics.filter.data.accepted.should.equal(n);
              i.events.remove('filter', ev);
            });
          });
        });
      });

      /**
       * @name skips items correctly when seeking:
       */
      it('skips items correctly when seeking', () => {

        let i = new iterator.array(numbers, {
          repeat: true
        });

        let ev = i.events.on('filter', (_r) => {
          return (_r >= 30);
        });

        return i.begin().then((_n) => {
          _n.should.equal(30);
          return i.next();
        }).then((_n) => {
          _n.should.equal(40);
          return i.seek(10);
        }).then((_n) => {
          _n.should.equal(30);
          i.events.remove('filter', ev);
        });
      });

      /**
       * @name reaches EOF via next:
       */
      it('reaches EOF via next', () => {

        const n = 40;

        let i = new iterator.array(numbers, {
          repeat: 2, ordered: true
        });

        let ev = i.events.on('filter', (_r) => {
          return (_r >= n);
        });

        return i.begin().then((_n) => {
          _n.should.equal(40);
          return i.next();
        }).then((_n) => {
          _n.should.equal(50);
          return i.next();
        }).then((_n) => {
          _n.should.equal(40);
          return i.seek(10);
        }).then((_n) => {
          should.not.exist(_n);
          i.is_eof().should.be.true;
          i.events.remove('filter', ev);
        });
      });

      /**
       * @name reaches EOF via seek:
       */
      it('reaches EOF via seek', () => {

        const n = 30;

        let i = new iterator.array(numbers, {
          repeat: 2, ordered: true
        });

        let ev = i.events.on('filter', (_r) => {
          return (_r >= n);
        });

        return i.begin().then((_n) => {
          _n.should.equal(30);
          return i.seek(10);
        }).then((_n) => {
          _n.should.equal(30);
          return i.seek(20);
        }).then((_n) => {
          should.not.exist(_n);
          i.is_eof().should.be.true;
          i.events.remove('filter', ev);
        });
      });

      /**
       * @name reaches EOF with a reject-all filter:
       */
      it('reaches EOF with a reject-all filter', () => {

        let i = new iterator.array(numbers, {
          repeat: 10
        });

        let ev = i.events.on('filter', (_r) => {
          return false;
        });

        return i.begin().then((_n) => {
          should.not.exist(_n);
          i.is_eof().should.be.true;
          i.events.remove('filter', ev);
        });
      });
    });
  });
});

