
'use strict';

import chai from 'chai';
import _ from '../utils/lodash';
import promises from 'chai-as-promised';
import schema from '../../../src/schema';
import promise from '../../../src/promise';
import iterator from '../../../src/iterator';
import join from '../../../src/iterator/join';
import exception from '../../../src/exception';
import fixture_utils from '../../include/fixtures';
import fixtures from '../../../output/fixtures/all';

/**
 * @name iterator:
 *   Unit tests for the asynchronous iterator.
 */
describe('iterator', () => {

  let should = chai.should();
  chai.use(promises);

  /**
   * @name join:
   */
  describe('join', () => {

    /**
     * @name array;
     */
    describe('array', () => {

      /**
       * @name requires ordered iterators:
       */
      it('requires ordered iterators', () => {

        let i1 = new iterator.array([]);
        let i2 = new iterator.array([]);

        (() => {
          new join(i1, i2);
        }).should.throw(exception.argument);
      });

      /**
       * @name is always ordered:
       */
      it('is always ordered', () => {

        let i1 = new iterator.array([], { ordered: true });
        let i2 = new iterator.array([], { ordered: true });
        let j = new join(i1, i2);

        j.is_ordered().should.be.true;
      });

      /**
       * @name requires key mapping to be a function:
       */
      it('requires key mapping to be a function', () => {

        let i1 = new iterator.array([], { ordered: true });
        let i2 = new iterator.array([], { ordered: true });

        (() => {
          let j = new join(i1, i2);
          j.events.on('key.subset', 'Not a function');
        }).should.throw(exception.argument);
      });

      /**
       * @name joins two empty iterators:
       */
      it('joins two empty iterators', () => {

        let i1 = new iterator.array([], { ordered: true });
        let i2 = new iterator.array([], { ordered: true });

        let j = new join(i1, i2);

        return j.begin().then(() => {
          j.is_eof().should.be.true;
        });
      });

      /**
       * @name joins two matching iterators:
       */
      it('joins two matching iterators', () => {

        let i1 = new iterator.array([ 2 ], { ordered: true });
        let i2 = new iterator.array([ 2 ], { ordered: true });

        let j = new join(i1, i2);

        return j.begin().then((_rv) => {
          should.exist(_rv);
          should.exist(_rv.left); _rv.left.should.equal(2);
          should.exist(_rv.right); _rv.right.should.equal(2);
          j.is_eof().should.be.false;
          return j.next();
        }).then((_rv) => {
          should.not.exist(_rv);
          j.is_eof().should.be.true;
        });
      });

      /**
       * @name joins two non-matching iterators:
       */
      it('joins two non-matching iterators', () => {

        let i1 = new iterator.array([ 1 ], { ordered: true });
        let i2 = new iterator.array([ 2 ], { ordered: true });

        let j = new join(i1, i2);

        return j.begin().then((_rv) => {
          should.not.exist(_rv);
          j.is_eof().should.be.true;
        });
      });

      /**
       * @name joins two ordered iterators:
       */
      it('joins two ordered iterators', () => {

        let results = [];
        let o = { ordered: true };

        let i1 = new iterator.array([
          [ 0, 'abc' ], [ 0, 'def' ], [ 0, 'ghi' ],
          [ 1, 'abcd' ], [ 1, 'efgh' ], [ 1, 'ijkl' ], [ 2, 'memes' ]
        ], o);

        let i2 = new iterator.array([
          [ 0, 'd' ], [ 0, 'ghi' ], [ 1, 'abcd' ], [ 1, 'efgh' ],
          [ 1, 'i' ], [ 1, 'jkl' ], [ 2, 'memes' ], [ 3, 'dank' ]
        ], o);

        let j = new join(i1, i2);

        return j.begin().then(() => {
          return j.each((_pair) => {
            results.push(_pair.left);
          });
        }).then(() => {
          i1.is_eof().should.be.true;
          i2.is_eof().should.be.false;
          results.should.deep.equal([
            [ 0, 'ghi' ], [ 1, 'abcd' ], [ 1, 'efgh' ], [ 2, 'memes' ]
          ]);
        });
      });

      /**
       * @name takes a cross-product for identical keys'
       */
      it('takes a cross-product for identical keys', () => {

        let test_fn = (_l, _r, _expect) => {

          let results = [];
          let o = { ordered: true };

          let i1 = new iterator.array(_l, o);
          let i2 = new iterator.array(_r, o);

          let j = new join(i1, i2);

          return j.begin().then(() => {
            return j.each((_pair) => {
              results.push(_pair.left);
            });
          }).then(() => {
            results.should.deep.equal(_expect);
          });
        };

        let fn = (_l, _r, _expect) => {

          return promise.all([
            test_fn(_l, _r, _expect),
            test_fn(_r, _l, _expect)
          ]);
        };

        return promise.all([
          fn([ 1, 4, 4, 6, 7, 9, 10 ],
             [ 0, 1, 2, 3, 4, 4, 5, 6, 9 ],
             [ 1, 4, 4, 4, 4, 6, 9 ]),
          fn([ 1, 4, 4, 4, 6, 7, 9, 10 ],
             [ 0, 1, 2, 3, 4, 4, 5, 6, 9 ],
             [ 1, 4, 4, 4, 4, 4, 4, 6, 9 ]),
          fn([ 1, 4, 4, 6, 7, 9, 10 ],
             [ 0, 1, 2, 3, 4, 4, 4, 5, 6, 9 ],
             [ 1, 4, 4, 4, 4, 4, 4, 6, 9 ])
        ]);
      });

      /**
       * @name copes with ordering constraint violations:
       */
      it('copes with ordering constraint violations', () => {

        let results = [];
        let o = { ordered: true };

        let i1 = new iterator.array([
          1, 2, 3, 4, 7, 3, 4, 3, 4, 3, 1, 2, 3, 4
        ], o);

        let i2 = new iterator.array([
          3, 4, 5, 2, 3, 4, 5, 6, 7
        ], o);

        let j = new join(i1, i2);

        return j.begin().then(() => {
          return j.each((_pair) => {
            results.push(_pair.left);
          });
        }).then(() => {
          results.should.deep.equal([ 3, 4 ]);
        });
      });

      /**
       * @name applies a mapping function to keys:
       */
      it('applies a mapping function to keys', () => {

        let results = [];
        let o = { ordered: true };

        let i1 = new iterator.array([
          [ 0, 'a', 1 ], [ 0, 'b', 0 ], [ 0, 'c', 10 ],
          [ 1, 'a', 1 ], [ 1, 'b', 100 ], [ 1, 'c', 99 ]
        ], o);

        let i2 = new iterator.array([
          [ 0, 'a', 0 ], [ 0, 'c', 100 ], [ 0, 'd', 1 ],
          [ 1, 'a', 100 ], [ 1, 'b', 10 ], [ 2, 'c', 1 ]
        ], o);

        let j = new join(i1, i2);

        j.events.on(
          'key.subset', (_r) => _r.slice(0, 2)
        );

        return j.begin().then(() => {
          return j.each((_pair) => {
            results.push(_pair);
          });
        }).then(() => {
          results.should.deep.equal([
            { left: [ 0, 'a', 1 ], right: [ 0, 'a', 0 ] },
            { left: [ 0, 'c', 10 ], right: [ 0, 'c', 100 ] },
            { left: [ 1, 'a', 1 ], right: [ 1, 'a', 100 ] },
            { left: [ 1, 'b', 100 ], right: [ 1, 'b', 10 ] }
          ]);
        });
      });

      /**
       * @name rewinds:
       */
      it('rewinds', () => {

        let results = [];
        let o = { ordered: true };

        let i1 = new iterator.array([
          1, 2, 3, 5, 8, 13
        ], o);

        let i2 = new iterator.array([
          2, 4, 6, 8, 10, 12
        ], o);

        let j = new join(i1, i2);

        return j.begin().then(() => {
          return j.each((_pair) => {
            results.push(_pair.left);
          });
        }).then(() => {
          return j.rewind();
        }).then(() => {
          return j.each((_pair) => {
            results.push(_pair.right);
          });
        }).then(() => {
          i1.is_eof().should.be.false;
          i2.is_eof().should.be.true;
          results.should.deep.equal([
            2, 8, 2, 8
          ]);
        });
      });
    });

    /**
     * @name view:
     */
    describe('view', () => {

      let fixture_set = null;
      const f = [ 'litany.txt', 'mantra.txt' ];

      let s = new schema.pouchdb(
        'output/databases/unit/iterator-join'
      );

      before(() => {
        return s.deploy().then(() => {
          return fixture_utils.store_fixtures(s, f, fixtures.unit.text);
        }).then((_f) => {
          fixture_set = _f;
        });
      });

      after(() => {
        return s.destroy().catch((_e) => {
          /* Ignore error */
        });
      });

      /**
       * @name self-join reaches end of file:
       */
      it('self-join reaches end-of-file', () => {

        let i1 = s.view_iterator('words');
        let i2 = s.view_iterator('words');

        let results = [];
        let j = new join(i1, i2);

        return j.begin().then(() => {
          return j.each((_r) => results.push(_r));
        }).then(() => {
          j.is_eof().should.be.true;
          i1.is_eof().should.be.true;
          i2.is_eof().should.be.true;
          results.length.should.equal(103);
        });
      });

      /**
       * @name joins with varying limits:
       */
      it('joins with varying limits', () => {

        let fn = (_limit) => {

          let i1 = s.view_iterator('words', {
            view: {
              include_docs: false, limit: _limit,
              startkey: [ 0 ], endkey: [ 0, 'past' ]
            }
          });

          let i2 = s.view_iterator('words', {
            view: {
              include_docs: false, limit: _limit,
              startkey: [ 0, 'not' ], endkey: [ 1 ]
            }
          });

          let results = [];

          let j = new join(i1, i2, {
            limit: _limit
          });

          return j.begin().then(() => {
            return j.each((_r) => results.push(_r));
          }).then(() => {

            let expect = [
              [ 0, 'not', 7 ], [ 0, 'nothing', 276 ],
              [ 0, 'obliteration', 85 ], [ 0, 'of', 64 ],
              [ 0, 'only', 285 ], [ 0, 'over', 145 ], [ 0, 'pass', 140 ]
            ];

            results.length.should.equal(expect.length);

            _.each(results, (_r, _i) => {
              fixture_utils.compare_word_index_key(_r.left.key, expect[_i]);
            });

            _.map(results, (_r) => _r.left).should.deep.equal(
              _.map(results, (_r) => _r.right)
            );
          });
        };

        return promise.all([
          fn(1), fn(2), fn(5), fn(7), fn(13)
        ]);
      });

      /**
       * @name joins with an array iterator:
       */
      it('joins with an array iterator', () => {

        let rows = [];
        let results = {};

        let build_fn = (_offsets) => {
          return _.mapValues(_offsets, (_pair) => {
            return _.reduce(_pair[0], (_rv, _l) => {
              _rv[_l] = _pair[1]; return _rv;
            }, {});
          });
        };

        /* Powersets */
        let expect = build_fn({
          that: [ [ 73 ], [ 67 ] ],
          my: [ [ 26, 182 ], [ 111 ] ],
          is: [ [ 3, 48, 159 ], [ 22, 47 ] ],
          it: [ [ 0, 45, 156 ], [ 134, 178 ] ],
          i: [ [ 20, 176 ], [ 0, 99, 120, 195, 290 ] ],
          the: [ [ 54, 102, 127 ], [ 25, 50, 207, 244 ] ],
          will: [ [ 9, 165 ], [ 101, 122, 197, 268, 292 ] ]
        });

        /* Find `mantra.txt` fixture */
        let it = s.view_iterator('nodes_by_title', {
          view: {
            include_docs: true,
            startkey: [ 0, 'mantra.txt' ], limit: 1
          }
        });

        return it.begin().then((_rv) => {

          /* ...and remove it */
          let p = promise.all([
            s.remove(_rv.doc),
            s.remove(fixture_set.pop().doc)
          ]);

          /* But save its word list */
          return p.then((_r) => {
            return _.sortBy(_rv.doc.data.words, 0);
          });

        }).then((_words) => {

          /* Create source iterators */
          let i1 = new iterator.array(_words, {
            repeat: false, ordered: true
          });

          let i2 = s.view_iterator('words', {
            view: { include_docs: false }
          });

          let j = new join(i1, i2);

          /* Test key format modification */
          i1.events.on('populate', (_rows) => {
            return _.map(_rows, (_r) => {
              should.exist(_r);
              return { test: [ 0, _r[0], _r[1] ] };
            });
          });

          /* Test key mapping */
          i1.events.on('key.map', (_r) => {
            should.exist(_r);
            return _r.test;
          });

          /* Test partial key comparison */
          j.events.on('key.subset', (_r) => {
            should.exist(_r);
            return _r.slice(0, 2);
          });

          /* Format results */
          return j.begin().then(() => {
            return j.each((_r) => {

              let word = _r.left.test[1];
              let left_offset = _r.left.test[2];
              let right_offset = _r.right.key[3];

              if (!results[word]) {
                results[word] = {};
              }

              if (!results[word][left_offset]) {
                results[word][left_offset] = [];
              }

              results[word][left_offset].push(right_offset);
              rows.push(_r);
            });
          });

        }).then(() => {

          /* Assert equality */
          rows.length.should.equal(47);
          expect.should.deep.equal(results);
        });
      });
    });
  });
});

