
'use strict';

import chai from 'chai';
import _ from '../utils/lodash';
import utils from '../../../src/utils';
import promises from 'chai-as-promised';
import schema from '../../../src/schema';
import promise from '../../../src/promise';
import indexer from '../../../src/indexer';
import iterator from '../../../src/iterator';
import fixtures from '../../include/fixtures';

/**
 * @name iterator:
 *   Unit tests for the asynchronous iterator.
 */
describe('iterator', () => {

  let should = chai.should();
  chai.use(promises);

  let s = new schema.pouchdb(
    'output/databases/unit/iterator-view'
  );

  /**
   * @name view:
   */
  describe('view', () => {

    /**
     * @name startup:
     */
    describe('startup', () => {

      /**
       * @name deploys:
       */
      it('deploys', () => {
        return s.deploy();
      });
    });

    /**
     * @name single document:
     */
    describe('single document', () => {

      let fixture = null;
      let total_words = 0;

      /**
       * @name before:
       */
      before(() => {
        return fixtures.store_fixture(s, 'lorem-ipsum.txt').then((_f) => {
          fixture = _f;
          total_words = _f.linked_doc.data.words.length;
        });
      });

      /**
       * @name after:
       */
      after(() => {
        return fixtures.remove_fixtures(s, fixture);
      });

      /**
       * @name iterates with low result limits:
       */
      it('iterates with low result limits', () => {

        let limit = 1;
        const max_limit = 16;

        return promise.while(() => (limit <= max_limit), () => {

          let total = 0;
          let previous = false;

          let i = new iterator.view(s, 'words', {
            view: { limit: limit, include_docs: false }
          });

          return i.begin().then(() => {
            return promise.while(() => !i.is_eof(), () => {

              if (previous) {
                /* Items must be monotonically increasing */
                s.collate(previous.key, i.current().key)
                  .should.be.below(0);
              }

              ++total;
              previous = i.current();

              return i.next();

            }).then(() => {
              ++limit;
              total.should.equal(total_words);
            });
          });
        });
      });

      /**
       * @name fills with low result limits:
       */
      it('fills with low result limits', () => {

        let limit = 1;
        const max_limit = 16;

        return promise.while(() => (limit <= max_limit), () => {

          let total = 0;
          let previous = false;

          let i = new iterator.view(s, 'words', {
            view: { limit: limit, include_docs: false }
          });

          return i.begin().then(() => {
            return promise.while(() => !i.is_eof(), () => {

              let r = i.drain();

              if (previous) {
                /* Groups must be monotonically increasing */
                s.collate(previous.key, _.first(r).key)
                  .should.be.below(0);
              }

              for (let j = 1, len = r.length; j < len; ++j) {
                /* Items must be monotonically increasing */
                s.collate(r[j - 1].key, r[j].key)
                  .should.be.below(0);
              }

              total += r.length;
              previous = _.last(r);

              return i.fill();

            }).then(() => {
              ++limit;
              total.should.equal(total_words);
            });
          });
        });
      });

      /**
       * @name buffers results:
       */
      it('buffers results', () => {

        let limit = total_words - 1;

        let i = new iterator.view(s, 'words', {
          view: { limit: limit, include_docs: false }
        });

        return i.begin().then(() => {
          i.statistics.requests.should.equal(1);
          return i.seek([ 0, 'veniam' ], true);
        }).then(() => {
          i.current().key.should.be.an.array;
          i.current().key[1].should.equal('veniam');
          i.statistics.requests.should.equal(1);
          return i.seek([ 0, 'voluptate' ], true);
        }).then(() => {
          i.current().key.should.be.an.array;
          i.current().key[1].should.equal('voluptate');
          i.statistics.requests.should.equal(2);
          return i.next();
        }).then(() => {
          i.is_eof().should.be.true;
        });
      });

      /**
       * @name changes buffer size:
       */
      it('changes buffer size', () => {

        const limit = 4;

        let i = new iterator.view(s, 'words', {
          view: { limit: limit, include_docs: false }
        });

        return i.begin().then(() => {
          i.statistics.requests.should.equal(1);
          return i.seek([ 0, 'culpa' ], true);
        }).then(() => {
          i.current().key.should.be.an.array;
          i.current().key[1].should.equal('culpa');
          i.statistics.requests.should.equal(2);
          return i.seek([ 0, 'do' ], true);
        }).then(() => {
          i.option('view.limit', 100);
          i.current().key.should.be.an.array;
          i.current().key[1].should.equal('do');
          i.statistics.requests.should.equal(2);
          return i.seek([ 0, 'dolor' ], true);
        }).then(() => {
          i.current().key.should.be.an.array;
          i.current().key[1].should.equal('dolor');
          i.statistics.requests.should.equal(3);
          i.drain().length.should.equal(54);
          return i.next();
        }).then(() => {
          i.is_eof().should.be.true;
        });
      });

      /**
       * @name seeks past EOF:
       */
      it('seeks past EOF', () => {

        const limit = 7;

        let i = new iterator.view(s, 'words', {
          view: { limit: limit, include_docs: false }
        });

        return i.begin().then(() => {
          return i.seek([ 0, 'zzyz' ]);
        }).then((_r) => {
          should.not.exist(_r);
          i.is_eof().should.be.true;
        });
      });

      /**
       * @name exactly matches the source view:
       */
      it('exactly matches the source view', () => {

        let offset = 0;
        const limits = [ 1, 3, 5, 7, 11, 13, 17, 19, 23 ];

        return promise.while(() => offset++ < limits.length, () => {

          let rows = [];

          let i = new iterator.view(s, 'words', {
            view: { limit: limits[offset], include_docs: false }
          });

          return i.begin().then(() => {
            return s.view('words');
          }).then((_view) => {
            return promise.while(() => !i.is_eof(), () => {
              rows.push(i.current());
              return i.next();
            }).then(() => {
              rows.should.deep.equal(_view.rows);
            });
          });
        });
      });

      /**
       * @name seeks in force-advance mode:
       */
      it('seeks in force-advance mode', () => {

        const limit = 3;

        let i = new iterator.view(s, 'words', {
          view: { limit: limit, include_docs: false }
        });

        return i.begin().then(() => {
          i.statistics.requests.should.equal(1);
          i.current().key[1].should.equal('ad');
          return i.seek([ 0 ], true);
        }).then(() => {
          i.statistics.requests.should.equal(1);
          i.current().key[1].should.equal('adipiscing');
          return i.seek([ 0 ], true, true);
        }).then(() => {
          i.statistics.requests.should.equal(1);
          i.current().key[1].should.equal('aliqua');
          return i.seek([ 0 ], true, true);
        }).then(() => {
          i.statistics.requests.should.equal(2);
          i.current().key[1].should.equal('aliquip');
          i.is_eof().should.be.false;
          return i.seek([ 1 ], true, true);
        }).then(() => {
          i.statistics.requests.should.equal(3);
          i.is_eof().should.be.true;
        });
      });

      /**
       * @name correctly reads the final item:
       */
      it('correctly reads the final item', () => {

        let offset = 0;
        const limits = [ 1, 2, 3, 4, 5, 7, 9, 15 ];

        return promise.while(() => offset < limits.length, () => {

          let final_item = null;

          let i = new iterator.view(s, 'words', {
            view: { limit: limits[offset], include_docs: false }
          });

          return i.begin().then(() => {
            return i.each((_i) => {
              final_item = _i;
            });
          }).then(() => {
            final_item.key[1].should.equal('voluptate');
            offset++;
          });
        });
      });

      /**
       * @name reads identically forward and backward:
       */
      it('reads identically forward and backward', () => {

        let offset = 0;
        const limits = [ 1, 2, 3, 5, 7, 13, 21, 25 ];

        return promise.while(() => offset < limits.length, () => {

          let forward = [], backward = [];

          let i = new iterator.view(s, 'words', {
            view: { limit: limits[offset], include_docs: false }
          });

          return i.all().then((_all) => {
            forward = _all;
            return i.previous();
          }).then(() => {
            return promise.while(() => !i.is_bof(), () => {
              backward.unshift(i.current());
              return i.previous();
            });
          }).then(() => {
            backward.should.deep.equal(forward);
            offset++;
          });
        });
      });

      /**
       * @name reads in both directions with reject-all filter:
       */
      it('reads in both directions with reject-all filter', () => {

        let offset = 0;
        const limits = [  1, 2, 3, 5, 7, 13 ];

        return promise.while(() => offset < limits.length, () => {

          let forward = [], backward = [];

          let i = new iterator.view(s, 'words', {
            view: { limit: limits[offset], include_docs: false }
          });

          i.events.on('filter', (_r) => false);

          return i.all().then((_all) => {
            forward = _all;
            return i.previous();
          }).then(() => {
            return promise.while(() => !i.is_bof(), () => {
              backward.unshift(i.current());
              return i.previous();
            });
          }).then(() => {
            backward.should.deep.equal(forward);
            offset++;
          });
        });
      });

      /**
       * @name reads in both directions with reject-some filter:
       */
      it('reads in both directions with reject-some filter', () => {

        let offset = 0;
        const limits = [  1, 2, 3, 5, 7, 9, 15, 21 ];

        return promise.while(() => offset < limits.length, () => {

          let forward = [], backward = [];

          let i = new iterator.view(s, 'words', {
            view: { limit: limits[offset], include_docs: false }
          });

          i.events.on('filter', (_r) => {
            return !!(_r.key[1].substring(0,1).charCodeAt(0) % 2);
          });

          return i.all().then((_all) => {
            forward = _all;
            return i.previous();
          }).then((_r) => {
            return promise.while(() => !i.is_bof(), () => {
              backward.unshift(i.current());
              return i.previous();
            });
          }).then(() => {
            backward.should.deep.equal(forward);
            offset++;
          });
        });
      });

      /**
       * @name has a working word index:
       */
      it('has a working word index', () => {

        return s.view('words', { include_docs: true }).then((_v) => {
          _v.should.be.an.object;
          _v.rows.should.be.an.array;
          _.each(_v.rows, (_r) => {
            _r.doc.type.should.equal('index');
            _r.doc.node_id.should.equal(fixture.doc._id);
            _r.doc._id.should.equal(fixture.linked_doc._id);
          });
        });
      });

      /**
       * @name iterates over words:
       */
      it('iterates over words', () => {

        let offset = 0;
        const limits = [  1, 2, 3, 5, 7, 9, 15, 21 ];

        return promise.while(() => offset < limits.length, () => {

          let i = s.view_iterator('words', {
            limit: limits[offset]
          });

          return i.begin().then((_r) => {
            fixtures.compare_word_index_key(_r.key, [ 0, 'ad', 132 ]);
            return i.seek([ 0, 'in' ]);
          }).then((_r) => {
            fixtures.compare_word_index_key(_r.key, [ 0, 'in', 255 ]);
            return i.seek([ 0, 'ipsum' ], true);
          }).then((_r) => {
            fixtures.compare_word_index_key(_r.key, [ 0, 'ipsum', 6 ]);
            return i.next();
          }).then((_r) => {
            fixtures.compare_word_index_key(_r.key, [ 0, 'irure', 243 ]);
            return i.seek([ 0, 'lorel' ], true);
          }).then((_r) => {
            fixtures.compare_word_index_key(_r.key, [ 0, 'lorem', 0 ]);
            offset++;
          });
        });
      });
    });

    /**
     * @name multiple documents:
     */
    describe('multiple documents', () => {

      let fixture_set = null;
      const f = [ 'litany.txt', 'mantra.txt' ];

      /**
       * @name before:
       */
      before(() => {
        return fixtures.store_fixtures(s, f).then((_f) => {
          fixture_set = _f;
        });
      });

      /**
       * @name after:
       */
      after(() => {
        return fixtures.remove_fixtures(s, fixture_set);
      });

      /**
       * @name iterates over words:
       */
      it('iterates over words', () => {

        let summary = _.reduce(f, (_rv, _f) => {
          return (_rv[_f] = {}, _rv);
        }, {});

        let i = s.view_iterator('words', {
          limit: 128, view: { include_docs: true }
        });

        return i.begin().then(() => {
          return i.seek([ 0, 'will' ], true);
        }).then(() => {
          return promise.while(() => !i.is_eof(), () => {
            let c = i.current();
            let k = c.key, v = c.value;
            let word = k[1], offset = k[3];
            summary[c.doc.data.title][offset] = word;
            return i.next();
          });
        }).then(() => {
          summary.should.deep.equal({
            'mantra.txt': {
              9: 'will', 165: 'will'
            },
            'litany.txt': {
              101: 'will', 122: 'will',
              197: 'will', 268: 'will', 292: 'will'
            }
          });
        });
      });
    });

    /**
     * @name shutdown:
     */
    describe('shutdown', () => {

      /**
       * @name destroys itself:
       */
      it('destroys itself', () => {
        return s.destroy();
      });
    });

  });
});

