
'use strict';

import chai from 'chai';
import _ from '../utils/lodash';
import utils from '../../../src/utils';
import promises from 'chai-as-promised';
import schema from '../../../src/schema';
import promise from '../../../src/promise';
import indexer from '../../../src/indexer';
import fixtures from '../../include/fixtures';
import join from '../../../src/iterator/join';
import exception from '../../../src/exception';

/**
 * @name check_document:
 *   Helper function for document verification.
 */
let check_document = (_doc, _id) => {

  if (!_.isNil(_id)) {
    _doc._id.should.equal(_id);
  }

  _doc._id.should.exist;
  _doc._id.should.not.be.empty;
  _doc._rev.should.exist;
  _doc._rev.should.not.be.empty;

  return _doc;
};

/**
 * @name schema:
 *   Unit tests for the schema module.
 */
describe('schema', () => {

  let should = chai.should();
  chai.use(promises);

  let s = new schema.pouchdb(
    'output/databases/unit/schema'
  );

  /**
   * @name pouchdb:
   */
  describe('pouchdb', () => {

    /**
     * @name after:
     *   Destroy and compact the database to release storage.
     */
    after(() => {

      return s.destroy();
    });

    /**
     * @name deployment:
     *   Tests related to schema deployment.
     */
    describe('deployment', () => {

      /**
       * @name builds a design document properly:
       */
      it('builds a design document properly', () => {

        var ddoc = s._build_design_document();

        ddoc._id.should.be.a.string;
        ddoc.version.should.be.an.integer;

        ddoc.views.should.be.an.object;
        ddoc.views.words.map.should.be.a.string;
        should.not.exist(ddoc.views.words.reduce);
      });

      /**
       * @name has a proper pre-deployment version:
       */
      it('has a proper pre-deployment version', () => {

        s.version().should.eventually.equal(0);
      });

      /**
       * @name deploys the schema successfully:
       */
      it('deploys the schema successfully', () => {

        return s.deploy();
      });

      /**
       * @name compacts the schema successfully:
       */
      it('compacts the schema successfully', () => {

        return s.compact();
      });

      /**
       * @name has a proper post-deployment version:
       */
      it('has a proper post-deployment version', () => {

        s.version().should.eventually.equal(s.latest_version);
      });

      /**
       * @name redeploys the same schema successfully:
       */
      it('redeploys the same schema successfully', () => {

        return s.deploy();
      });

      /**
       * @name undeploys an existing schema successfully:
       */
      it('undeploys an existing schema successfully', () => {

        return s.undeploy();
      });

      /**
       * @name reverts to version zero after destruction:
       */
      it('reverts to version zero after destruction', () => {

        s.version().should.eventually.equal(0);
      });

      /**
       * @name redeploys the schema successfully:
       */
      it('redeploys the schema successfully', () => {

        return s.deploy();
      });
    });

    /**
     * @name manipulation:
     *   Document manipulation and retrieval tests.
     */
    describe('manipulation', () => {

      /**
       * @name store:
       */
      describe('store', () => {

        /**
         * @name with an identifier:
         */
        it('with an identifier', () => {

          return s.store({ _id: 'i1' }).then((_doc) => {
            check_document(_doc, 'i1');
            return s.remove(_doc);
          });
        });

        /**
         * @name without an identifier:
         */
        it('without an identifier', () => {

          return s.store({ test: true }).then((_doc) => {
            check_document(_doc);
            return s.remove(_doc);
          });
        });

        /**
         * @name with indexer output:
         */
        it('with indexer output', () => {

          let i = new indexer.text();

          const f = 'lorem-ipsum.txt';
          let text = fixtures.all().unit.text['lorem-ipsum.txt'];

          return s.store(i.build(text, {
            tags: [ 'important', 'simple' ]
          }).result).then((_doc) => {
            check_document(_doc);
            return s.remove(_doc);
          });
        });
      });

      /**
       * @name store linked:
       */
      describe('store linked', () => {

        /**
         * @name without _id:
         */
        it('without _id', () => {

          return s.store({}).then((_doc) => {
            return s.store_linked(_doc, { test: true }).then((_r) => {
              check_document(_r.doc);
              check_document(_r.linked_doc);
              _r.doc._id.should.equal(_r.linked_doc.linked_id);
              return _r;
            });
          }).then((_r) => {
            return fixtures.remove_fixtures(s, _r);
          });
        });

        /**
         * @name simple conflict:
         */
        it('simple conflict', () => {

          let doc = null;

          return s.store({}).then((_doc) => {
            return s.store(_doc).then((_d) => {
              doc = _d;
              return _doc; /* Previous revision */
            });
          }).then((_doc) => {
            return s.store_linked(_doc, { t: 1 }).catch((_e) => {
              _e.status.should.equal(409);
              should.not.exist(_e.linked);
            });
          }).then(() => {
            return s.remove(doc);
          });
        });

        /**
         * @name simple conflict with _id and retries:
         */
        it('simple conflict with _id and retries', () => {

          let final_doc = null;

          return s.store({ _id: 'i2' }).then((_doc) => {
            return s.store(_doc).then((_d) => {
              final_doc = _d; return _doc;
            });
          }).then((_doc) => {
            let p = s.store_linked(
              _doc, { t: 1 }, 'a', { retries: 5 }
            );
            return p.catch((_e) => {
              _e.status.should.equal(409);
              should.not.exist(_e.linked);
              return s.remove(final_doc);
            });
          });
        });

        /**
         * @name complex conflict:
         */
        it('complex conflict', () => {

          let n = 5;
          let final_doc = null;

          let o = {
            retries: (_doc) => {
              return s.get(_doc._id).then((_d) => {
                return s.store(_d); /* Force conflict */
              }).then((_d) => {
                return --n;
              });
            }
          };

          return s.store({}).then((_doc) => {
            return s.store(_doc).then((_d) => {
              final_doc = _d;
              return _doc;
            });
          }).then((_doc) => {
            return s.store_linked(_doc, { t: 1 }, 'b', o).then((_r) => {
              should.exist(null); /* Not reached */
            }).catch((_e) => {
              n.should.equal(0);
              should.exist(_e.linked);
              should.exist(_e.linked.doc);
              should.exist(_e.linked.doc._id);
              _e.status.should.equal(409);
              _e.linked.status.should.equal(409);

              return s.get(_e.linked.doc._id).then((_d) => {
                return s.remove(_d);
              }).then((_d) => {
                return s.remove(final_doc);
              });
            });
          });
        });
      });

      /**
       * @name remove linked:
       */
      describe('remove linked', () => {

        /**
         * @name removes a document:
         */
        it('removes a document', () => {

          return s.store({}).then((_doc) => {
            return s.store_linked(_doc, { test: true }).then((_rv) => {
              return s.remove_linked(_rv.doc);
            });
          }).then((_r) => {
            return fixtures.remove_fixtures(s, _r);
          });
        });

        /**
         * @name resists conflicts:
         */
        it('resists conflicts', () => {

          return s.store({}).then((_doc) => {
            return s.store_linked(_doc, { test: true }).then((_rv) => {
              return s.store(_rv.linked_doc);
            });
          }).then((_doc) => {
            return s.remove_linked(_doc);
          }).then((_r) => {
            return fixtures.remove_fixtures(s, _r);
          });
        });
      });

      /**
       * @name get:
       */
      describe('get', () => {

        /**
         * @name with _id:
         */
        it('with _id', () => {

          let doc = null;
          return s.store({ _id: 'i3' }).then((_doc) => {
            doc = check_document(_doc, 'i3');
            return s.get(doc._id);
          }).then((_d) => {
            doc._id.should.equal(_d._id);
            return s.remove(_d);
          });
        });

        /**
         * @name without _id:
         */
        it('without _id', () => {

          let doc = null;
          return s.store({}).then((_doc) => {
            doc = check_document(_doc);
            return s.get(doc._id);
          }).then((_d) => {
            doc._id.should.equal(_d._id);
            return s.remove(_d);
          });
        });
      });

      /**
       * @name update:
       */
      describe('update', () => {

        /**
         * @name with _id:
         */
        it('with _id', () => {

          return s.store({ _id: 'i4' }).then((_doc) => {
            check_document(_doc, 'i4');
            return s.store(_.assignIn(_doc, { updated: true }));
          }).then((_doc) => {
            _doc.updated.should.be.true;
            check_document(_doc, 'i4');
            return s.remove(_doc);
          });
        });

        /**
         * @name without _id:
         */
        it('without _id', () => {

          return s.store({ test: 1 }).then((_doc) => {
            check_document(_doc);
            return s.store(_.assignIn(_doc, { test: 2 }));
          }).then((_doc) => {
            _doc.test.should.equal(2);
            check_document(_doc);
            return s.store(_.assignIn(_doc, { test: 3 }));
          }).then((_doc) => {
            _doc.test.should.equal(3);
            check_document(_doc);
            return s.remove(_doc);
          });
        });

        /**
         * @name simple conflict:
         */
        it('simple conflict', () => {

          let doc = null;

          return s.store({}).then((_doc) => {
            doc = _doc;
            return s.store(_.omit(_doc, '_rev'));
          }).then(() => {
            should.exist(null); /* Not reached */
          }).catch((_e) => {
            _e.error.should.be.true;
            _e.status.should.equal(409);
            return s.remove(doc);
          });
        });

        /**
         * @name remove:
         */
        describe('remove', () => {

          /**
           * @name normal:
           */
          it('normal', () => {

            return s.store({}).then((_doc) => {
              return s.remove(_doc);
            });
          });

          /**
           * @name forced:
           */
          it('forced', () => {

            return s.store({}).then((_doc) => {
              return s.remove(_doc, { force: true });
            });
          });
        });
      });

      /**
       * @name all:
       */
      describe('all', () => {

        /**
         * @name simple:
         */
        it('simple', () => {

          return s.all({ include_docs: true }).then((_r) => {
            _r.should.be.an.object;
            _r.total_rows.should.equal(1); /* Design document */
          });
        });
      });

      /**
       * @name text encoding:
       */
      describe('text encoding', () => {

        /**
         * @name collates properly:
         */
        it('collates properly', () => {

          /* Simple cases */
          s.collate(' ', ' ').should.equal(0);
          s.collate('a', 'z').should.equal(-1);
          s.collate('zz', 'zy').should.equal(1);
          s.collate('mémé', 'meme').should.equal(1);

          /* Not locale-aware */
          s.collate('Č', 'D').should.equal(1);
          s.collate('Č', 'Ć').should.equal(1);
          s.collate('Ł', 'Z').should.equal(1);
        });
      });
    });

    /**
     * @name view:
     *   Tests related to view and query functionality.
     */
    describe('view', () => {

      /**
       * @name direct:
       */
      describe('direct', () => {

        let results = null;
        let i = new indexer.text();

        const o = { include_docs: true };
        const f = [ 'litany.txt', 'mantra.txt' ];

        /**
         * @name before:
         */
        before(() => {
          return fixtures.store_fixtures(s, f).then((_rv) => {
            results = _rv;
          });
        });

        /**
         * @name before:
         */
        after(() => {
          return fixtures.remove_fixtures(s, results);
        });

        /**
         * @name retrieves words:
         */
        it('retrieves words', () => {

          return s.view('words', o).then((_v) => {
            _.each(_v.rows, (_r) => {

              /* Word tuple format:
                  0: zone_id, 1: index_id,
                  2: word, 3: offset, 4: length, 5: ordinal */

              let word = _r.key[1];
              let offset = _r.key[3], ordinal = _r.key[5];

              _r.key[0].should.equal(0);

              word.should.be.a('string');
              _r.doc.data.words[ordinal][0].should.equal(word);

              word.should.equal(
                i.splitter.fold(
                  _r.doc.data.content.substring(
                    offset, offset + word.length
                  )
                )
              );
            });
          });
        });

        /**
         * @name retrieves nodes by zone:
         */
        it('retrieves nodes by zone', () => {

          return s.view('nodes_by_zone', o).then((_v) => {

            let rv = _.map(_v.rows, (_r) => {
              _r.doc.type.should.equal('index');
              _r.value.type.should.equal('node');
              return [
                _r.key[0], _r.key[1], _r.value.tags
              ];
            });

            rv.should.deep.equal([
              [ 0, 'litany.txt', [ 'litany.txt fixture', 'test' ] ],
              [ 0, 'mantra.txt', [ 'mantra.txt fixture', 'test' ] ]
            ]);
          });
        });

        /**
         * @name retrieves nodes by title:
         */
        it('retrieves nodes by title', () => {

          return s.view('nodes_by_title', o).then((_v) => {

            let rv = _.map(_v.rows, (_r) => {
              _r.doc.type.should.equal('index');
              _r.value.type.should.equal('node');
              return [
                _r.key[0], _r.key[1], _r.value.tags
              ];
            });

            rv.should.deep.equal([
              [ 0, 'litany.txt', [ 'litany.txt fixture', 'test' ] ],
              [ 0, 'mantra.txt', [ 'mantra.txt fixture', 'test' ] ]
            ]);
          });
        });

        /**
         * @name retrieves nodes by tag:
         */
        it('retrieves nodes by tag', () => {

          return s.view('nodes_by_tag', o).then((_v) => {

            let rv = _.map(_v.rows, (_r) => {
              _r.doc.type.should.equal('node');
              _r.value.type.should.equal('index');
              _r.key[3].should.be.a('string'); /* Id */
              return [
                _r.key[0], _r.key[1], _r.key[2], _r.value.tag
              ];
            });

            rv.should.deep.equal([
              [ 0, 'litany.txt', 0,
                [ 'litany.txt fixture', [
                  [ 'litany.txt', 0, 10, 0 ], [ 'fixture', 11, 7, 1 ] ] ] ],
              [ 0, 'mantra.txt', 0,
                [ 'mantra.txt fixture', [
                  [ 'mantra.txt', 0, 10, 0 ], [ 'fixture', 11, 7, 1 ] ] ] ],
              [ 0, 'test', 1,
                [ 'test', [ [ 'test', 0, 4, 0 ] ] ] ],
              [ 0, 'test', 1,
                [ 'test', [ [ 'test', 0, 4, 0 ] ] ] ]
            ]);
          });
        });
      });
    });
  });
});

