
'use strict';

import chai from 'chai';
import _ from '../utils/lodash';
import promises from 'chai-as-promised';
import transform from '../../../src/schema/transform';
import words_transform from '../../../src/schema/transform/words';

/**
 * @name transform:
 *   Unit tests for the transformmodule.
 */
describe('transform', () => {

  let should = chai.should();
  chai.use(promises);

  /**
   * @name simple:
   */
  describe('simple', () => {

    /**
     * @name maps an empty set of documents:
     */
    it('maps an empty set of documents', () => {

      let t = new transform({
        map: (_fn, _d) => _fn(_d)
      });

      t.map([]).length.should.equal(0);
    });

    /**
     * @name maps a simple set of documents:
     */
    it('maps a simple set of documents', () => {

      let t = new transform({
        map: (_fn, _d) => _fn(_d.memes, { _id: _d._id })
      });

      t.map([
        { a: 0 },
        { memes: 'dank', b: 1 },
        { memes: 'memes', c: 2 },
        { _id: 'memes', memes: 'dank' },
        { _id: 'bepis', memes: 'memes', d: null }
      ]).should.deep.equal([
        { id: undefined, key: undefined, value: { _id: undefined } },
        { id: undefined, key: 'dank', value:  { _id: undefined } },
        { id: undefined, key: 'memes', value:  { _id: undefined } },
        { id: 'memes', key: 'dank', value:  { _id: 'memes' } },
        { id: 'bepis', key: 'memes', value:  { _id: 'bepis' } }
      ]);
    });

    /**
     * @name allows mapped values to be omitted:
     */
    it('allows mapped values to be omitted', () => {

      let t = new transform({
        map: (_fn, _d) => _fn(_d.memes)
      });

      t.map([
        { a: 0 },
        { memes: 'dank', b: 1 },
        { memes: 'memes', c: 2 },
        { _id: 'memes', memes: 'dank' },
        { _id: 'bepis', memes: 'memes', d: null }
      ]).should.deep.equal([
        { id: undefined, key: undefined, value: undefined },
        { id: undefined, key: 'dank', value: undefined },
        { id: undefined, key: 'memes', value: undefined },
        { id: undefined, key: 'dank', value: undefined },
        { id: undefined, key: 'memes', value: undefined }
      ]);
    });

    /**
     * @name maps documents one-to-many:
     */
    it('maps documents one-to-many', () => {

      let t = new transform({
        map: (_fn, _d) => {
          _fn({ dank: _d.memes }, 1);
          _fn({ memes: _d.dank }, 2);
        }
      });

      t.map([
        { dank: 'memes' },
        { memes: 'dank' }
      ]).should.deep.equal([
        { id: undefined, key: { dank: undefined }, value: 1 },
        { id: undefined, key: { memes: 'memes' }, value: 2 },
        { id: undefined, key: { dank: 'dank' }, value: 1 },
        { id: undefined, key: { memes: undefined }, value: 2 }
      ]);
    });

    /**
     * @name sorts results:
     */
    it('sorts results', () => {

      let o = { sort: true };

      let t = new transform({
        map: (_fn, _d) => {
          _fn([ 2, _d.memes ], 1);
          _fn([ 1, _d.memes ], 2);
        }
      }, o);

      t.map([
        { memes: 'a' },
        { memes: 'b' }
      ]).should.deep.equal([
        { id: undefined, key: [ 1, 'a' ], value: 2 },
        { id: undefined, key: [ 1, 'b' ], value: 2 },
        { id: undefined, key: [ 2, 'a' ], value: 1 },
        { id: undefined, key: [ 2, 'b' ], value: 1 }
      ]);
    });

    /**
     * @name sorts results with custom collator:
     */
    it('sorts results with custom collator', () => {

      let collator = {
        collate: (_l, _r) => {
          let i = (_l[0] == _r[0] ? 1 : 0);
          return (_l[i] <= _r[i] ? 1 : -1);
        }
      };

      let o = {
        sort: true,
        collator: collator
      };

      let t = new transform({
        map: (_fn, _d) => {
          _fn([ 1, _d.memes ], 2);
          _fn([ 2, _d.memes ], 1);
        }
      }, o);

      t.map([
        { memes: 'b' },
        { memes: 'a' }
      ]).should.deep.equal([
        { id: undefined, key: [ 2, 'b' ], value: 1 },
        { id: undefined, key: [ 2, 'a' ], value: 1 },
        { id: undefined, key: [ 1, 'b' ], value: 2 },
        { id: undefined, key: [ 1, 'a' ], value: 2 }
      ]);
    });

    /**
     * @name maps words:
     */
    it('maps words', () => {

      let t = new transform(words_transform);

      t.map([
        { type: 'other' },
        { type: 'index',
          _id: 1, _rev: 'memes',
          node_id: 'dank', zone_id: 'memes',
          data: {
            title: 'Meme',
            words: [
              [ 0, 'dank', 'memes', 5, 5, 1 ],
              [ 1, 'memes', 'dank', 0, 4, 0 ]
            ]
          }
        },
        { type: 'other' },
        { type: 'index',
          _id: 2, _rev: 'dank',
          node_id: 'memes', zone_id: 'dank',
          data: {
            title: 'Dank',
            words: [
              [ 0, 'memes', 'dank', 0, 4, 0 ],
              [ 1, 'dank', 'memes', 5, 5, 1 ]
            ]
          }
        },
        { type: 'other' }
      ]).should.deep.equal([
        { id: 1,
          key: [ 'memes', 0, 'dank', 'dank', 'memes', 5 ],
          value: { type: 'index', _id: 1, _rev: 'memes', title: 'Meme'} },
        { id: 1,
          key: [ 'memes', 1, 'dank', 'memes', 'dank', 0 ],
          value: { type: 'index', _id: 1, _rev: 'memes', title: 'Meme'} },
        { id: 2,
          key: [ 'dank', 0, 'memes', 'memes', 'dank', 0 ],
          value: { type: 'index', _id: 2, _rev: 'dank', title: 'Dank' } },
        { id: 2,
          key: [ 'dank', 1, 'memes', 'dank', 'memes', 5 ],
          value: { type: 'index', _id: 2, _rev: 'dank', title: 'Dank'} }
      ]);
    });

    /**
     * @name includes docs:
     */
    it('includes docs', () => {

      let o = { include_docs: true };

      let t = new transform({
        map: (_fn, _d) => {
          _fn(0, { memes: 'dank', _id: _d._id });
          _fn(1, { memes: 'very dank', _id: _d._id });
        }
      }, o);

      t.map([
        { _id: 2 },
        { _id: 3 }
      ]).should.deep.equal([
        { id: 2, key: 0,
          value: { _id: 2, memes: 'dank', doc: { _id: 2 } } },
        { id: 2, key: 1,
          value: { _id: 2, memes: 'very dank', doc: { _id: 2 } } },
        { id: 3, key: 0,
          value: { _id: 3, memes: 'dank', doc: { _id: 3 } } },
        { id: 3, key: 1,
          value: { _id: 3, memes: 'very dank', doc: { _id: 3 } } }
      ]);
    });
  });

});

