
'use strict';

import chai from 'chai';
import _ from '../utils/lodash';
import splitter from '../../../src/splitter';
import exception from '../../../src/exception';
import fixtures from '../../../output/fixtures/all';

/**
 * @name text:
 *   Unit tests for the Unicode word splitter.
 */
describe('text', () => {

  let should = chai.should();

  let s = new splitter.text();
  let sf = new splitter.text(null, { fold: true });
  let sa = new splitter.text(null, { autoreset: false });

  /**
   * @name split:
   */
  describe('split', () => {

    /**
     * @name splits ASCII text:
     */
    it('splits ASCII text', () => {

      let r = s.split('Hello, world!').result;
      s.verify().ok.should.be.true;

      r.should.deep.equal([
        [ 'Hello', 0, 5, 0 ], [ 'world', 7, 5, 1 ]
      ]);
    });

    /**
     * @name splits Unicode text:
     */
    it('splits Unicode text', () => {

      let r = s.split('مرحبا. كيف حالك اليوم؟').result;

      r.length.should.equal(4);
      s.verify().ok.should.be.true;

      r.should.deep.equal([
        [ 'مرحبا', 0, 5, 0 ], [ 'كيف', 7, 3, 1 ],
        [ 'حالك', 11, 4, 2 ], [ 'اليوم', 16, 5, 3 ]
      ]);
    });

    /**
     * @name splits Emoji text:
     */
    it('splits Emoji text', () => {

      let r = s.split('👌👌👌 💯💯 😂').result;

      r.length.should.equal(3);
      s.verify().ok.should.be.true;

      r.should.deep.equal([
        [ '👌👌👌', 0, 6, 0 ],
        [ '💯💯', 7, 4, 1 ], [ '😂', 12, 2, 2 ]
      ]);
    });

    /**
     * @name splits a single word:
     */
    it('splits a single word', () => {

      let r = s.split('memes').result;

      r.length.should.equal(1);
      s.verify().ok.should.be.true;

      r.should.deep.equal([
        [ 'memes', 0, 5, 0 ]
      ]);
    });

    /**
     * @name does not split internal punctuation:
     */
    it('does not split internal punctuation', () => {

      let r = s.split('meme me.me me-me me\'me').result;

      r.length.should.equal(4);
      s.verify().ok.should.be.true;

      r.should.deep.equal([
        [ 'meme', 0, 4, 0 ],
        [ 'me.me', 5, 5, 1 ],
        [ 'me-me', 11, 5, 2 ],
        [ 'me\'me', 17, 5, 3 ]
      ]);
    });

    /**
     * @name ignores external punctuation:
     */
    it('ignores external punctuation', () => {

      let r = s.split('"meme" .meme. !me.me! \'me-me\'').result;

      r.length.should.equal(4);
      s.verify().ok.should.be.true;

      r.should.deep.equal([
        [ 'meme', 1, 4, 0 ],
        [ 'meme', 8, 4, 1 ],
        [ 'me.me', 15, 5, 2 ],
        [ 'me-me', 23, 5, 3 ]
      ]);
    });

    /**
     * @name ignores standalone punctuation:
     */
    it('ignores standalone punctuation', () => {

      let r = s.split('meme - meme .  meme ! meme').result;

      r.length.should.equal(4);
      s.verify().ok.should.be.true;

      r.should.deep.equal([
        [ 'meme', 0, 4, 0 ],
        [ 'meme', 7, 4, 1 ],
        [ 'meme', 15, 4, 2 ],
        [ 'meme', 22, 4, 3 ]
      ]);
    });

    /**
     * @name handles all-punctuation:
     */
    it('handles all-punctuation', () => {

      let r = s.split('(!@#$%^&*)').result;

      r.length.should.equal(0);
      s.verify().ok.should.be.true;
      r.should.deep.equal([]);
    });

    /**
     * @name handles preceeding and trailing punctuation:
     */
    it('handles preceeding and trailing punctuation', () => {

      let r = s.split(
        '|}{":?><+_)(*&^%$#@!memes|}{":?><+_)(*&^%$#@!'
      ).result;

      r.length.should.equal(1);
      s.verify().ok.should.be.true;

      r.should.deep.equal([
        [ 'memes', 20, 5, 0 ]
      ]);
    });

    /**
     * @name splits fitzpatrick modifiers:
     */
    it('splits fitzpatrick modifiers', () => {

      let r = s.split('✊🏿✊🏾✊🏽 memes ✊🏼✊🏻✊').result;

      r.length.should.equal(3);
      s.verify().ok.should.be.true;

      r.should.deep.equal([
        [ '✊🏿✊🏾✊🏽', 0, 9, 0 ],
        [ 'memes', 10, 5, 1 ], [ '✊🏼✊🏻✊', 16, 7, 2 ]
      ]);
    });

    /**
     * @name splits an example corpus:
     */
    it('splits an example corpus', () => {

      const f = 'lorem-emoji.txt';
      let r = sf.split(fixtures.unit.text[f]).result;

      s.split(fixtures.unit.text[f]);
      s.verify().ok.should.be.true;

      r.should.deep.include.members([
        [ '☁️⛅️⛈', 215, 5, 31 ],
        [ '✍🏾✍🏾', 661, 6, 96 ],
        [ '👂🏿👂🏽👂🏻', 1074, 12, 157 ],
        [ '👋🏼👋🏽👋🏾👋🏿', 1849, 16, 265 ],
        [ '⛄️', 2373, 2, 338 ],
        [ '👊🏿👊🏾👊🏽👊🏼', 2603, 16, 373 ]
      ]);
    });

    /**
     * @name splits unicode spaces:
     */
    it('splits unicode spaces', () => {

      let c = [
        0x0020, 0x00a0, 0x1680, 0x2000, 0x2001,
        0x2002, 0x2003, 0x2004, 0x2005, 0x2006, 0x2007,
        0x2008, 0x2009, 0x200a, 0x200b, 0x202f, 0x205f, 0x3000
      ];

      let str = _.reduce(c, (_left, _n) => {
        return (
          _left + '0x' +
            _.padStart(_n.toString(16), 4, '0') +
            String.fromCharCode(_n) + 'a '
        );
      }, '');

      let r = sf.split(str).result;
      sf.verify().ok.should.be.true;

      _.each(_.chunk(r, 2), (_pair) => {
        _pair[1][0].should.equal('a');
      });
    });

    /**
     * @name rejects non-strings:
     */
    it('rejects non-strings', () => {

      let e = exception.argument;

      (() => { s.split(1); }).should.throw(e);
      (() => { s.split({}); }).should.throw(e);
      (() => { s.split([]); }).should.throw(e);
      (() => { s.reset().split(); }).should.throw(e);
      (() => { s.string = undefined; }).should.throw(e);
      (() => { s.reset().split(null); }).should.throw(e);
      (() => { s.reset().split(undefined); }).should.throw(e);
      (() => { s.string = 'a'; s.split(); }).should.not.throw();
      (() => { s.string = 'a'; s.split(undefined); }).should.not.throw();
    });

    /**
     * @name correctly handles joining characters:
     */
    it('correctly handles joining characters', () => {

      let str = 'षःशँहंक्ष्त्रृज्ञायीऋि७ॠेऌैश्रोॐश्रौ';
      let r = s.split(str).result;

      r.length.should.equal(1);
      r[0].should.deep.equal([ str, 0, 36, 0 ]);
    });
  });

  /**
   * @name verify:
   */
  describe('verify', () => {

    /**
     * @name refuses to verify an empty result:
     */
    it('refuses to verify an empty result', () => {

      s.reset().string = 'hello';
      s.verify().ok.should.be.false;
      s.reset().verify().ok.should.be.false;
    });

    /**
     * @name identifies incorrect string:
     */
    it('identifies incorrect string', () => {

      let r = s.split('hello, world').result;
      s.verify().ok.should.be.true;

      s.string = 'goodbye, world';
      s.verify().ok.should.be.false;
    });

    /**
     * @name identifies incorrect result:
     */
    it('identifies incorrect result', () => {

      const f = 'lorem-emoji.txt';
      let r = s.split(fixtures.unit.text[f]).result;

      r[0] = [ 0, 'Incorrect', 0 ];
      s.verify().ok.should.be.false;
    });
  });

  /**
   * @name retains a string correctly:
   */
  it('retains a string correctly', () => {

    s.split('memes!');
    s.verify().ok.should.be.true;
    s.string.should.equal('memes!');
    s.reset().string = 'meems!';
    s.string.should.equal('meems!');
  });

  /**
   * @name clears string and results correctly:
   */
  it('clears string and results correctly', () => {

    s.split('memes memes memes').reset();
    should.equal(s.string, undefined);
    should.equal(s.result, undefined);
  });

  /**
   * @name rejects non-strings in its constructor:
   */
  it('rejects non-strings in its constructor', () => {

    let e = exception.argument;

    (() => { new splitter.text(1); }).should.throw(e);
    (() => { new splitter.text(true); }).should.throw(e);
    (() => { new splitter.text(false); }).should.throw(e);
    (() => { new splitter.text(null); }).should.not.throw();
    (() => { new splitter.text(undefined); }).should.not.throw();
  });

  /**
   * @name folds case correctly:
   */
  it('folds case correctly', () => {

    let r = sf.split('Hello, World!').result;

    r.should.deep.equal([
      ['hello', 0, 5, 0 ], [ 'world', 7, 5, 1 ]
    ]);

    let o = s.options;
    s.options = { fold: true };

    r = s.split('Hello, World!').result;

    r.should.deep.equal([
      ['hello', 0, 5, 0 ], [ 'world', 7, 5, 1 ]
    ]);

    s.options = o;

    r = s.split('Hello, World!').result;

    r.should.deep.equal([
      ['Hello', 0, 5, 0 ], [ 'World', 7, 5, 1 ]
    ]);

    r = sf.split('ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞ').result;

    r.should.deep.equal([
      [ 'àáâãäåæçèéêëìíîïðñòóôõöøùúûüýþ', 0, 30, 0 ]
    ]);
  });

  /**
   * @name stems text correctly:
   */
  it('stems text correctly', () => {

    let o = s.options;
    s.options = { stem: true };

    s.split('Greetings, worlds!');
    s.verify().ok.should.be.true;

    s.options = { stem: 'en' };
    let r = s.split('Greetings, worlds!').result;
    s.verify().ok.should.be.true;

    s.options = o;

    r.should.deep.equal([
      [ 'Greetings', 0, 9, 0, { stem: 'greet' } ],
      [ 'worlds', 11, 6, 1, { stem: 'world' } ]
    ]);
  });

  /**
   * @name rejects unknown stemming languages:
   */
  it('rejects unknown stemming languages', () => {

    (() => {
      s.stem('memes', 'dank');
    }).should.throw(exception.argument);
  });

  /**
   * @name splits with autoreset disabled:
   */
  it('splits with autoreset disabled', () => {

    let r = sa.split('Hello, world!').result;
    s.verify().ok.should.be.true;

    r.should.deep.equal([
      [ 'Hello', 0, 5, 0 ], [ 'world', 7, 5, 1 ]
    ]);

    r = sa.split('Hello, world.').result;
    s.verify().ok.should.be.true;

    r.should.deep.equal([
      [ 'Hello', 0, 5, 2 ], [ 'world', 7, 5, 3 ]
    ]);
  });
});

