
'use strict';

import chai from 'chai';
import _ from './utils/lodash';
import utils from '../../src/utils';
import exception from '../../src/exception';

/**
 * @name utils:
 *   Unit tests for the utilities library.
 */
describe('utils', () => {

  let should = chai.should();

  /**
   * @name search.binary:
   */
  describe('search.binary', () => {

    let s = utils.search;

    let map = (_r) => _r.key;
    let cmp = (_l, _r) => (_l - _r);

    let a = [ 1, 2, 3, 5, 8, 13, 21, 34, 55 ];
    let b = _.map(a, (_n) => { return { key: _n }; });

    /**
     * @name handles empty input:
     */
    it('handles empty input', () => {

      s.binary([], 42, cmp).should.equal(false);
    });

    /**
     * @name finds simple items correctly:
     */
    it('finds simple items correctly', () => {

      s.binary(a, 0, cmp).should.equal(~0);
      s.binary(a, 1, cmp).should.equal(0);
      s.binary(a, 2, cmp).should.equal(1);
      s.binary(a, 3, cmp).should.equal(2);
      s.binary(a, 4, cmp).should.equal(~3);
      s.binary(a, 5, cmp).should.equal(3);
    });

    /**
     * @name finds mapped items correctly:
     */
    it('finds mapped items correctly', () => {

      s.binary(b, 12, cmp, map).should.equal(~5);
      s.binary(b, 13, cmp, map).should.equal(5);
      s.binary(b, 21, cmp, map).should.equal(6);
      s.binary(b, 34, cmp, map).should.equal(7);
      s.binary(b, 40, cmp, map).should.equal(~8);
      s.binary(b, 55, cmp, map).should.equal(8);
      s.binary(b, 56, cmp, map).should.equal(~9);
    });

    /**
     * @name finds mapped items within bounds:
     */
    it('finds mapped items within bounds', () => {

      s.binary(b, 1, cmp, map, 2, 6).should.equal(~2);
      s.binary(b, 2, cmp, map, 2, 6).should.equal(~2);
      s.binary(b, 3, cmp, map, 2, 6).should.equal(2);
      s.binary(b, 20, cmp, map, 2, 6).should.equal(~6);
      s.binary(b, 21, cmp, map, 2, 6).should.equal(6);
      s.binary(b, 34, cmp, map, 2, 6).should.equal(~7);
      s.binary(b, 55, cmp, map, 2, 6).should.equal(~7);
      s.binary(b, 56, cmp, map, 2, 6).should.equal(~7);
    });

    /**
     * @name throws on invalid bounds:
     */
    it('throws on invalid bounds', () => {

      let fn = (_n, _l, _r) => {
        return s.binary(b, _n, cmp, map, _l, _r);
      };

      (() => { fn(1, -1); }).should.throw(exception.range);
      (() => { fn(2, 1, 10); }).should.throw(exception.range);
      (() => { fn(3, -1, 8); }).should.throw(exception.range);
      (() => { fn(5, 1, 0); }).should.throw(exception.range);
      (() => { fn(8, -10, 100); }).should.throw(exception.range);
    });

    /**
     * @name returns false for empty array:
     */
    it('returns false for empty array', () => {

      s.binary([], 1, cmp).should.be.false;
    });
  });

  /**
   * @name uuid:
   */
  describe('uuid', () => {

    /**
     * @name is monotonically increasing:
     */
    it('is monotonically increasing', () => {

      let uuid = '';
      const iterations = 1000;

      for (var i = 0; i < iterations; ++i) {
        let u = utils.uuid.compact();
        uuid.should.be.below(u);
        uuid = u;
      }
    });
  });

  /**
   * @name is_subset:
   */
  describe('is_subset', () => {

    /**
     * @name handles scalar values:
     */
    it('handles scalar values', () => {

      utils.is_subset(1, 1).should.be.true;
      utils.is_subset(2, 1).should.be.false;
      utils.is_subset('dank', 'dank').should.be.true;
      utils.is_subset('dank', 'memes').should.be.false;
      utils.is_subset('memememe', 'meme').should.be.false;
    });

    /**
     * @name handles arrays:
     */
    it('handles arrays', () => {

      utils.is_subset([], []).should.be.true;
      utils.is_subset([ 1 ], [ 1 ]).should.be.true;
      utils.is_subset([ 1, 2, 3, 4 ], [ 1, 2 ]).should.be.true;
      utils.is_subset([ 1, 2, 3, 4 ], [ 2, 1 ]).should.be.false;
      utils.is_subset([ 1, 2, 3, 4 ], [ 1, 2, null ]).should.be.false;
      utils.is_subset([ 1, 2, 3, 4 ], [ 1, 2, 3, 4, 5 ]).should.be.false;
    });

    /**
     * @name handles objects:
     */
    it('handles objects', () => {

      utils.is_subset({}, {}).should.be.true;
      utils.is_subset({ memes: 1 }, { memes: 1 }).should.be.true;
      utils.is_subset({ memes: 1 }, { memes: 2 }).should.be.false;
      utils.is_subset({ memes: 1, dank: 2 }, { memes: 1 }).should.be.true;
      utils.is_subset({ memes: 1 }, { memes: 1, dank: 2 }).should.be.false;
      utils.is_subset({ memes: 1 }, { memes: null }).should.be.false;
      utils.is_subset({ memes: false }, { memes: false }).should.be.true;
    });

    /**
     * @name handles nested structures:
     */
    it('handles nested structures', () => {

      utils.is_subset(
        [ 1, 2, { memes: { dank: null, get: 'out' } }, 3, 4 ],
        [ 1, 2, { memes: { dank: null } } ]
      ).should.be.true;

      utils.is_subset(
        { memes: { dank: true, list: [ 1, 2, 3, 4 ] }, ok: 1 },
        { memes: { list: [ ] } }
      ).should.be.true;

      utils.is_subset(
        { memes: { dank: false, list: [ 1, 2, 3, 4 ] }, ok: 1 },
        { memes: { dank: false, list: [ 1 ] } }
      ).should.be.true;

      utils.is_subset(
        [ 1, 2, { memes: { dank: true, get: 'out' } }, 3, 4 ],
        [ 1, 2, { meme: { dank: true } } ]
      ).should.be.false;

      utils.is_subset(
        { memes: { dank: true, list: [ 1, 2, 3, 4 ] }, ok: 1 },
        { memes: { list: [ 2, 3, 4 ] } }
      ).should.be.false;

      utils.is_subset(
        { memes: { dank: null, list: [ false, { a: 2 } ] }, ok: 1 },
        { memes: { dank: false, list: [ null, { a: 2 } ] } }
      ).should.be.false;
    });

    /**
     * @name limits recursion:
     */
    it('limits recursion', () => {

      var array = [];
      var object = {};

      array.push(array);
      object.circular = object;

      (() => {
        utils.is_subset(array, array);
      }).should.throw(exception.argument);

      (() => {
        utils.is_subset(object, object);
      }).should.throw(exception.argument);

      (() => {
        utils.is_subset(object, array);
      }).should.not.throw;

      (() => {
        utils.is_subset(array, object);
      }).should.not.throw;
    });
  });

  /**
   * @name make_promise:
   */
  describe('make_promise', () => {

    /**
     * @name allows zero arguments:
     */
    it('allows zero arguments', () => {

      return utils.make_promise().then((_nil) => {
        should.not.exist(_nil);
      });
    });

    /**
     * @name creates a promise from a scalar:
     */
    it('creates a promise from a scalar', () => {

      return utils.make_promise(123).then((_n) => {
        _n.should.equal(123);
      });
    });

    /**
     * @name creates a promise from a function:
     */
    it('creates a promise from a function', () => {

      return utils.make_promise((_a, _b, _c) => {
        return (_c + 1);
      }, 1, 2, 3).then((_n) => {
        _n.should.equal(4);
      });
    });
  });

  /**
   * @name paths_for:
   */
  describe('paths_for', () => {

    /**
     * @name enumerates simple object paths:
     */
    it('enumerates simple object paths', () => {

      utils.paths_for({ foo: 1 })
        .should.deep.equal({ foo: true });

      utils.paths_for({ memes: true })
        .should.deep.equal({ memes: true });

      utils.paths_for({ bar: 2, baz: 3 })
        .should.deep.equal({ bar: true, baz: true });
    });

    /**
     * @name enumerates nested object paths:
     */
    it('enumerates nested object paths', () => {

      utils.paths_for({ foo: { bar: 1 } })
        .should.deep.equal({ 'foo.bar': true });

      utils.paths_for({ foo: { bar: { baz: 1 } } })
        .should.deep.equal({ 'foo.bar.baz': true });

      utils.paths_for({ foo: 1, baz: { quux: [ { zzyz: 1 } ] } })
        .should.deep.equal({ 'foo': true, 'baz.quux[0].zzyz': true });

      utils.paths_for({ baz: { quux: [ 1, {} ] } })
        .should.deep.equal({ 'baz.quux[0]': true, 'baz.quux[1]': true });
    });

    /**
     * @name handles empty objects:
     */
    it('handles empty objects', () => {

      utils.paths_for({})
        .should.deep.equal({});
    });

    /**
     * @name limits recursion:
     */
    it('limits recursion', () => {

      var array = [];
      var object = {};

      array.push(array);
      object.circular = object;

      (() => {
        utils.paths_for(array);
      }).should.throw(exception.argument);

      (() => {
        utils.paths_for(object);
      }).should.throw(exception.argument);
    });
  });

  /**
   * @name logical:
   */
  describe('logical', () => {

    /**
     * @name or:
     */
    describe('or', () => {

      /**
       * @name passes arguments:
       */
      it('passes arguments', () => {

        utils.logical.or((..._args) => {
          _args.should.have.lengthOf(5);
          _.each(_args, (_n, _i) => _n.should.equal(_i + 1));
        })(1, 2, 3, 4, 5);
      });

      /**
       * @name returns a function:
       */
      it('returns a function', () => {

        utils.logical.or().should.be.a('function');
      });

      /**
       * @name returns false for an empty set:
       */
      it('returns false for an empty set', () => {

        utils.logical.or()().should.be.false;
      });

      /**
       * @name computes a false result:
       */
      it('computes a false result', () => {

        utils.logical.or(() => false, () => false)().should.be.false;
      });

      /**
       * @name computes a true result:
       */
      it('computes a true result', () => {

        utils.logical.or(() => false, () => true)().should.be.true;
      });

      /**
       * @name short circuits:
       */
      it('short circuits', () => {

        utils.logical.or(
          () => false , () => true, () => { should.be.true(false); }
        )().should.be.true;
      });
    });

    /**
     * @name and:
     */
    describe('and', () => {

      /**
       * @name passes arguments:
       */
      it('passes arguments', () => {

        utils.logical.and((..._args) => {
          _args.should.have.lengthOf(6);
          _.each(_args, (_n, _i) => _n.should.equal(_i * 10));
        })(0, 10, 20, 30, 40, 50);
      });

      /**
       * @name returns a function:
       */
      it('returns a function', () => {

        utils.logical.and().should.be.a('function');
      });

      /**
       * @name returns true for an empty set:
       */
      it('returns true for an empty set', () => {

        utils.logical.and()().should.be.true;
      });

      /**
       * @name computes a false result:
       */
      it('computes a false result', () => {

        utils.logical.and(() => true, () => false)().should.be.false;
      });

      /**
       * @name computes a true result:
       */
      it('computes a true result', () => {

        utils.logical.and(() => true, () => true)().should.be.true;
      });

      /**
       * @name short circuits:
       */
      it('short circuits', () => {

        utils.logical.and(
          () => true, () => false, () => { should.be.true(false); }
        )().should.be.false;
      });

    });
  });
});

