
'use strict';

import chai from 'chai';
import _ from './utils/lodash';
import utils from '../../src/utils';
import validator from '../../src/validator';

/**
 * @name validator:
 *   Unit tests for the schema/constraint validation library.
 */
describe('validator', () => {

  let v = new validator();
  let should = chai.should();

  /**
   * @name check:
   *   Check the results of a validation operation. The `_r` argument
   *   should be a return value from `validate`. The `_pass` argument
   *   should be true if success is expected, or false otherwise. The
   *   `_key` argument is the property to be validated, `_i` is the
   *   function offset to check in the selected property's subresult
   *   object, and `_constraint_pass` indicates whether or not the
   *   subresult object is expected to validate cleanly.
   */
  let check = (_r, _pass, _key, _i, _constraint_pass) => {

    /* Optional */
    if (_.isNil(_constraint_pass)) {
      _constraint_pass = _pass;
    }

    /* Check top-level result */
    _r.ok.should.equal(_pass);
    _r.success.should.be.an('object');
    _r.failure.should.be.an('object');

    if (_pass) {
      _.keys(_r.failure).should.have.length.of(0);
      _.keys(_r.success).should.have.length.above(0);
    } else {
      _.keys(_r.failure).should.have.length.above(0);
    }

    /* Check detailed constraint result */
    let d = (_pass ? _r.success : _r.failure)[_key];

    d.should.be.an('object');
    d.ok.should.equal(_pass);

    if (_constraint_pass) {
      d.success.should.be.an('array');
      d.success[_i].should.be.an('object');
      d.success[_i].ok.should.be.true;
    } else {
      if (!d.internal) {
        d.failure.should.be.an('array');
        d.failure[_i].should.be.an('object');
        d.failure[_i].ok.should.be.false;
      }
    }
  };

  /**
   * @name simple:
   *   A simplified test. Create a schema with `_constraint` for
   *   the property named `_key`, validate `_object`, then call
   *   `check` using the remaining arguments.
   */
  let simple = (_key, _constraint, _object, _pass) => {

    let schema = {};
    schema[_key] = _constraint;

    let r = v.validate(schema, _object, false);
    check(r, _pass, _key, 0);
  };

  /**
   * @name evaluates simple passing constraints:
   */
  it('evaluates simple passing constraints', () => {

    simple('foo', _.isInteger, { foo: 1 }, true);
    simple('foo', _.isString, { foo: 'bar' }, true);
    simple('foo', _.isPlainObject, { foo: {} }, true);
  });

  /**
   * @name evaluates nested passing constraints:
   */
  it('evaluates nested passing constraints', () => {

    simple('foo.bar', _.isNumber, { foo: { bar: 4.2 } }, true);
    simple('foo.bar[0]', _.isString, { foo: { bar: [ 'a' ] } }, true);
    simple('foo.bar[0]', _.isNil, { foo: { bar: [ null ] } }, true);
  });

  /**
   * @name evaluates simple failing constraints:
   */
  it('evaluates simple failing constraints', () => {

    simple('foo', _.isInteger, { foo: 'a' }, false);
    simple('foo', _.isString, { foo: 12345 }, false);
    simple('foo', _.isPlainObject, { foo: new Date() }, false);
  });

  /**
   * @name evaluates nested failing constraints:
   */
  it('evaluates nested failing constraints', () => {

    simple('foo.bar', _.isNumber, { foo: { bar: {} } }, false);
    simple('foo.bar[0]', _.isString, { foo: { bar: [ [] ] } }, false);
    simple('foo.bar[1]', _.isNil, { foo: { bar: [ true, 1 ] } }, false);
  });

  /**
   * @name evaluates multiple passing constraints:
   */
  it('evaluates multiple passing constraints', () => {

    let r = v.validate(
      { 'foo': _.isString, 'baz': _.isNumber },
      { foo: 'bar', baz: 123 }
    );

    check(r, true, 'foo', 0);
    check(r, true, 'baz', 0);
  });

  /**
   * @name evaluates multiple functions in a passing constraint:
   */
  it('evaluates multiple functions in a passing constraint', () => {

    let r = v.validate(
      { 'foo': [ _.isObject, _.isArray ] }, { foo: [] }
    );

    check(r, true, 'foo', 0);
    check(r, true, 'foo', 1);
  });

  /**
   * @name evaluates multiple functions in a failing constraint:
   */
  it('evaluates multiple functions in a failing constraint', () => {

    let r = v.validate(
      { 'foo': [ _.isPlainObject, _.isArray ] }, { foo: [], bar: 1 }
    );

    check(r, false, 'foo', 0, true);
    check(r, false, 'foo', 0, false);
  });

  /**
   * @name allows null value with optional constraint:
   */
  it('allows null value with optional constraint', () => {

    let r = v.validate(
      { foo: [ v.optional, _.isObject, _.isArray ] }, { foo: null }
    );

    check(r, true, 'foo', 0, true);

    r.success.foo.failure.should.have.length.of(0);
    r.success.foo.success.should.have.length.of(1);
    r.success.foo.success[0].short_circuit.should.be.true;
  });

  /**
   * @name allows non-null value with optional constraint:
   */
  it('allows non-null value with optional constraint', () => {

    let r = v.validate(
      { foo: [ v.optional, _.isObject, _.isArray ] }, { foo: [] }
    );

    check(r, true, 'foo', 0, true);

    r.success.foo.failure.should.have.length.of(0);
    r.success.foo.success.should.have.length.of(3);
    should.not.exist(r.success.foo.success[0].short_circuit);
  });

  /**
   * @name rejects extraneous properties:
   */
  it('rejects extraneous properties', () => {

    let r = v.validate(
      { foo: _.isString, baz: _.isNumber },
      { foo: 'bar', baz: 123, extraneous: { value: 1 } }, false
    );

    r.ok.should.be.false;
    r.unexpected.should.be.an('object');
    r.unexpected['extraneous.value'].should.be.true;

    r.success.foo.failure.should.have.length.of(0);
    r.success.foo.success.should.have.length.of(1);
    r.success.baz.failure.should.have.length.of(0);
    r.success.baz.success.should.have.length.of(1);
  });

  /**
   * @name does not treat optional array members as extraneous:
   */
  it('does not treat optional array members as extraneous', () => {

    let r = v.validate(
      { foo: [ v.optional, _.isArray ], bar: [ v.optional, _.isArray ] },
      { foo: [ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ] }, false
    );

    check(r, true, 'foo', 0);
    check(r, true, 'foo', 1);
    check(r, true, 'bar', 0);
  });

  /**
   * @name accepts extraneous properties:
   */
  it('accepts extraneous properties', () => {

    let r = v.validate(
      { 'foo': _.isString, 'baz': _.isNumber },
      { foo: 'bar', baz: 123, extraneous: { value: 2 } }, true
    );

    check(r, true, 'foo', 0);
    check(r, true, 'baz', 0);
  });

  /**
   * @name rejects invalid constraints:
   */
  it('rejects invalid constraints', () => {

    let r = v.validate(
      { foo: 'invalid', baz: [ 'invalid' ] },
      { foo: 'bar', baz: 'quux' }
    );

    check(r, false, 'foo', 0);
    check(r, false, 'baz', 0);

    r.failure.foo.internal.should.be.true;
    r.failure.baz.internal.should.be.true;
  });

  /**
   * @name rejects invalid schemas:
   */
  it('rejects invalid schemas', () => {

    let r = v.validate(
      419, { foo: 'bar' }
    );

    r.ok.should.be.false;
    r.internal.should.be.true;
  });

  /**
   * @name rejects invalid objects:
   */
  it('rejects invalid objects', () => {

    let r = v.validate(
      { foo: _.isString }, 420
    );

    r.ok.should.be.false;
    r.internal.should.be.true;
  });

  /**
   * @name rejects on invalid constraint return values:
   */
  it('rejects on invalid constraint return values', () => {

    let r = v.validate(
      { foo: [ () => 1, _.isInteger ] }, {}
    );

    r.ok.should.be.false;
    r.failure.foo.failure.length.should.equal(1);
    r.failure.foo.failure[0].short_circuit.should.be.true;
  });

  /**
   * @name does not set message for valid fields:
   */
  it('does not set message for valid fields', () => {

    let r = v.validate(
      { foo: [ v.optional, _.isInteger ] }, {}
    );

    r.ok.should.be.true;
    should.exist(r.message);
    should.exist(r.success.foo.message);
    should.not.exist(r.success.message);
    r.message.should.not.have.string('undefined');
  });
});

