
'use strict';

import chai from 'chai';
import _ from './utils/lodash';
import worker from '../../src/worker';
import promise from '../../src/promise';
import promises from 'chai-as-promised';
import test_utils from '../include/utils';
import exception from '../../src/exception';


/**
 * @name test_worker:
 *   A worker subclass for testing purposes.
 */
let test_worker = class extends worker {

  /**
   * @name main:
   */
  main () {

    self.onmessage = (_ev) => {

      if (_ev.data === false) {
        self.postMessage(true);
        return self.close();
      }

      if (_ev.data === true) {
        for (let i = 0; i < 10; ++i) {
          self.postMessage(i);
        }
      }
    };
  }
};

/**
 * @name import_worker:
 *   A worker subclass for testing a small script import.
 */
let import_worker = class extends worker {

  /**
   * @name main:
   */
  main () {

    self.onmessage = (_ev) => {
      self.postMessage(self.test_result);
    };
  }
};

/**
 * @name api_worker:
 *   A worker subclass for testing a full API import.
 */
let api_worker = class extends worker {

  /**
   * @name main:
   */
  main () {

    self.onmessage = (_ev) => {
      self.postMessage(Object.keys(self.api));
      self.close();
    };
  }
};

/**
 * @name args_worker:
 *   A worker subclass that sends an array of its
 *   `run` arguments when any message is received.
 */
let args_worker = class extends worker {

  /**
   * @name main:
   */
  main (..._args) {

    self.onmessage = (_ev) => {
      self.postMessage(_args);
    };
  }
};


/**
 * @name worker:
 *   Unit tests for the worker library.
 */
describe('worker', () => {

  let should = chai.should();
  chai.use(promises);

  let base_dir = test_utils.base_directory();

  /**
   * @name spawns:
   */
  it('spawns', () => {

    let w = new test_worker();

    w.run();
    w.send(false);

    return w.receive().then((_m) => {
      _m.should.be.true;
    })
  });

  /**
   * @name exchanges messages:
   */
  it('exchanges messages', () => {

    let rv = [ [], [] ];
    let w = new test_worker();

    w.receive().then((_m) => rv[0].push(_m));
    w.receive().then((_m) => rv[0].push(_m));

    w.send(null);
    w.run();
    w.send(true);

    return promise.while(() => rv[1].length < 8, () => {
      return w.receive().then((_m) => rv[1].push(_m));
    }).then(() => {
      rv.should.deep.equal([ [ 0, 1 ], [ 2, 3, 4, 5, 6, 7, 8, 9 ] ]);
      w.terminate();
    });
  });

  /**
   * @name receive rejects receive after termination if empty:
   */
  it('receive rejects after termination if empty', () => {

    let w = new test_worker();

    w.terminate().receive().should.eventually.be.rejectedWith(
      exception.runtime
    );
  });

  /**
   * @name send throws after termination:
   */
  it('send throws after termination', () => {

    let w = new test_worker();

    (() => { w.terminate().send(1); }).should.throw(
      exception.argument
    );
  });

  /**
   * @name imports scripts:
   */
  it('imports scripts', () => {

    let w = new import_worker();

    w.run([ `${base_dir}/tests/harness/test.js` ]);
    w.send(true);

    return w.receive().then((_m) => {
      _m.should.equal('ok');
    }).then(() => {
      w.terminate();
    });
  });

  /**
   * @name imports api:
   */
  it('imports api', () => {

    let w = new api_worker();
    w.send(true);

    w.run([
      `${base_dir}/dist/pouchdb.min.js`,
      `${base_dir}/dist/api.js`
    ]);

    return w.receive().then((_m) => {
      _m.should.deep.equal([
        'base', 'engine', 'exception', 'promise', 'proxy',
          'schema', 'unicode', 'utils', 'view', 'worker', 'default'
      ]);
    });
  });

  /**
   * @name reports termination after send:
   */
  it('reports termination after send', () => {

    let w = new test_worker();

    w.run();
    w.send(false);

    return w.receive().then((_m) => {
      (() => w.send(true)).should.throw(exception.argument);
      w.is_running().should.be.false;
    });
  });

  /**
   * @name cancels receive on worker exit:
   */
  it('cancels receive on worker exit', () => {

    let w = new test_worker();

    w.receive().should.eventually.equal(true);
    w.receive().should.eventually.be.rejectedWith(exception.runtime);
    w.receive().should.eventually.be.rejectedWith(exception.runtime);

    w.send(false);
    w.receive().should.eventually.be.rejectedWith(exception.runtime);

    w.run();
    w.receive().should.eventually.be.rejectedWith(exception.runtime);
  });

  /**
   * @name cancels receive on worker termination:
   */
  it('cancels receive on worker termination', () => {

    let w = new test_worker();

    w.receive().should.eventually.be.rejectedWith(exception.runtime);
    w.run().terminate();
  });

  /**
   * @name receives after termination when messages remain:
   */
  it('receives after termination when messages remain', () => {

    let w = new test_worker();

    w.run();
    w.send(false);

    return w.receive().then((_m) => {
      _m.should.be.true;
      w.receive().should.eventually.be.rejectedWith(exception.runtime);
    }).then(() => {
      for (var i = 0; i < 1000; ++i) {
        w.receive().should.eventually.be.rejectedWith(exception.runtime);
      }
    });
  });

});

