
'use strict';

import chai from 'chai';
import _ from '../utils/lodash';
import utils from '../../../src/utils';
import promises from 'chai-as-promised';
import promise from '../../../src/promise';
import test_utils from '../../include/utils';
import proxy from '../../../src/worker/proxy';
import exception from '../../../src/exception';

/**
 * @name empty_proxy:
 *   A Proxy subclass that instantiates an empty object.
 */
let empty_proxy = class extends proxy {

  /** @static: **/
  instantiate () {

    return {};
  }
};

/**
 * @name base_proxy:
 *   A Proxy subclass that instantiates a base object.
 */
let base_proxy = class extends proxy {

  /** @static: **/
  instantiate (_options) {

    return new self.api.base(_options);
  }
};

/**
 * @name promise_proxy:
 *   A Proxy subclass that instantiates a promise object.
 */
let promise_proxy = class extends proxy {

  /** @static: **/
  instantiate (_options) {

    return self.api.promise;
  }
};

/**
 * @name engine_proxy:
 *   A Proxy subclass that instantiates an Engine object.
 */
let engine_proxy = class extends proxy {

  /** @static: **/
  instantiate (_options) {

    return new self.api.engine(_options);
  }
};

/**
 * @name instantiation_error_proxy:
 *   A Proxy subclass that throws during instantiation.
 */
let instantiation_error_proxy = class extends proxy {

  /** @static: **/
  instantiate (_options) {

    throw new Error();
  }
};

/**
 * @name method_error_proxy:
 *   A Proxy subclass that raises errors on method invocation.
 */
let method_error_proxy = class extends proxy {

  /** @static: **/
  instantiate (_options) {

    return Object.create({
      sync: () => {
        throw new Error('sync');
      },
      async: () => {
        return Promise.reject(new Error('async'));
      }
    });
  }
};

/**
 * @name proxy:
 *   Unit tests for the worker proxy library.
 */
describe('proxy', () => {

  let should = chai.should();
  chai.use(promises);

  let engine_options = {
    type: 'pouchdb',
    path: 'output/databases/unit/engine'
  };

  let base_dir = test_utils.base_directory();

  let dependencies = [
    `${base_dir}/node_modules/babel-polyfill/dist/polyfill.js`,
    `${base_dir}/dist/pouchdb.min.js`,
    `${base_dir}/dist/api.js`
  ];

  /**
   * @name throws if not subclassed:
   */
  it('throws if not subclassed', () => {

    let p = new proxy();

    (() => p.run()).should.throw(exception.argument);
  });

  /**
   * @name starts correctly:
   */
  it('starts correctly', () => {

    let p = new empty_proxy();

    return p.run().then(() => {
      p.is_running().should.be.true;
      p.terminate();
      p.is_running().should.be.false;
    });
  });

  /**
   * @name instantiates an object:
   */
  it('instantiates an object', () => {

    let p = new base_proxy();

    return p.run(dependencies, { hello: 'memes' }).then(() => {
      p.is_running().should.be.true;
      p.call('options').should.eventually.equal(engine_options);
    });
  });

  /**
   * @name handles instantiation errors:
   */
  it('handles instantiation errors', () => {

    let p = new instantiation_error_proxy();

    p.run().should.eventually.be.rejectedWith(
      exception.runtime
    );
  });

  /**
   * @name handles method errors:
   */
  it('handles method errors', () => {

    let p = new method_error_proxy();

    return p.run().then(() => {
      p.call('sync').should.eventually.be.rejectedWith(exception.runtime);
      p.call('none').should.eventually.be.rejectedWith(exception.argument);
      p.call('async').should.eventually.be.rejectedWith(exception.runtime);
    });
  });

  /**
   * @name handles promises:
   */
  it('handles promises', () => {

    let p = new promise_proxy();
    let expect = { memes: 'dank' };

    return p.run(dependencies).then(() => {
      p.is_running().should.be.true;
      p.call('resolve', expect).should.eventually.equal(expect);
    });
  });

  /**
   * @name instantiates an engine:
   */
  it('instantiates an engine', () => {

    if (utils.is_node()) {
      return;
    }

    let p = new engine_proxy();

    return p.run(dependencies, engine_options).then(() => {
      p.is_running().should.be.true;
      return p.call('options');
    }).then((_options) => {
      _options.should.deep.equal(engine_options);
      return p.call('start');
    }).then(() => {
      return p.call('create_node', {
        title: 'Test', content: 'Hello, world!'
      });
    }).then((_rv) => {
      _rv.node.zone_id.should.equal(0);
      _rv.index.zone_id.should.equal(0);
      _rv.node.tags.should.deep.equal([]);
      _rv.node._id.should.equal(_rv.index.node_id);
      _rv.index._id.should.equal(_rv.node.index_id);
      return p.call('get_node', _rv.node._id);
    }).then((_node) => {
      _node.content.should.equal('Hello, world!');
    });
  });
});

